
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxCore.cpp */


#include "bxCore.h"
#include "bxCoreSub.h"
#include "bxCoreGblSub.h"
#include "bxRect.h"
#include "bxTri.h"

using namespace NameBeziex;



BxErr  BxCore::
MakeVHandle( EdgeVertexInfoBoundRot0& src, bool& isEnable, BxVec3& dst )
{
	BxErr  result;

	result = GBxCoreSub()->IsEnable( src, isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	result = GBxEdge()->MakeVHandle( src, dst );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxCore::
MakeBNormal( SurfaceInfo& info, bool& isEnable )
{
	BxErr  result;

	result = GBxCoreSub()->IsEnable( info, isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	for( int i=0; i<info.numVertex; i++ ) {
		result = GBxSurface()->MakeBNormal( i, info.numVertex, info.vertex );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxCore::
ConvSurface( SurfaceInfo& src, bool& isEnable, PatchRec& dst )
{
	RectPatch*  rectInfo;
	TriPatch*   triInfo;
	BxErr       result;

	result = GBxCoreSub()->IsEnable( src, isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	dst.numVertex = src.numVertex;

	switch( src.numVertex ) {
		case 3:
			triInfo = &( TriPatch& )dst;
			result = GBxTri()->ConvSurface( src.vertex, triInfo->tri );
			if( result == kBxErrFailure )
				return kBxErrFailure;
			break;
		case 4:
			rectInfo = &( RectPatch& )dst;
			result = GBxRect()->ConvSurface( src.vertex, rectInfo->rect );
			if( result == kBxErrFailure )
				return kBxErrFailure;
			break;
		default:
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}


BxCore::InfoDivRect::
InfoDivRect()
{
	boundVertex = 0;
	inVertex    = 0;
	face        = 0;
}

BxCore::InfoDivRect::
~InfoDivRect()
{
	if( boundVertex != 0 )
		delete[] boundVertex;
	if( inVertex != 0 )
		delete[] inVertex;
	if( face != 0 )
		delete[] face;
}

BxCore::InfoDivTri::
InfoDivTri()
{
	boundVertex = 0;
	inVertex    = 0;
	face        = 0;
}

BxCore::InfoDivTri::
~InfoDivTri()
{
	if( boundVertex != 0 )
		delete[] boundVertex;
	if( inVertex != 0 )
		delete[] inVertex;
	if( face != 0 )
		delete[] face;
}


int  BxCore::
SizeDivBoundVertex( const int numDiv )
{
	return numDiv - 1;
}

int  BxCore::
SizeDivInVertex( const int numVertex, const int numDiv )
{
	switch( numVertex ) {
	case 4:
		return ( numDiv - 1 ) * ( numDiv - 1 );
	case 3:
		if( numDiv <= 2 )
			return 0;
		return ( numDiv - 2 ) * ( numDiv - 1 ) / 2;
	}

	return -1;
}

int  BxCore::
SizeDivFace( const int numVertex, const int numDiv )
{
	switch( numVertex ) {
	case 4:
		return numDiv * numDiv;
	case 3:
		return numDiv * numDiv;
	}

	return -1;
}

void  BxCore::
InitDivRect( const int numDiv, DivRect& dst )
{
	GBxRect()->InitDiv( numDiv, dst );
}

void  BxCore::
InitDivTri( const int numDiv, DivTri& dst )
{
	GBxTri()->InitDiv( numDiv, dst );
}

BxErr  BxCore::
DivideRect( const RectPatchInfo& src, DivRect& dst )
{
	return GBxRect()->Divide( src, dst );
}

BxErr  BxCore::
DivideTri( const TriPatchInfo& src, DivTri& dst )
{
	return GBxTri()->Divide( src, dst );
}


////////
