
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxCore.h */


#if !defined( BX_CORE_H )
#define BX_CORE_H

#include "bxMath.h"


class BxCore {
public:
	enum {
		kVertexInFace,
		kVertexOnBound,
		kVertexOnVertex
	};

public:
	struct EdgeVertexInfoBoundRot0 {
		BxVec3  radialVec;
		BxVec3  vertexNormal[ 2 ];
	};


	struct SurfaceInfoEdge {
		BxVec3  wingVec;
		BxVec3  wingNormal[ 2 ];
		BxDir   wingBoundRot;
		BxVec2  ihandle;
		bool    wingPairEnable;
		bool    wingPairEnableBHandle;
		BxVec3  wingPairNormal[ 2 ];
		BxVec3  bNormal;
	};
	struct SurfaceInfoVertex {
		BxVec3           vertexPos;
		BxVec3           vertexNormal;
		BxDir            boundRot;
		bool             pairEnable;
		bool             pairEnableBHandle;
		SurfaceInfoEdge  edge[ 2 ];
	};
	struct SurfaceInfo {
		int                numVertex;
		SurfaceInfoVertex  vertex[ 4 ];
	};


	struct PatchRec {
		int  numVertex;
	};

	struct RectPatchInfo {
		BxVecAry3  bound[ 4 ];
		BxVecAry5  in[ 4 ];
	};
	struct RectPatch : public PatchRec {
		RectPatchInfo  rect;
	};

	struct TriPatchIn {
		BxVecAry2  dt2;
		BxVecAry5  dt5;
	};
	struct TriPatchInfo {
		BxVecAry3   bound[ 3 ];
		TriPatchIn  in[ 3 ];
	};
	struct TriPatch : public PatchRec {
		TriPatchInfo  tri;
	};


	struct DivVertexFace {
		int  faceNo;
		int  vertexInFace;
	};
	struct DivVertexVertex {
		BxVec3         pos;
		DivVertexFace  face;
	};

	struct DivBoundVertexRect {
		BxVec3         pos;
		DivVertexFace  face[ 2 ];
	};
	struct DivInVertexRect {
		BxVec3         pos;
		DivVertexFace  face[ 4 ];
	};
	struct DivBoundVertexTri {
		BxVec3         pos;
		DivVertexFace  face[ 3 ];
	};
	struct DivInVertexTri {
		BxVec3         pos;
		DivVertexFace  face[ 6 ];
	};

	struct DivFaceVertex {
		int  kind;
		int  no;
	};
	struct DivFace {
		DivFaceVertex  vertexNo;
		BxVec3         normal;
	};

	struct DivPatch {
		int  numVertex;
		int  numDiv;
	};

	struct DivFaceRect {
		DivFace  vertex[ 4 ];
	};
	struct InfoDivRect {
		DivVertexVertex      vertexVertex[ 4 ];
		DivBoundVertexRect*  boundVertex;        // Array size : SizeDivBoundVertex( numDiv ) * 4;
		DivInVertexRect*     inVertex;           // Array size : SizeDivInVertex( 4, numDiv );
		DivFaceRect*         face;               // Array size : SizeDivFace( 4, numDiv );
	public:
		InfoDivRect();
		virtual ~InfoDivRect();
	};
	struct DivRect : public DivPatch {
		InfoDivRect  rect;
	};

	struct DivFaceTri {
		DivFace  vertex[ 3 ];
	};
	struct InfoDivTri {
		DivVertexVertex     vertexVertex[ 3 ];
		DivBoundVertexTri*  boundVertex;         // Array size : SizeDivBoundVertex( numDiv ) * 3;
		DivInVertexTri*     inVertex;            // Array size : SizeDivInVertex( 3, numDiv );
		DivFaceTri*         face;                // Array size : SizeDivFace( 3, numDiv );
	public:
		InfoDivTri();
		virtual ~InfoDivTri();
	};
	struct DivTri : public DivPatch {
		InfoDivTri  tri;
	};

public:
	BxErr
	MakeVHandle( EdgeVertexInfoBoundRot0& src, bool& isEnable, BxVec3& dst );

	BxErr
	MakeBNormal( SurfaceInfo& src, bool& isEnable );

	BxErr
	ConvSurface( SurfaceInfo& src, bool& isEnable, PatchRec& dst );

	int
	SizeDivBoundVertex( const int numDiv );
	int
	SizeDivInVertex( const int numVertex, const int numDiv );
	int
	SizeDivFace( const int numVertex, const int numDiv );

	void
	InitDivRect( const int numDiv, DivRect& dst );
	void
	InitDivTri( const int numDiv, DivTri& dst );
	BxErr
	DivideRect( const RectPatchInfo& src, DivRect& dst );
	BxErr
	DivideTri( const TriPatchInfo& src, DivTri& dst );
};


#endif // #if !defined( BX_CORE_H )

////////
