
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxCoreSub.cpp */


#include <math.h>

#include "bxCoreSub.h"
#include "bxEdgeErrChk.h"
#include "bxSurfaceErrChk.h"

using namespace NameBeziex;



BxErr  BxCoreSub::
IsEnable( const BxCore::EdgeVertexInfoBoundRot0& data, bool& isEnable )
{
	BxEdgeErrChk  data0;

	data0.Conv( data );

	return data0.IsEnable( isEnable );
}

BxErr  BxCoreSub::
IsEnable( BxCore::SurfaceInfo& data, bool& isEnable )
{
	BxSurfaceErrChk  data0;

	data0.Conv( data );

	BxErr  result = data0.IsEnable( isEnable );
	if( result == kBxErrSuccess )
		data0.SetPairEnable( data );

	return result;
}

BxErr  BxCoreSub::
SameNormal( const BxVec3& normal0, const BxVec3& normal1, bool& sameNormal, BxVec3* aveNormal )
{
	BxErr  errResult;

	sameNormal = false;
	
	double  sinDt = Cross( normal0, normal1 ).Length();
	if( sinDt < MaxSinVertexNormal() ) {
		sameNormal = true;

		if( aveNormal != 0 ) {
			*aveNormal = ( normal0 + normal1 ).Normalize( &errResult );
			if( errResult == kBxErrFailure )
				return kBxErrFailure;
		}
	}

	return kBxErrSuccess;
}

const double  BxCoreSub::
MaxSinVertexNormal()
{
	static const double  maxDeg = 0.5;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = sin( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
};


////////
