
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxEdgeErrChk.cpp */


#include "bxCoreGblSub.h"
#include "bxEdgeErrChk.h"
#include "bxSurfaceErrChk.h"

using namespace NameBeziex;



void  BxEdgeErrChk::
Conv( const BxCore::EdgeVertexInfoBoundRot0& src )
{
	radialVec = src.radialVec;

	for( int i=0; i<2; i++ )
		vertexNormal[ i ] = src.vertexNormal[ i ];
}

BxErr  BxEdgeErrChk::
IsEnable( bool& isEnable )
{
	IsEnableVertexPos( isEnable );
	if( isEnable == false )
		return kBxErrSuccess;

	IsEnableEdgeNormal( isEnable );
	if( isEnable == false )
		return kBxErrSuccess;

	BxErr  result = IsEnableRadial( isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	return kBxErrSuccess;
}

void  BxEdgeErrChk::
IsEnableVertexPos( bool& isEnable )
{
	double  len;

	isEnable = true;

	len = radialVec.Length();
	if( len < GBxNearZero() ) {
		isEnable = false;
		return;
	}
}

void  BxEdgeErrChk::
IsEnableEdgeNormal( bool& isEnable )
{
	isEnable = true;

	double  cosDt = Dot( vertexNormal[ 0 ], vertexNormal[ 1 ] );
	if( cosDt < BxSurfaceErrChkVertex::MaxAngCosEdgeNormal() )
		isEnable = false;
}

BxErr  BxEdgeErrChk::
IsEnableRadial( bool& isEnable )
{
	BxVec3  pairNormal[ 2 ], handleVec, radialVecUnit;
	BxErr   errResult;

	isEnable = true;
	
	pairNormal[ 0 ] = GBxZero3();
	pairNormal[ 1 ] = GBxZero3();

	BxErr  result = GBxOrgTo3D()->BHandleVec( vertexNormal, radialVec, 0.0, false, pairNormal, handleVec );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	radialVecUnit = radialVec.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	handleVec  = handleVec.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	double  cosDt = Dot( radialVecUnit, handleVec );
	if( cosDt < BxSurfaceErrChkVertex::MaxAngCosRadialToHandle() )
		isEnable = false;

	return kBxErrSuccess;
}


////////
