
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxMath.h */


#if !defined( BX_MATH_H )
#define BX_MATH_H

#include "bxBase.h"


class BxMathBase {
public:
	enum {
		kElmX,
		kElmY,
		kElmZ
	};
	enum {
		kElmU,
		kElmV
	};
	enum {
		kNumVec2  = 2,
		kNumVec3  = 3
	};
};



class BxMat2x3;
class BxVec2;

class BxVec2 : public BxMathBase {
protected:
	double  fDt[ kNumVec2 ];

public:
	BxVec2()    {  }
	BxVec2( const BxVec2& a );
	BxVec2( const double x, const double y );

	double&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const double&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	double&
	X()        { return fDt[ kElmX ]; }
	double&
	Y()        { return fDt[ kElmY ]; }
	const double&
	X() const  { return fDt[ kElmX ]; }
	const double&
	Y() const  { return fDt[ kElmY ]; }

	double&
	U()        { return fDt[ kElmU ]; }
	double&
	V()        { return fDt[ kElmV ]; }
	const double&
	U() const  { return fDt[ kElmU ]; }
	const double&
	V() const  { return fDt[ kElmV ]; }

	BxVec2&
	operator+=( const BxVec2& a );
	BxVec2&
	operator-=( const BxVec2& a );
	BxVec2&
	operator*=( const double a );
	BxVec2&
	operator/=( const double a );

	const BxVec2
	operator+( const BxVec2& a ) const;
	const BxVec2
	operator-( const BxVec2& a ) const;
	const BxVec2
	operator/( const double a ) const;

	friend BxVec2
	operator*( const BxVec2& a, const double b );
	friend BxVec2
	operator*( const double a, const BxVec2& b );

	friend double
	Dot( const BxVec2& a, const BxVec2& b );
	friend double
	Cross( const BxVec2& a, const BxVec2& b );
	friend BxVec2
	Arc( const BxVec2& vecPntToPnt, const BxVec2& unitTan, BxErr* errResult = 0 );

	const double
	Length() const;
	const double
	LengthSquared() const;
	const BxVec2
	Normalize() const;
	const BxVec2
	Normalize( const BxVec2& resultForErr ) const;
	const BxVec2
	Normalize( BxErr* errResult ) const;

	const BxVec2
	operator-() const;
	const BxVec2
	Abs() const;

	const BxVec2
	Point( const BxMat2x3& a ) const;
	const BxVec2
	Vector( const BxMat2x3& a ) const;

};

class BxDir : public BxVec2 {
public:
	BxDir()    {  }
	BxDir( const BxDir& a );
	BxDir( const BxVec2& a );
	BxDir( const double x, const double y );
	BxDir( const double angle );

	const BxDir
	operator-() const;

	const double
	Sin( BxErr* errResult = 0 ) const;
	const double
	Cos( BxErr* errResult = 0 ) const;
	const double
	SignedAngle( BxErr* errResult = 0 ) const;
	const double
	UnsignedAngle( BxErr* errResult = 0 ) const;

	bool
	GreaterSignedAngle( const BxDir& cmpDt, BxErr* errResult = 0 );
	bool
	LessSignedAngle( const BxDir& cmpDt, BxErr* errResult = 0 );
	bool
	GreaterUnsignedAngle( const BxDir& cmpDt, BxErr* errResult = 0 );
	bool
	LessUnsignedAngle( const BxDir& cmpDt, BxErr* errResult = 0 );

private:
	BxErr
	SignedAngleSub( const BxDir& cmpDt, double& ang0, double& ang1 );
	BxErr
	UnsignedAngleSub( const BxDir& cmpDt, double& ang0, double& ang1 );
};



class BxMat3x4;

class BxVec3 : public BxMathBase {
private:
	double  fDt[ kNumVec3 ];

public:
	BxVec3()    {  }
	BxVec3( const BxVec3& a );
	BxVec3( const double x, const double y, const double z );

	double&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const double&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	double&
	X()        { return fDt[ kElmX ]; }
	double&
	Y()        { return fDt[ kElmY ]; }
	double&
	Z()        { return fDt[ kElmZ ]; }
	const double&
	X() const  { return fDt[ kElmX ]; }
	const double&
	Y() const  { return fDt[ kElmY ]; }
	const double&
	Z() const  { return fDt[ kElmZ ]; }

	BxVec3&
	operator+=( const BxVec3& a );
	BxVec3&
	operator-=( const BxVec3& a );
	BxVec3&
	operator*=( const double a );
	BxVec3&
	operator/=( const double a );

	const BxVec3
	operator+( const BxVec3& a ) const;
	const BxVec3
	operator-( const BxVec3& a ) const;
	const BxVec3
	operator/( const double a ) const;

	friend BxVec3
	operator*( const BxVec3& a, const double b );
	friend BxVec3
	operator*( const double a, const BxVec3& b );

	friend double
	Dot( const BxVec3& a, const BxVec3& b );
	friend BxVec3
	Cross( const BxVec3& a, const BxVec3& b );
	friend BxVec3
	Arc( const BxVec3& vecPntToPnt, const BxVec3& unitTan, BxErr* errResult = 0 );

	const double
	Length() const;
	const double
	LengthSquared() const;
	const BxVec3
	Normalize() const;
	const BxVec3
	Normalize( const BxVec3& resultForErr ) const;
	const BxVec3
	Normalize( BxErr* errResult ) const;

	const BxVec3
	operator-() const;
	const BxVec3
	Abs() const;

	const BxVec3
	Point( const BxMat3x4& a ) const;
	const BxVec3
	Vector( const BxMat3x4& a ) const;
};



class BxAxisXYZ;

class BxAxis : public BxMathBase {
public:
	enum {
		kAxisMain,
		kAxisSub,
		kNumAxis
	};

private:
	BxVec3  fDt[ kNumAxis ];

public:
	BxAxis()    {  }
	BxAxis( const BxAxis& a );
	BxAxis( const BxVec3& mainAxis, const BxVec3& subAxis );

	BxVec3&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const BxVec3&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	BxVec3&
	MainAxis()        { return fDt[ kAxisMain ]; }
	BxVec3&
	SubAxis()         { return fDt[ kAxisSub  ]; }
	const BxVec3&
	MainAxis() const  { return fDt[ kAxisMain ]; }
	const BxVec3&
	SubAxis()  const  { return fDt[ kAxisSub  ]; }

	const BxAxis
	RegularAxis( BxErr* errResult = 0 ) const;
	friend BxAxis
	MakeSubAxis( const BxVec3& mainAxis, BxErr* errResult = 0 );

	const BxVec3
	GetAxisZ( BxErr* errResult = 0 ) const;
	const BxAxisXYZ
	Conv( BxErr* errResult = 0 ) const;
};



class BxAxisXYZ : public BxMathBase {
public:
	enum {
		kAxisX,
		kAxisY,
		kAxisZ,
		kNumAxisXYZ
	};

private:
	BxVec3  fDt[ kNumAxisXYZ ];

public:
	BxAxisXYZ()    {  }
	BxAxisXYZ( const BxAxisXYZ& a );
	BxAxisXYZ( const BxVec3& axisX, const BxVec3& axisY, const BxVec3& axisZ );

	BxVec3&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const BxVec3&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	BxVec3&
	AxisX()        { return fDt[ kAxisX ]; }
	BxVec3&
	AxisY()        { return fDt[ kAxisY ]; }
	BxVec3&
	AxisZ()        { return fDt[ kAxisZ ]; }
	const BxVec3&
	AxisX() const  { return fDt[ kAxisX ]; }
	const BxVec3&
	AxisY() const  { return fDt[ kAxisY ]; }
	const BxVec3&
	AxisZ() const  { return fDt[ kAxisZ ]; }
};



class BxMat2x3 : public BxMathBase {
public:
	class Base {
	private:
		double  fDt[ 3 ];

	public:
		double&
		operator[]( const int v )        { return fDt[ v ]; }
		const double&
		operator[]( const int v ) const  { return fDt[ v ]; }
	};

private:
	Base  fDt[ 2 ];

public:
	BxMat2x3();
	BxMat2x3( const BxMat2x3& a );
	BxMat2x3( const double a[ 2 ][ 3 ] );
	BxMat2x3( const double a00, const double a10,
	          const double a01, const double a11,
		  const double a02, const double a12 );

	Base&
	operator[]( const int u )        { return fDt[ u ]; }
	const Base&
	operator[]( const int u ) const  { return fDt[ u ]; }

	friend BxMat2x3
	operator*( const BxMat2x3& a, const BxMat2x3& b );

	friend BxMat2x3
	Translete( const double x, const double y );
	friend BxMat2x3
	Scale( const double x, const double y );
	friend BxMat2x3
	Rotate( const double a );
	friend BxMat2x3
	Rotate( const BxDir& a, BxErr* errResult = 0 );

private:
	BxMat2x3&
	RotateSub( const double dtSin, const double dtCos );
};



class BxMat3x4 : public BxMathBase {
public:
	class Base {
	private:
		double  fDt[ 4 ];

	public:
		double&
		operator[]( const int v )        { return fDt[ v ]; }
		const double&
		operator[]( const int v ) const  { return fDt[ v ]; }
	};

private:
	Base  fDt[ 3 ];

public:
	BxMat3x4();
	BxMat3x4( const BxMat3x4& a );
	BxMat3x4( const double a[ 3 ][ 4 ] );
	BxMat3x4( const double a00, const double a10, const double a20,
	          const double a01, const double a11, const double a21,
		  const double a02, const double a12, const double a22,
		  const double a03, const double a13, const double a23 );

	Base&
	operator[]( const int u )        { return fDt[ u ]; }
	const Base&
	operator[]( const int u ) const  { return fDt[ u ]; }

	friend BxMat3x4
	operator*( const BxMat3x4& a, const BxMat3x4& b );

	friend BxMat3x4
	Translete( const double x, const double y, const double z );
	friend BxMat3x4
	Scale( const double x, const double y, const double z );
	friend BxMat3x4
	RotateX( const double x );
	friend BxMat3x4
	RotateY( const double y );
	friend BxMat3x4
	RotateZ( const double z );
	friend BxMat3x4
	RotateAxis( const double angle, const BxVec3& axis, BxErr* errResult = 0 );
	friend BxMat3x4
	Scale( const BxDir& a, BxErr* errResult );
	friend BxMat3x4
	RotateX( const BxDir& a, BxErr* errResult = 0 );
	friend BxMat3x4
	RotateY( const BxDir& a, BxErr* errResult = 0 );
	friend BxMat3x4
	RotateZ( const BxDir& a, BxErr* errResult = 0 );
	friend BxMat3x4
	RotateAxis( const BxDir& a, const BxVec3& axis, BxErr* errResult = 0 );

	friend BxMat3x4
	RotWorldToLocal( const BxAxis& axis, BxErr* errResult );
	friend BxMat3x4
	RotLocalToWorld( const BxAxis& axis, BxErr* errResult );
	friend BxMat3x4
	RotateAxis( const BxAxis& srcAxis, const BxAxis& dstAxis, BxErr* errResult = 0 );

private:
	BxMat3x4&
	RotateXSub( const double dtSin, const double dtCos );
	BxMat3x4&
	RotateYSub( const double dtSin, const double dtCos );
	BxMat3x4&
	RotateZSub( const double dtSin, const double dtCos );
	BxMat3x4&
	RotateAxisSub( const double dtSin, const double dtCos, const BxVec3& axis, BxErr& errResult );

	friend BxMat3x4
	RotWorldLocalSub( const BxAxisXYZ& srcAxis, const BxAxisXYZ& dstAxis );
};



struct BxVecAry2 {
	BxVec3  p[ 2 ];
};
struct BxVecAry3 {
	BxVec3  p[ 3 ];
};
struct BxVecAry4 {
	BxVec3  p[ 4 ];
};
struct BxVecAry5 {
	BxVec3  p[ 5 ];
};
struct BxVecAry6 {
	BxVec3  p[ 6 ];
};
struct BxVecAry7 {
	BxVec3  p[ 7 ];
};



class Bezier3Line {
private:
	BxVec3  fDt[ 4 ];

public:
	Bezier3Line()    {  }
	Bezier3Line( const Bezier3Line& a );
	Bezier3Line( const BxVecAry4& a );
	Bezier3Line( const BxVec3 a[ 4 ] );

	BxVec3&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const BxVec3&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	friend BxVec3
	Pos( const Bezier3Line& bezierLine, const double t );
	friend BxVec3
	Diff( const Bezier3Line& bezierLine, const double t );
};

class Bezier6Line {
private:
	BxVec3  fDt[ 7 ];

public:
	Bezier6Line()    {  }
	Bezier6Line( const Bezier6Line& a );
	Bezier6Line( const BxVecAry7& a );
	Bezier6Line( const BxVec3 a[ 7 ] );

	BxVec3&
	operator[]( const int idx )        { return fDt[ idx ]; }
	const BxVec3&
	operator[]( const int idx ) const  { return fDt[ idx ]; }

	friend BxVec3
	Pos( const Bezier6Line& bezierLine, const double t );
	friend BxVec3
	Diff( const Bezier6Line& bezierLine, const double t );
};



namespace NameBeziex {
	const double     GBxPi();

	const BxVec2     GBxZero2();
	const BxVec3     GBxZero3();

	const double     GBxNearZero();
	const BxVec2     GBxNearZero2();
	const BxVec3     GBxNearZero3();
	const BxVec2     GBxMinusNearZero2();
	const BxVec3     GBxMinusNearZero3();

	const BxDir      GBxZeroDir();
	const BxAxisXYZ  GBxWorldAxis();

	const double     GBxNegateAngle( const double angle );
	const double     GBxSignedRegularAngle( const double angle );
	const double     GBxUnsignedRegularAngle( const double angle );
};


#endif  // !defined( BX_MATH_H )

////////
