
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxOrgTo3D.cpp */


#include <math.h>
#include "bxOrgTo3D.h"

using namespace NameBeziex;



BxErr  BxOrgTo3D::
BHandleVec( const BxVec3 normal[ 2 ], const BxVec3& radialVec, const BxDir& boundRot, const bool pairEnableBHandle,
	const BxVec3 pairNormal[ 2 ], BxVec3& dst )
{
	BxVec3  normal0[ 2 ];
	BxErr   errResult;

	for( int i=0; i<2; i++ ) {
		if( pairEnableBHandle == true ) {
			normal0[ i ] = ( normal[ i ] + pairNormal[ i ] ).Normalize( &errResult );
			if( errResult == kBxErrFailure )
				return kBxErrFailure;
		}
		else
			normal0[ i ] = normal[ i ];
	}

	return BHandleVecSub( normal0, radialVec, boundRot, dst );
}

BxErr  BxOrgTo3D::
BHandleVecSub( const BxVec3 normal[ 2 ], const BxVec3& radialVec, const BxDir& boundRot, BxVec3& dst )
{
	BxErr  errResult;

	BxVec3  normal0     = normal[ 0 ] + normal[ 1 ];
	BxVec3  boundNormal = Cross( radialVec, normal0 ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxMat3x4  mat = RotateAxis( boundRot, radialVec, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	boundNormal = boundNormal.Vector( mat );

	BxVec3  handleVec0 = Cross( normal[ 0 ], boundNormal ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxVec3  handleVec = Arc( radialVec, handleVec0, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	dst = handleVec;

	return kBxErrSuccess;
}

BxErr  BxOrgTo3D::
IHandleVec( const BxVec3& vNormal, const BxVec3& bNormal, const BxVec3& handleOut, const BxVec3& vecHandleToHandle,
	const BxVec2& srcIHandle, BxVec3& dst )
{
	BxErr   errResult;
	BxVec3  unitIHandleTmp;

	unitIHandleTmp = Cross( vNormal, bNormal );
	if( unitIHandleTmp.Length() < GBxNearZero() )
		unitIHandleTmp = handleOut;
	else if( Dot( handleOut, unitIHandleTmp ) < 0 )
		unitIHandleTmp = -unitIHandleTmp;

	BxVec3  unitIHandle = unitIHandleTmp.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxVec3  iHandleVec0 = Arc( vecHandleToHandle, unitIHandle, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxMat3x4  mat1 = RotateAxis( srcIHandle, bNormal, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxMat3x4  mat2 = Scale( srcIHandle, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;
	
	dst = iHandleVec0.Vector( mat1 * mat2 );

	return kBxErrSuccess;
}

BxErr  BxOrgTo3D::
TriVHandleVec( const BxVec3& vNormal, const BxVec3& bNormal, const BxVec3& handleOut, const BxVec3& vecHandleToHandle,
	const BxVec2& srcIHandle, const BxVec3& normal, BxVec3& dst )
{
	BxVec3  antiIHandle, iHandleNormal;
	BxErr   errResult;

	BxErr  result = IHandleVec( vNormal, bNormal, handleOut, vecHandleToHandle, srcIHandle, antiIHandle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	BxVec3  mVecHandleToHandle = -vecHandleToHandle;

	BxVec3  unitProjVec = Cross( normal, Cross( mVecHandleToHandle, normal ) ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;
	BxVec3  projPnt = Dot( unitProjVec, vecHandleToHandle ) * unitProjVec;
	double  len     = BxVec3( projPnt - vecHandleToHandle ).Length();

	BxVec3  iHandleAxis2    = Cross( vecHandleToHandle, antiIHandle );
	double  lenIHandleAxis2 = iHandleAxis2.Length();
	if( lenIHandleAxis2 < GBxNearZero() )
		iHandleNormal = normal;
	else {
		iHandleAxis2 /= lenIHandleAxis2;
		
		iHandleNormal = Cross( iHandleAxis2, vecHandleToHandle ).Normalize( &errResult );
		if( errResult == kBxErrFailure )
			return kBxErrFailure;

		if( Dot( iHandleNormal, normal ) < 0.0 )
			iHandleNormal = -iHandleNormal;
	}

	double  cosDt = Dot( normal, iHandleNormal );
	if( cosDt < GBxNearZero() ) {
		if( len >= GBxNearZero() )
			return kBxErrFailure;

		iHandleNormal = GBxZero3();
	}
	else {
		iHandleNormal = iHandleNormal * len / cosDt;
	}

	BxVec3  unitIHandle = BxVec3( mVecHandleToHandle + iHandleNormal ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxVec3  iHandleVec0 = Arc( mVecHandleToHandle, unitIHandle, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxMat3x4  mat = Scale( srcIHandle, &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;
	
	dst = iHandleVec0.Vector( mat );

	return kBxErrSuccess;
}

BxErr  BxOrgTo3D::
BNormal( const BxVec3& vertexNormal, const BxVec3& radialVec, const BxVec3& iHandleAxis, const BxVec3& vecHandleToHandle,
	BxVec3& bNormal )
{
	BxErr  errResult;

	BxVec3  radialVecE = radialVec.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	double  lenCross  = Dot( vertexNormal, radialVecE );
	double  lenNormal = sqrt( 1.0 - lenCross * lenCross );

	BxVec3  iHandleAxisE = iHandleAxis.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;
	BxVec3  vecHandleToHandleE = vecHandleToHandle.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;
	BxVec3  bNormalTmp = Cross( iHandleAxisE, vecHandleToHandleE ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	bNormal = ( lenCross * vecHandleToHandleE + lenNormal * bNormalTmp ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}


////////
