
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxRect.cpp */


#include "bxCoreGbl.h"
#include "bxCoreGblSub.h"
#include "bxRect.h"
#include "bxVector.h"

using namespace NameBeziex;



BxErr  BxRect::
ConvSurface( const BxCore::SurfaceInfoVertex vertex[ 4 ], BxCore::RectPatchInfo& rect )
{
	int    i;
	BxErr  result;
	
	for( i=0; i<4; i++ ) {
		int  m = ( i + 4 - 1 ) % 4;

		result = GBxVector()->Pos2( m, 4, vertex, rect.bound[ i ].p[ 0 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		GBxVector()->Pos0( i, 4, vertex, rect.bound[ i ].p[ 1 ] );
		result = GBxVector()->Pos1( i, 4, vertex, rect.bound[ i ].p[ 2 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	for( i=0; i<4; i++ ) {
		result = GBxSurface()->IHandlePos( i, 4, vertex, rect.in[ i ].p );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

void  BxRect::
RectToPatch4774( const BxCore::RectPatchInfo& rect, BxSurface::Patch4774& patch )
{
	patch.v0.p[ 0 ] = rect.bound[ 0 ].p[ 1 ];
	patch.v0.p[ 1 ] = rect.bound[ 0 ].p[ 2 ];
	patch.v0.p[ 2 ] = rect.bound[ 1 ].p[ 0 ];
	patch.v0.p[ 3 ] = rect.bound[ 1 ].p[ 1 ];

	patch.v1.p[ 0 ] = rect.bound[ 0 ].p[ 0 ];
	patch.v1.p[ 1 ] = rect.in[ 0 ].p[ 0 ];
	patch.v1.p[ 2 ] = rect.in[ 0 ].p[ 1 ];
	patch.v1.p[ 3 ] = rect.in[ 0 ].p[ 2 ];
	patch.v1.p[ 4 ] = rect.in[ 0 ].p[ 3 ];
	patch.v1.p[ 5 ] = rect.in[ 0 ].p[ 4 ];
	patch.v1.p[ 6 ] = rect.bound[ 1 ].p[ 2 ];

	patch.v2.p[ 0 ] = rect.bound[ 3 ].p[ 2 ];
	patch.v2.p[ 1 ] = rect.in[ 2 ].p[ 4 ];
	patch.v2.p[ 2 ] = rect.in[ 2 ].p[ 3 ];
	patch.v2.p[ 3 ] = rect.in[ 2 ].p[ 2 ];
	patch.v2.p[ 4 ] = rect.in[ 2 ].p[ 1 ];
	patch.v2.p[ 5 ] = rect.in[ 2 ].p[ 0 ];
	patch.v2.p[ 6 ] = rect.bound[ 2 ].p[ 0 ];

	patch.v3.p[ 0 ] = rect.bound[ 3 ].p[ 1 ];
	patch.v3.p[ 1 ] = rect.bound[ 3 ].p[ 0 ];
	patch.v3.p[ 2 ] = rect.bound[ 2 ].p[ 2 ];
	patch.v3.p[ 3 ] = rect.bound[ 2 ].p[ 1 ];
}

void  BxRect::
RectToPatch4774T( const BxCore::RectPatchInfo& rect, BxSurface::Patch4774& patch )
{
	patch.v0.p[ 0 ] = rect.bound[ 0 ].p[ 1 ];
	patch.v0.p[ 1 ] = rect.bound[ 0 ].p[ 0 ];
	patch.v0.p[ 2 ] = rect.bound[ 3 ].p[ 2 ];
	patch.v0.p[ 3 ] = rect.bound[ 3 ].p[ 1 ];

	patch.v1.p[ 0 ] = rect.bound[ 0 ].p[ 2 ];
	patch.v1.p[ 1 ] = rect.in[ 3 ].p[ 4 ];
	patch.v1.p[ 2 ] = rect.in[ 3 ].p[ 3 ];
	patch.v1.p[ 3 ] = rect.in[ 3 ].p[ 2 ];
	patch.v1.p[ 4 ] = rect.in[ 3 ].p[ 1 ];
	patch.v1.p[ 5 ] = rect.in[ 3 ].p[ 0 ];
	patch.v1.p[ 6 ] = rect.bound[ 3 ].p[ 0 ];

	patch.v2.p[ 0 ] = rect.bound[ 1 ].p[ 0 ];
	patch.v2.p[ 1 ] = rect.in[ 1 ].p[ 0 ];
	patch.v2.p[ 2 ] = rect.in[ 1 ].p[ 1 ];
	patch.v2.p[ 3 ] = rect.in[ 1 ].p[ 2 ];
	patch.v2.p[ 4 ] = rect.in[ 1 ].p[ 3 ];
	patch.v2.p[ 5 ] = rect.in[ 1 ].p[ 4 ];
	patch.v2.p[ 6 ] = rect.bound[ 2 ].p[ 2 ];

	patch.v3.p[ 0 ] = rect.bound[ 1 ].p[ 1 ];
	patch.v3.p[ 1 ] = rect.bound[ 1 ].p[ 2 ];
	patch.v3.p[ 2 ] = rect.bound[ 2 ].p[ 0 ];
	patch.v3.p[ 3 ] = rect.bound[ 2 ].p[ 1 ];
}

void  BxRect::
PosSurface( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& pos )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			pos = rect.bound[ 0 ].p[ 1 ];
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			pos = rect.bound[ 1 ].p[ 1 ];
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		if( uv.U() < GBxNearZero() ) {
			pos = rect.bound[ 3 ].p[ 1 ];
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			pos = rect.bound[ 2 ].p[ 1 ];
			return;
		}
	}

	PosSurfaceMain( rect, uv, pos );
}


void  BxRect::
PosSurfaceMain( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& pos )
{
	BxVec3  posU, posV;

	PosSurfaceU( rect, uv, posU );
	PosSurfaceV( rect, uv, posV );

	double  uu = uv.U() * ( 1.0 - uv.U() );
	double  vv = uv.V() * ( 1.0 - uv.V() );

	pos = ( vv * posU + uu * posV ) / ( vv + uu );
}

void  BxRect::
PosSurfaceU( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& posU )
{
	BxSurface::Patch4774  patch;
	BxVec2                uv2;

	RectToPatch4774T( rect, patch );

	uv2.U() = uv.V();
	uv2.V() = uv.U();

	GBxSurface()->PosSurface4774( patch, uv2, posU );
}

void  BxRect::
PosSurfaceV( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& posV )
{
	BxSurface::Patch4774  patch;

	RectToPatch4774( rect, patch );
	GBxSurface()->PosSurface4774( patch, uv, posV );
}

void  BxRect::
NormalSurface( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& normalVec )
{
	BxVec3  du, dv;

	DuSurface( rect, uv, du );
	DvSurface( rect, uv, dv );
	normalVec = Cross( du, dv ).Normalize( GBxZero3() );
}

void  BxRect::
DuSurface( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& du )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			du = ( rect.bound[ 0 ].p[ 2 ] - rect.bound[ 0 ].p[ 1 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			du = ( rect.bound[ 1 ].p[ 1 ] - rect.bound[ 1 ].p[ 0 ] ) * 3.0;
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		if( uv.U() < GBxNearZero() ) {
			du = ( rect.bound[ 3 ].p[ 0 ] - rect.bound[ 3 ].p[ 1 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			du = ( rect.bound[ 2 ].p[ 1 ] - rect.bound[ 2 ].p[ 2 ] ) * 3.0;
			return;
		}
	}

	DuSurfaceMain( rect, uv, du );
}

void  BxRect::
DuSurfaceMain( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& du )
{
	BxVec3  posU, posV, vecU, vecV;

	PosSurfaceU( rect, uv, posU );
	PosSurfaceV( rect, uv, posV );
	DuSurfaceU( rect, uv, vecU );
	DuSurfaceV( rect, uv, vecV );

	double  uu = uv.U() * ( 1.0 - uv.U() );
	double  vv = uv.V() * ( 1.0 - uv.V() );
	double  param0 = uu + vv;
	double  param1 = ( ( 1.0 - 2.0 * uv.U() ) * vv ) / ( param0 * param0 );

	du = param1 * ( -posU + posV ) + ( vv * vecU + uu * vecV ) / param0;
}

void  BxRect::
DuSurfaceU( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& vecU )
{
	BxSurface::Patch4774  patch;
	BxVec2                uv2;

	RectToPatch4774T( rect, patch );

	uv2.U() = uv.V();
	uv2.V() = uv.U();

	GBxSurface()->DvSurface4774( patch, uv2, vecU );
}

void  BxRect::
DuSurfaceV( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& vecV )
{
	BxSurface::Patch4774  patch;

	RectToPatch4774( rect, patch );
	GBxSurface()->DuSurface4774( patch, uv, vecV );
}

void  BxRect::
DvSurface( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& dv )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			dv = ( rect.bound[ 0 ].p[ 0 ] - rect.bound[ 0 ].p[ 1 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			dv = ( rect.bound[ 1 ].p[ 2 ] - rect.bound[ 1 ].p[ 1 ] ) * 3.0;
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		if( uv.U() < GBxNearZero() ) {
			dv = ( rect.bound[ 3 ].p[ 1 ] - rect.bound[ 3 ].p[ 2 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			dv = ( rect.bound[ 2 ].p[ 1 ] - rect.bound[ 2 ].p[ 0 ] ) * 3.0;
			return;
		}
	}

	DvSurfaceMain( rect, uv, dv );
}

void  BxRect::
DvSurfaceMain( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& dv )
{
	BxVec3  posU, posV, vecU, vecV;

	PosSurfaceU( rect, uv, posU );
	PosSurfaceV( rect, uv, posV );
	DvSurfaceU( rect, uv, vecU );
	DvSurfaceV( rect, uv, vecV );

	double  uu = uv.U() * ( 1.0 - uv.U() );
	double  vv = uv.V() * ( 1.0 - uv.V() );
	double  param0 = uu + vv;
	double  param1 = ( uu * ( 1.0 - 2.0 * uv.V() ) ) / ( param0 * param0 );

	dv = param1 * ( posU - posV ) + ( vv * vecU + uu * vecV ) / param0;
}

void  BxRect::
DvSurfaceU( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& vecU )
{
	BxSurface::Patch4774  patch;
	BxVec2                uv2;

	RectToPatch4774T( rect, patch );

	uv2.U() = uv.V();
	uv2.V() = uv.U();

	GBxSurface()->DuSurface4774( patch, uv2, vecU );
}

void  BxRect::
DvSurfaceV( const BxCore::RectPatchInfo& rect, const BxVec2& uv, BxVec3& vecV )
{
	BxSurface::Patch4774  patch;

	RectToPatch4774( rect, patch );
	GBxSurface()->DvSurface4774( patch, uv, vecV );
}

void  BxRect::
InitDiv( const int numDiv, BxCore::DivRect& dst )
{
	dst.numVertex = 4;
	dst.numDiv    = numDiv;

	if( dst.rect.boundVertex != 0 )
		delete[] dst.rect.boundVertex;
	if( dst.rect.inVertex != 0 )
		delete[] dst.rect.inVertex;
	if( dst.rect.face != 0 )
		delete[] dst.rect.face;

	dst.rect.boundVertex = new BxCore::DivBoundVertexRect[ GBxCore()->SizeDivBoundVertex( numDiv ) * 4 ];
	dst.rect.inVertex    = new BxCore::DivInVertexRect[ GBxCore()->SizeDivInVertex( 4, numDiv ) ];
	dst.rect.face        = new BxCore::DivFaceRect[ GBxCore()->SizeDivFace( 4, numDiv ) ];

	InitDivFaceBound( numDiv, dst.rect );
	InitDivFaceIn( numDiv, dst.rect );
}

void  BxRect::
InitDivFaceBound( const int numDiv, BxCore::InfoDivRect& rect )
{
	int  i, ofs;

	ofs = 0;
	for( i=0; i<( numDiv-1 ); i++ ) {
		rect.face[ i     ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ i     ].vertex[ 1 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 0 ].faceNo       = i;
		rect.boundVertex[ i + ofs ].face[ 0 ].vertexInFace = 1;

		rect.face[ i + 1 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ i + 1 ].vertex[ 0 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 1 ].faceNo       = i + 1;
		rect.boundVertex[ i + ofs ].face[ 1 ].vertexInFace = 0;
	}

	ofs = numDiv - 1;
	for( i=0; i<( numDiv-1 ); i++ ) {
		rect.face[ ( i + 1 ) * numDiv - 1 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ ( i + 1 ) * numDiv - 1 ].vertex[ 2 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 0 ].faceNo       = ( i + 1 ) * numDiv - 1;
		rect.boundVertex[ i + ofs ].face[ 0 ].vertexInFace = 2;

		rect.face[ ( i + 2 ) * numDiv - 1 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ ( i + 2 ) * numDiv - 1 ].vertex[ 1 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 1 ].faceNo       = ( i + 2 ) * numDiv - 1;
		rect.boundVertex[ i + ofs ].face[ 1 ].vertexInFace = 1;
	}

	ofs = ( numDiv - 1 ) * 2;
	for( i=0; i<( numDiv-1 ); i++ ) {
		rect.face[ numDiv * numDiv - 1 - i ].vertex[ 3 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ numDiv * numDiv - 1 - i ].vertex[ 3 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 0 ].faceNo       = numDiv * numDiv - 1 - i;
		rect.boundVertex[ i + ofs ].face[ 0 ].vertexInFace = 3;

		rect.face[ numDiv * numDiv - 2 - i ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ numDiv * numDiv - 2 - i ].vertex[ 2 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 1 ].faceNo       = numDiv * numDiv - 2 - i;
		rect.boundVertex[ i + ofs ].face[ 1 ].vertexInFace = 2;
	}

	ofs = ( numDiv - 1 ) * 3;
	for( i=0; i<( numDiv-1 ); i++ ) {
		rect.face[ numDiv * ( numDiv - 1 - i ) ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ numDiv * ( numDiv - 1 - i ) ].vertex[ 0 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 0 ].faceNo       = numDiv * ( numDiv - 1 - i );
		rect.boundVertex[ i + ofs ].face[ 0 ].vertexInFace = 0;

		rect.face[ numDiv * ( numDiv - 2 - i ) ].vertex[ 3 ].vertexNo.kind = BxCore::kVertexOnBound;
		rect.face[ numDiv * ( numDiv - 2 - i ) ].vertex[ 3 ].vertexNo.no   = i + ofs;
		rect.boundVertex[ i + ofs ].face[ 1 ].faceNo       = numDiv * ( numDiv - 2 - i );
		rect.boundVertex[ i + ofs ].face[ 1 ].vertexInFace = 3;
	}

	rect.face[ 0                       ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnVertex;
	rect.face[ 0                       ].vertex[ 0 ].vertexNo.no   = 0;
	rect.vertexVertex[ 0 ].face.faceNo       = 0;
	rect.vertexVertex[ 0 ].face.vertexInFace = 0;

	rect.face[ numDiv - 1              ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnVertex;
	rect.face[ numDiv - 1              ].vertex[ 1 ].vertexNo.no   = 1;
	rect.vertexVertex[ 1 ].face.faceNo       = numDiv - 1;
	rect.vertexVertex[ 1 ].face.vertexInFace = 1;

	rect.face[ numDiv * numDiv - 1     ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnVertex;
	rect.face[ numDiv * numDiv - 1     ].vertex[ 2 ].vertexNo.no   = 2;
	rect.vertexVertex[ 2 ].face.faceNo       = numDiv * numDiv - 1;
	rect.vertexVertex[ 2 ].face.vertexInFace = 2;

	rect.face[ numDiv * ( numDiv - 1 ) ].vertex[ 3 ].vertexNo.kind = BxCore::kVertexOnVertex;
	rect.face[ numDiv * ( numDiv - 1 ) ].vertex[ 3 ].vertexNo.no   = 3;
	rect.vertexVertex[ 3 ].face.faceNo       = numDiv * ( numDiv - 1 );
	rect.vertexVertex[ 3 ].face.vertexInFace = 3;
}

void  BxRect::
InitDivFaceIn( const int numDiv, BxCore::InfoDivRect& rect )
{
	for( int j=0; j<( numDiv-1 ); j++ ) {
		for( int i=0; i<( numDiv-1 ); i++ ) {
			rect.face[ j * numDiv + i             ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexInFace;
			rect.face[ j * numDiv + i             ].vertex[ 2 ].vertexNo.no   = j * ( numDiv - 1 ) + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 0 ].faceNo          = j * numDiv + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 0 ].vertexInFace    = 2;

			rect.face[ j * numDiv + i + 1         ].vertex[ 3 ].vertexNo.kind = BxCore::kVertexInFace;
			rect.face[ j * numDiv + i + 1         ].vertex[ 3 ].vertexNo.no   = j * ( numDiv - 1 ) + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 1 ].faceNo          = j * numDiv + i + 1;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 1 ].vertexInFace    = 3;

			rect.face[ ( j + 1 ) * numDiv + i     ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexInFace;
			rect.face[ ( j + 1 ) * numDiv + i     ].vertex[ 1 ].vertexNo.no   = j * ( numDiv - 1 ) + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 2 ].faceNo          = ( j + 1 ) * numDiv + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 2 ].vertexInFace    = 1;

			rect.face[ ( j + 1 ) * numDiv + i + 1 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexInFace;
			rect.face[ ( j + 1 ) * numDiv + i + 1 ].vertex[ 0 ].vertexNo.no   = j * ( numDiv - 1 ) + i;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 3 ].faceNo          = ( j + 1 ) * numDiv + i + 1;
			rect.inVertex[ j * ( numDiv - 1 ) + i ].face[ 3 ].vertexInFace    = 0;
		}
	}
}

BxErr  BxRect::
Divide( const BxCore::RectPatchInfo& src, BxCore::DivRect& dst )
{
	if( dst.rect.boundVertex == 0 || dst.rect.inVertex == 0 || dst.rect.face == 0 )
		return kBxErrFailure;

	DivideBound( src, dst );
	DivideIn( src, dst );

	return kBxErrSuccess;
}

void  BxRect::
DivideBound( const BxCore::RectPatchInfo& src, BxCore::DivRect& dst )
{
	int     i, ofs;
	double  u, v;
	BxCore::DivBoundVertexRect*  bVertex;
	BxCore::DivVertexVertex*     vVertex;

	BxCore::InfoDivRect&  rect = dst.rect;
	int  numDiv = dst.numDiv;

	ofs = 0;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &rect.boundVertex[ i + ofs ];
		u = ( double )( i + 1 ) / numDiv;
		DivideOne( src, u, 0.0, bVertex->pos, 2, bVertex->face, rect.face );
	}

	ofs = numDiv - 1;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &rect.boundVertex[ i + ofs ];
		v = ( double )( i + 1 ) / numDiv;
		DivideOne( src, 1.0, v, bVertex->pos, 2, bVertex->face, rect.face );
	}

	ofs = ( numDiv - 1 ) * 2;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &rect.boundVertex[ i + ofs ];
		u = ( double )( numDiv - 1  - i ) / numDiv;
		DivideOne( src, u, 1.0, bVertex->pos, 2, bVertex->face, rect.face );
	}

	ofs = ( numDiv - 1 ) * 3;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &rect.boundVertex[ i + ofs ];
		v = ( double )( numDiv - 1  - i ) / numDiv;
		DivideOne( src, 0.0, v, bVertex->pos, 2, bVertex->face, rect.face );
	}

	vVertex = &rect.vertexVertex[ 0 ];
	DivideOne( src, 0.0, 0.0, vVertex->pos, 1, &vVertex->face, rect.face );

	vVertex = &rect.vertexVertex[ 1 ];
	DivideOne( src, 1.0, 0.0, vVertex->pos, 1, &vVertex->face, rect.face );

	vVertex = &rect.vertexVertex[ 2 ];
	DivideOne( src, 1.0, 1.0, vVertex->pos, 1, &vVertex->face, rect.face );

	vVertex = &rect.vertexVertex[ 3 ];
	DivideOne( src, 0.0, 1.0, vVertex->pos, 1, &vVertex->face, rect.face );
}

void  BxRect::
DivideIn( const BxCore::RectPatchInfo& src, BxCore::DivRect& dst )
{
	BxCore::DivInVertexRect*  vertex;

	BxCore::InfoDivRect&  rect = dst.rect;
	int  numDiv = dst.numDiv;

	for( int j=0; j<( numDiv-1 ); j++ ) {
		for( int i=0; i<( numDiv-1 ); i++ ) {
			vertex = &rect.inVertex[ j * ( numDiv - 1 ) + i ];
			double  u = ( double )( i + 1 ) / numDiv;
			double  v = ( double )( j + 1 ) / numDiv;

			DivideOne( src, u, v, vertex->pos, 4, vertex->face, rect.face );
		}
	}
}

void  BxRect::
DivideOne( const BxCore::RectPatchInfo& src, const double u, const double v, BxVec3& pos,
	const int numNormal, BxCore::DivVertexFace vertexFace[], BxCore::DivFaceRect face[] )
{
	BxVec3  normal;

	BxVec2  uv( u, v );

	PosSurface( src, uv, pos );
	NormalSurface( src, uv, normal );

	for( int i=0; i<numNormal; i++ ) {
		int  faceNo       = vertexFace[ i ].faceNo;
		int  vertexInFace = vertexFace[ i ].vertexInFace;
		face[ faceNo ].vertex[ vertexInFace ].normal = normal;
	}
}


////////
