
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxSurface.cpp */


#include "bxCoreGblSub.h"
#include "bxSurface.h"
#include "bxVector.h"

using namespace NameBeziex;



BxErr  BxSurface::
MakeBNormal( const int no, const int numVertex, BxCore::SurfaceInfoVertex vertex[] )
{
	BxErr  result;

	result = GBxVector()->Normal1( no, numVertex, vertex, vertex[ no ].edge[ 0 ].bNormal );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Normal2( no, numVertex, vertex, vertex[ no ].edge[ 1 ].bNormal );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxSurface::
IHandlePos( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 pos[ 5 ] )
{
	JointInfo  a;
	BxErr      result;
	BxVec3     base[ 5 ], vec[ 5 ];

	if( vertex[ no ].pairEnable == true ) {
		result = JointVec0Enable( no, numVertex, vertex, a.jointVec0 );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		result = JointVec3Enable( no, numVertex, vertex, a.jointVec3 );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}
	else {
		result = JointVec0Disable( no, numVertex, vertex, a.jointVec0 );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		result = JointVec3Disable( no, numVertex, vertex, a.jointVec3 );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	result = JointRatio0( no, numVertex, vertex, a.jointVec0, a.k0, a.h0 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = JointRatio3( no, numVertex, vertex, a.jointVec3, a.k3, a.h3 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = JointVec1( no, numVertex, vertex, a.jointVec0, a.k0, a.h0, a.jointVec1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = JointVec2( no, numVertex, vertex, a.jointVec3, a.k3, a.h3, a.jointVec2 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = JointRatio1( no, numVertex, vertex, a.jointVec0, a.k1, a.h1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = JointRatio2( no, numVertex, vertex, a.jointVec3, a.k2, a.h2 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = BasePos( no, numVertex, vertex, base );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = HandleVec( no, numVertex, vertex, a, vec );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	VecToPos( base, vec, pos );

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointVec0Enable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& jointVec0 )
{
	BxVec3  tan, handle, antiHandle;
	BxErr   result;

	result = GBxVector()->Tan0( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle0( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->AntiHandle0( no, numVertex, vertex, antiHandle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return JointVecOut( tan, handle, antiHandle, jointVec0 );
}

BxErr  BxSurface::
JointVec3Enable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& jointVec3 )
{
	BxVec3  tan, handle, antiHandle;
	BxErr   result;

	result = GBxVector()->Tan2( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle3( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->AntiHandle3( no, numVertex, vertex, antiHandle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return JointVecOut( -tan, handle, antiHandle, jointVec3 );
}

BxErr  BxSurface::
JointVecOut( const BxVec3& tan, const BxVec3& handle, const BxVec3& antiHandle, BxVec3& jointVec )
{
	BxErr  errResult;

	BxVec3  tanE = tan.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	double  sinDt0 = Cross( tanE, handle ).Length();
	double  sinDt1 = Cross( tanE, antiHandle ).Length();
	if( sinDt0 < GBxNearZero() || sinDt1 < GBxNearZero() )
		return kBxErrFailure;

	BxVec3  tanNorE = Cross( tanE, Cross( tanE, handle ) ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	double  cosDt1 = Dot( tanE, antiHandle );

	BxVec3  handle0     = handle / sinDt0;
	BxVec3  antiHandle0 = ( cosDt1 * tanE + sinDt1 * tanNorE ) / sinDt1;

	double  len = ( ( handle0 + antiHandle0 ) / 2.0 ).Length();
	if( Dot( ( handle0 + antiHandle0 ), tanE ) < 0.0 )
		len = -len;

	jointVec = ( handle0 - len * tanE ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointVec0Disable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& jointVec0 )
{
	BxVec3  handle;
	BxErr   errResult;

	BxErr  result = GBxVector()->Handle0( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	
	jointVec0 = handle.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointVec3Disable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& jointVec3 )
{
	BxVec3  handle;
	BxErr   errResult;

	BxErr  result = GBxVector()->Handle3( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	
	jointVec3 = handle.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointRatio0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec0,
	double& k0, double& h0 )
{
	BxVec3  tan, handle;
	BxErr   result;

	result = GBxVector()->Tan0( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle0( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return JointRatioMain( tan, handle, jointVec0, k0, h0 );
}

BxErr  BxSurface::
JointRatio3( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec3,
	double& k3, double& h3 )
{
	BxVec3  tan, handle;
	BxErr   result;

	result = GBxVector()->Tan2( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle3( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	BxErr  err = JointRatioMain( -tan, handle, jointVec3, k3, h3 );
	if( err == kBxErrSuccess )
		h3 = -h3;

	return err;
}

BxErr  BxSurface::
JointRatioMain( const BxVec3& tan, const BxVec3& handle, const BxVec3& jointVec, double& k, double& h )
{
	double  lenTan = tan.Length();
	if( lenTan < GBxNearZero() )
		return kBxErrFailure;
	BxVec3  tanE = tan / lenTan;

	double  sinHandle = Cross( tanE, handle ).Length();
	double  sinJoint  = Cross( tanE, jointVec ).Length();
	if( sinJoint < GBxNearZero() )
		return kBxErrFailure;

	k = sinHandle / sinJoint;

	BxVec3  hVec    = handle - ( k * jointVec );
	double  lenHVec = hVec.Length();

	h = lenHVec / ( lenTan * 3.0 / 4.0 );
	if( Dot( hVec, tan ) < 0 )
		h = -h;

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointVec1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec0,
	const double k0, const double h0, BxVec3& jointVec1 )
{
	BxVec3  pos0, pos1, pos2;
	BxErr   result;

	GBxVector()->Pos0( no, numVertex, vertex, pos0 );
	result = GBxVector()->Pos1( no, numVertex, vertex, pos1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Pos2( no, numVertex, vertex, pos2 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	BxVec3  pos01 = ( pos0 + 3.0 * pos1 ) / 4.0;
	BxVec3  pos12 = ( pos1 + pos2 ) / 2.0;

	BxVec3  tan0 = pos01 - pos0;
	BxVec3  tan1 = pos12 - pos01;

	BxVec3  vNormal = vertex[ no ].vertexNormal;
	BxVec3  bNormal = vertex[ no ].edge[ 0 ].bNormal;

	return JointVecIn( tan0, tan1, jointVec0, k0, h0, vNormal, bNormal, jointVec1 );
}

BxErr  BxSurface::
JointVec2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec3,
	const double k3, const double h3, BxVec3& jointVec2 )
{
	BxVec3  pos1, pos2, pos3;
	BxErr   result;

	result = GBxVector()->Pos1( no, numVertex, vertex, pos1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Pos2( no, numVertex, vertex, pos2 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	GBxVector()->Pos3( no, numVertex, vertex, pos3 );

	BxVec3  pos12 = ( pos1 + pos2 ) / 2.0;
	BxVec3  pos23 = ( 3.0 * pos2 + pos3 ) / 4.0;

	BxVec3  tan0 = pos23 - pos3;
	BxVec3  tan1 = pos12 - pos23;

	int  p = ( no + 1 ) % numVertex;
	BxVec3  vNormal = vertex[ p  ].vertexNormal;
	BxVec3  bNormal = vertex[ no ].edge[ 1 ].bNormal;

	return JointVecIn( tan0, tan1, jointVec3, k3, -h3, vNormal, bNormal, jointVec2 );
}

BxErr  BxSurface::
JointVecIn( const BxVec3& tan0, const BxVec3& tan1, const BxVec3& jointVecOut, const double k, const double h,
	const BxVec3& vNormal, const BxVec3& bNormal, BxVec3& jointVecIn )
{
	BxErr   errResult;
	BxVec3  jointVecInTmp;

	if( k < GBxNearZero() )
		return kBxErrFailure;

	jointVecInTmp = jointVecOut - ( h / k ) * ( tan1 - tan0 );
	double  lenJointVecInTmp = jointVecInTmp.Length();
	if( lenJointVecInTmp < GBxNearZero() )
		return kBxErrFailure;

	jointVecInTmp = Cross( vNormal, bNormal );
	if( jointVecInTmp.Length() < GBxNearZero() )
		jointVecInTmp = jointVecOut;
	else if( Dot( jointVecOut, jointVecInTmp ) < 0 )
		jointVecInTmp = -jointVecInTmp;

	jointVecIn = jointVecInTmp.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	jointVecIn *= lenJointVecInTmp;

	return kBxErrSuccess;
}

BxErr  BxSurface::
JointRatio1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec0,
	double& k1, double& h1 )
{
	BxVec3  tan, handle;
	BxErr   result;

	result = GBxVector()->Tan0( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle1( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return JointRatioMain( tan, handle, jointVec0, k1, h1 );
}

BxErr  BxSurface::
JointRatio2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec3,
	double& k2, double& h2 )
{
	BxVec3  tan, handle;
	BxErr   result;

	result = GBxVector()->Tan2( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = GBxVector()->Handle2( no, numVertex, vertex, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	BxErr  err = JointRatioMain( -tan, handle, jointVec3, k2, h2 );
	if( err == kBxErrSuccess )
		h2 = -h2;

	return err;
}

BxErr  BxSurface::
BasePos( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 base[ 5 ] )
{
	BxVec3  p3[ 4 ], p4[ 5 ], p5[ 6 ];

	BxErr  result = GBxVector()->Pos( no, numVertex, vertex, p3 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	p4[ 0 ] = p3[ 0 ];
	p4[ 1 ] = (       p3[ 0 ] + 3.0 * p3[ 1 ] ) / 4.0;
	p4[ 2 ] = (       p3[ 1 ] +       p3[ 2 ] ) / 2.0;
	p4[ 3 ] = ( 3.0 * p3[ 2 ] +       p3[ 3 ] ) / 4.0;
	p4[ 4 ] = p3[ 3 ];

	p5[ 0 ] = p4[ 0 ];
	p5[ 1 ] = (       p4[ 0 ] + 4.0 * p4[ 1 ] ) / 5.0;
	p5[ 2 ] = ( 2.0 * p4[ 1 ] + 3.0 * p4[ 2 ] ) / 5.0;
	p5[ 3 ] = ( 3.0 * p4[ 2 ] + 2.0 * p4[ 3 ] ) / 5.0;
	p5[ 4 ] = ( 4.0 * p4[ 3 ] +       p4[ 4 ] ) / 5.0;
	p5[ 5 ] = p4[ 4 ];

	base[ 0 ] = (       p5[ 0 ] + 5.0 * p5[ 1 ] ) / 6.0;
	base[ 1 ] = (       p5[ 1 ] + 2.0 * p5[ 2 ] ) / 3.0;
	base[ 2 ] = (       p5[ 2 ] +       p5[ 3 ] ) / 2.0;
	base[ 3 ] = ( 2.0 * p5[ 3 ] +       p5[ 4 ] ) / 3.0;
	base[ 4 ] = ( 5.0 * p5[ 4 ] +       p5[ 5 ] ) / 6.0;

	return kBxErrSuccess;
}

BxErr  BxSurface::
HandleVec( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const JointInfo& a,
	BxVec3 vec[ 5 ] )
{
	BxVec3  p3[ 4 ], p4[ 5 ];

	BxErr  result = GBxVector()->Pos( no, numVertex, vertex, p3 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	p4[ 0 ] = p3[ 0 ];
	p4[ 1 ] = (       p3[ 0 ] + 3.0 * p3[ 1 ] ) / 4.0;
	p4[ 2 ] = (       p3[ 1 ] +       p3[ 2 ] ) / 2.0;
	p4[ 3 ] = ( 3.0 * p3[ 2 ] +       p3[ 3 ] ) / 4.0;
	p4[ 4 ] = p3[ 3 ];

	BxVec3  tan0 = p4[ 1 ] - p4[ 0 ];
	BxVec3  tan1 = p4[ 2 ] - p4[ 1 ];
	BxVec3  tan2 = p4[ 3 ] - p4[ 2 ];
	BxVec3  tan3 = p4[ 4 ] - p4[ 3 ];

	vec[ 0 ] = ( a.k0 * a.jointVec1
		   + a.k1 * a.jointVec0
		   + a.h0 * tan1
		   + a.h1 * tan0        ) / 2.0;

	vec[ 1 ] = (       a.k0 * a.jointVec2 
		   + 3.0 * a.k1 * a.jointVec1
		   +       a.k2 * a.jointVec0
		   +       a.h0 * tan2
		   + 3.0 * a.h1 * tan1
		   +       a.h2 * tan0        ) / 5.0;

	vec[ 2 ] = (       a.k0 * a.jointVec3 
		   + 9.0 * a.k1 * a.jointVec2
		   + 9.0 * a.k2 * a.jointVec1
		   +       a.k3 * a.jointVec0
		   +       a.h0 * tan3
		   + 9.0 * a.h1 * tan2
		   + 9.0 * a.h2 * tan1
		   +       a.h3 * tan0        ) / 20.0;
	
	vec[ 3 ] = (       a.k1 * a.jointVec3 
		   + 3.0 * a.k2 * a.jointVec2
		   +       a.k3 * a.jointVec1
		   +       a.h1 * tan3
		   + 3.0 * a.h2 * tan2
		   +       a.h3 * tan1        ) / 5.0;

	vec[ 4 ] = ( a.k2 * a.jointVec3
		   + a.k3 * a.jointVec2
		   + a.h2 * tan3
		   + a.h3 * tan2        ) / 2.0;

	return kBxErrSuccess;
}

void  BxSurface::
VecToPos( const BxVec3 base[ 5 ], const BxVec3 vec[ 5 ], BxVec3 pos[ 5 ] )
{
	for( int i=0; i<5; i++ )
		pos[ i ] = base[ i ] + vec[ i ];
}

BxErr  BxSurface::
TriVHandlePos( const int no, const BxCore::SurfaceInfoVertex vertex[], BxVec3 pos[ 2 ] )
{
	BxVec3  vec[ 2 ];

	BxErr  result = GBxVector()->TriVHandle( no, vertex, vec );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	int  m = ( no + 3 - 1 ) % 3;

	for( int i=0; i<2; i++ )
		pos[ i ] = vertex[ m ].vertexPos + vec[ i ];

	return kBxErrSuccess;
}

void  BxSurface::
PosSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& pos )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier6Line  v2( patch.v2 );
	Bezier3Line  v3( patch.v3 );
	BxVec3       p0[ 4 ];

	p0[ 0 ] = Pos( v0, uv.U() );
	p0[ 1 ] = Pos( v1, uv.U() );
	p0[ 2 ] = Pos( v2, uv.U() );
	p0[ 3 ] = Pos( v3, uv.U() );

	Bezier3Line  p( p0 );
	pos = Pos( p, uv.V() );
}

void  BxSurface::
DuSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& du )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier6Line  v2( patch.v2 );
	Bezier3Line  v3( patch.v3 );
	BxVec3       d0[ 4 ];

	d0[ 0 ] = Diff( v0, uv.U() );
	d0[ 1 ] = Diff( v1, uv.U() );
	d0[ 2 ] = Diff( v2, uv.U() );
	d0[ 3 ] = Diff( v3, uv.U() );

	Bezier3Line  d( d0 );
	du = Pos( d, uv.V() );
}

void  BxSurface::
DvSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& dv )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier6Line  v2( patch.v2 );
	Bezier3Line  v3( patch.v3 );
	BxVec3       p0[ 4 ];

	p0[ 0 ] = Pos( v0, uv.U() );
	p0[ 1 ] = Pos( v1, uv.U() );
	p0[ 2 ] = Pos( v2, uv.U() );
	p0[ 3 ] = Pos( v3, uv.U() );

	Bezier3Line  p( p0 );
	dv = Diff( p, uv.V() );
}

void  BxSurface::
PosSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& pos )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ];

	p0[ 0 ] = Pos( v0, uv.U() );
	p0[ 1 ] = Pos( v1, uv.U() );
	p0[ 2 ] = Pos( v2, uv.U() );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	pos = Pos( p, uv.V() );
}

void  BxSurface::
DuSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& du )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       d0[ 4 ];

	d0[ 0 ] = Diff( v0, uv.U() );
	d0[ 1 ] = Diff( v1, uv.U() );
	d0[ 2 ] = Diff( v2, uv.U() );
	d0[ 3 ] = GBxZero3();

	Bezier3Line  d( d0 );
	du = Pos( d, uv.V() );
}

void  BxSurface::
DvSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& dv )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ];

	p0[ 0 ] = Pos( v0, uv.U() );
	p0[ 1 ] = Pos( v1, uv.U() );
	p0[ 2 ] = Pos( v2, uv.U() );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	dv = Diff( p, uv.V() );
}

void  BxSurface::
PosSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& pos )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ];

	double  ur = ParamUr( uvT );
	double  vr = ParamVr( uvT );

	p0[ 0 ] = Pos( v0, ur );
	p0[ 1 ] = Pos( v1, ur );
	p0[ 2 ] = Pos( v2, ur );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	pos = Pos( p, vr );
}

void  BxSurface::
DutSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ], d0[ 4 ];

	double  ur  = ParamUr( uvT );
	double  vr  = ParamVr( uvT );
	double  dur = DuParamUr( uvT );
	double  dvr = DuParamVr( uvT );

	p0[ 0 ] = Pos( v0, ur );
	p0[ 1 ] = Pos( v1, ur );
	p0[ 2 ] = Pos( v2, ur );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	BxVec3  du0 = Diff( p, vr ) * dvr;

	d0[ 0 ] = Diff( v0, ur ) * dur;
	d0[ 1 ] = Diff( v1, ur ) * dur;
	d0[ 2 ] = Diff( v2, ur ) * dur;
	d0[ 3 ] = GBxZero3();

	Bezier3Line  d( d0 );
	BxVec3  du1 = Pos( d, vr );

	dut = du0 + du1;
}

void  BxSurface::
DvtSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& dvt )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ], d0[ 4 ];

	double  ur  = ParamUr( uvT );
	double  vr  = ParamVr( uvT );
	double  dur = DvParamUr( uvT );
	double  dvr = DvParamVr( uvT );

	p0[ 0 ] = Pos( v0, ur );
	p0[ 1 ] = Pos( v1, ur );
	p0[ 2 ] = Pos( v2, ur );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	BxVec3  dv0 = Diff( p, vr ) * dvr;

	d0[ 0 ] = Diff( v0, ur ) * dur;
	d0[ 1 ] = Diff( v1, ur ) * dur;
	d0[ 2 ] = Diff( v2, ur ) * dur;
	d0[ 3 ] = GBxZero3();

	Bezier3Line  d( d0 );
	BxVec3  dv1 = Pos( d, vr );

	dvt = dv0 + dv1;
}

void  BxSurface::
PosSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& pos )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ];

	double  us = ParamUs( uvT );
	double  vs = ParamVs( uvT );

	p0[ 0 ] = Pos( v0, us );
	p0[ 1 ] = Pos( v1, us );
	p0[ 2 ] = Pos( v2, us );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	pos = Pos( p, vs );
}

void  BxSurface::
DutSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ], d0[ 4 ];

	double  us  = ParamUs( uvT );
	double  vs  = ParamVs( uvT );
	double  dus = DuParamUs( uvT );
	double  dvs = DuParamVs( uvT );

	p0[ 0 ] = Pos( v0, us );
	p0[ 1 ] = Pos( v1, us );
	p0[ 2 ] = Pos( v2, us );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	BxVec3  du0 = Diff( p, vs ) * dvs;

	d0[ 0 ] = Diff( v0, us ) * dus;
	d0[ 1 ] = Diff( v1, us ) * dus;
	d0[ 2 ] = Diff( v2, us ) * dus;
	d0[ 3 ] = GBxZero3();

	Bezier3Line  d( d0 );
	BxVec3  du1 = Pos( d, vs );

	dut = du0 + du1;
}

void  BxSurface::
DvtSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& dvt )
{
	Bezier3Line  v0( patch.v0 );
	Bezier6Line  v1( patch.v1 );
	Bezier3Line  v2( patch.v2 );
	BxVec3       p0[ 4 ], d0[ 4 ];

	double  us  = ParamUs( uvT );
	double  vs  = ParamVs( uvT );
	double  dus = DvParamUs( uvT );
	double  dvs = DvParamVs( uvT );

	p0[ 0 ] = Pos( v0, us );
	p0[ 1 ] = Pos( v1, us );
	p0[ 2 ] = Pos( v2, us );
	p0[ 3 ] = patch.v3;

	Bezier3Line  p( p0 );
	BxVec3  dv0 = Diff( p, vs ) * dvs;

	d0[ 0 ] = Diff( v0, us ) * dus;
	d0[ 1 ] = Diff( v1, us ) * dus;
	d0[ 2 ] = Diff( v2, us ) * dus;
	d0[ 3 ] = GBxZero3();

	Bezier3Line  d( d0 );
	BxVec3  dv1 = Pos( d, vs );

	dvt = dv0 + dv1;
}

double  BxSurface::
ParamUr( const BxVec2& uvT )
{
	return uvT.V() / ( uvT.U() + uvT.V() - uvT.U() * uvT.V() );
}

double  BxSurface::
ParamVr( const BxVec2& uvT )
{
	return ( 1.0 - uvT.U() ) * ( 1.0 - uvT.V() );
}

double  BxSurface::
ParamUs( const BxVec2& uvT )
{
	double  a = 1.0 - uvT.U() + uvT.U() * uvT.V();

	return ( a - uvT.V() ) / a;
}

double  BxSurface::
ParamVs( const BxVec2& uvT )
{
	return uvT.U() * ( 1.0 - uvT.V() );
}

double  BxSurface::
DuParamUr( const BxVec2& uvT )
{
	double  a = uvT.U() + uvT.V() - uvT.U() * uvT.V();

	return ( uvT.V() * ( uvT.V() - 1.0 ) ) / ( a * a );
}

double  BxSurface::
DuParamVr( const BxVec2& uvT )
{
	return uvT.V() - 1.0;
}

double  BxSurface::
DuParamUs( const BxVec2& uvT )
{
	double  a = 1.0 - uvT.U() + uvT.U() * uvT.V();

	return ( uvT.V() * ( 1.0 - uvT.V() ) ) / ( a * a );
}

double  BxSurface::
DuParamVs( const BxVec2& uvT )
{
	return 1.0 - uvT.V();
}

double  BxSurface::
DvParamUr( const BxVec2& uvT )
{
	double  a = uvT.U() + uvT.V() - uvT.U() * uvT.V();

	return uvT.U() / ( a * a );
}

double  BxSurface::
DvParamVr( const BxVec2& uvT )
{
	return uvT.U() - 1.0;
}

double  BxSurface::
DvParamUs( const BxVec2& uvT )
{
	double  a = 1.0 - uvT.U() + uvT.U() * uvT.V();

	return ( uvT.U() - 1.0 ) / ( a * a );
}

double  BxSurface::
DvParamVs( const BxVec2& uvT )
{
	return -uvT.U();
}


////////
