
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxSurface.h */


#if !defined( BX_SURFACE_H )
#define BX_SURFACE_H

#include "bxCore.h"


class BxSurface {
public:
	struct JointInfo {
		BxVec3  jointVec0, jointVec1, jointVec2, jointVec3;
		double  k0, k1, k2, k3, h0, h1, h2, h3;
	};

	struct Patch4774 {
		BxVecAry4  v0;
		BxVecAry7  v1;
		BxVecAry7  v2;
		BxVecAry4  v3;
	};
	struct Patch4741 {
		BxVecAry4  v0;
		BxVecAry7  v1;
		BxVecAry4  v2;
		BxVec3     v3;
	};

public:
	BxErr
	MakeBNormal( const int no, const int numVertex, BxCore::SurfaceInfoVertex vertex[] );

	BxErr
	IHandlePos( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 pos[ 5 ] );
	BxErr
	JointVec0Enable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		BxVec3& jointVec0 );
	BxErr
	JointVec3Enable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		BxVec3& jointVec3 );
	BxErr
	JointVec0Disable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		BxVec3& jointVec0 );
	BxErr
	JointVec3Disable( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		BxVec3& jointVec3 );
	BxErr
	JointRatio0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		const BxVec3& jointVec0, double& k0, double& h0 );
	BxErr
	JointRatio3( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		const BxVec3& jointVec3, double& k3, double& h3 );
	BxErr
	JointVec1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec0,
		const double k0, const double h0, BxVec3& jointVec1 );
	BxErr
	JointVec2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const BxVec3& jointVec3,
		const double k3, const double h3, BxVec3& jointVec2 );
	BxErr
	JointRatio1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		const BxVec3& jointVec0, double& k1, double& h1 );
	BxErr
	JointRatio2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[],
		const BxVec3& jointVec3, double& k2, double& h2 );

	BxErr
	TriVHandlePos( const int no, const BxCore::SurfaceInfoVertex vertex[], BxVec3 pos[ 2 ] );

	void
	PosSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& pos );
	void
	DuSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& du );
	void
	DvSurface4774( const Patch4774& patch, const BxVec2& uv, BxVec3& dv );

	void
	PosSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& pos );
	void
	DuSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& du );
	void
	DvSurface4741( const Patch4741& patch, const BxVec2& uv, BxVec3& dv );

	void
	PosSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& pos );
	void
	DutSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut );
	void
	DvtSurface4741R( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut );

	void
	PosSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& pos );
	void
	DutSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut );
	void
	DvtSurface4741S( const Patch4741& patch, const BxVec2& uvT, BxVec3& dut );

private:
	BxErr
	JointVecOut( const BxVec3& tan, const BxVec3& handle, const BxVec3& antiHandle, BxVec3& jointVec );
	BxErr
	JointRatioMain( const BxVec3& tan, const BxVec3& handle, const BxVec3& jointVec, double& k, double& h );
	BxErr
	JointVecIn( const BxVec3& tan0, const BxVec3& tan1, const BxVec3& jointVecOut, const double k, const double h,
		const BxVec3& vNormal, const BxVec3& bNormal, BxVec3& jointVecIn );
	BxErr
	BasePos( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 base[ 5 ] );
	BxErr
	HandleVec( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], const JointInfo& a,
		BxVec3 vec[ 5 ] );
	void
	VecToPos( const BxVec3 base[ 5 ], const BxVec3 vec[ 5 ], BxVec3 pos[ 5 ] );

	double
	ParamUr( const BxVec2& uvT );
	double
	ParamVr( const BxVec2& uvT );
	double
	ParamUs( const BxVec2& uvT );
	double
	ParamVs( const BxVec2& uvT );
	double
	DuParamUr( const BxVec2& uvT );
	double
	DuParamVr( const BxVec2& uvT );
	double
	DuParamUs( const BxVec2& uvT );
	double
	DuParamVs( const BxVec2& uvT );
	double
	DvParamUr( const BxVec2& uvT );
	double
	DvParamVr( const BxVec2& uvT );
	double
	DvParamUs( const BxVec2& uvT );
	double
	DvParamVs( const BxVec2& uvT );
};


#endif // #if !defined( BX_SURFACE_H )

////////
