
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxSurfaceErrChk.cpp */


#include <math.h>
#include "bxCoreGblSub.h"
#include "bxSurfaceErrChk.h"
#include "bxOrgTo3D.h"

using namespace NameBeziex;



void  BxSurfaceErrChk::
Conv( const BxCore::SurfaceInfo& src )
{
	BxSurfaceErrChkVertex::InfoVertex*  vertex0;
	BxSurfaceErrChkVertex::InfoWing*    wing[ 2 ];
	int  j;

	numVertex = src.numVertex;

	for( int i=0; i<src.numVertex; i++ ) {
		int  m  = ( i + src.numVertex - 1 ) % src.numVertex;
		int  p  = ( i + 1 ) % src.numVertex;
		int  p2 = ( i + 2 ) % src.numVertex;


		vertex0 = &vertex[ i ].vertex[ 0 ];

		vertex0->vertexPos =  src.vertex[ i ].vertexPos;
		vertex0->boundRot  =  src.vertex[ i ].boundRot;

		
		vertex0 = &vertex[ i ].vertex[ 1 ];

		vertex0->vertexPos =  src.vertex[ p ].vertexPos;
		vertex0->boundRot  = -src.vertex[ i ].boundRot;


		vertex[ i ].pairEnable        = src.vertex[ i ].pairEnable;
		vertex[ i ].pairEnableBHandle = src.vertex[ i ].pairEnableBHandle;

		for( j=0; j<2; j++ )
			wing[ j ] = &vertex[ i ].wing[ j ];


		wing[ 0 ]->edge[ 0 ].wingVec               =  src.vertex[ m ].vertexPos - src.vertex[ i ].vertexPos;
		wing[ 0 ]->edge[ 0 ].wingNormal[ 0 ]       =  src.vertex[ i ].vertexNormal;
		wing[ 0 ]->edge[ 0 ].wingNormal[ 1 ]       =  src.vertex[ m ].vertexNormal;
		wing[ 0 ]->edge[ 0 ].wingBoundRot          = -src.vertex[ m ].boundRot;
		wing[ 0 ]->edge[ 0 ].wingPairEnable        =  src.vertex[ m ].pairEnable;
		wing[ 0 ]->edge[ 0 ].wingPairEnableBHandle =  src.vertex[ m ].pairEnableBHandle;
		wing[ 0 ]->edge[ 0 ].wingPairNormal[ 0 ]   =  src.vertex[ m ].edge[ 1 ].wingNormal[ 0 ];
		wing[ 0 ]->edge[ 0 ].wingPairNormal[ 1 ]   =  src.vertex[ m ].edge[ 0 ].wingNormal[ 0 ];

		wing[ 0 ]->edge[ 1 ].wingVec               =  src.vertex[ p2 ].vertexPos - src.vertex[ p ].vertexPos;
		wing[ 0 ]->edge[ 1 ].wingNormal[ 0 ]       =  src.vertex[ p  ].vertexNormal;
		wing[ 0 ]->edge[ 1 ].wingNormal[ 1 ]       =  src.vertex[ p2 ].vertexNormal;
		wing[ 0 ]->edge[ 1 ].wingBoundRot          =  src.vertex[ p  ].boundRot;
		wing[ 0 ]->edge[ 1 ].wingPairEnable        =  src.vertex[ p  ].pairEnable;
		wing[ 0 ]->edge[ 1 ].wingPairEnableBHandle =  src.vertex[ p  ].pairEnableBHandle;
		wing[ 0 ]->edge[ 1 ].wingPairNormal[ 0 ]   =  src.vertex[ p  ].edge[ 0 ].wingNormal[ 0 ];
		wing[ 0 ]->edge[ 1 ].wingPairNormal[ 1 ]   =  src.vertex[ p  ].edge[ 1 ].wingNormal[ 0 ];

		for( j=0; j<2; j++ ) {
			wing[ 1 ]->edge[ j ].wingVec               =  src.vertex[ i ].edge[ j ].wingVec;
			wing[ 1 ]->edge[ j ].wingNormal[ 0 ]       =  src.vertex[ i ].edge[ j ].wingNormal[ 0 ];
			wing[ 1 ]->edge[ j ].wingNormal[ 1 ]       =  src.vertex[ i ].edge[ j ].wingNormal[ 1 ];
			wing[ 1 ]->edge[ j ].wingBoundRot          =  src.vertex[ i ].edge[ j ].wingBoundRot;
			wing[ 1 ]->edge[ j ].wingPairEnable        =  src.vertex[ i ].edge[ j ].wingPairEnable;
			wing[ 1 ]->edge[ j ].wingPairEnableBHandle =  src.vertex[ i ].edge[ j ].wingPairEnableBHandle;
			wing[ 1 ]->edge[ j ].wingPairNormal[ 0 ]   =  src.vertex[ i ].edge[ j ].wingPairNormal[ 0 ];
			wing[ 1 ]->edge[ j ].wingPairNormal[ 1 ]   =  src.vertex[ i ].edge[ j ].wingPairNormal[ 1 ];
		}
	}
}

BxErr  BxSurfaceErrChk::
IsEnable( bool& isEnable )
{
	for( int i=0; i<numVertex; i++ ) {
		BxErr  result = vertex[ i ].IsEnable( isEnable );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		else if( isEnable == false )
			return kBxErrSuccess;
	}

	return kBxErrSuccess;
}


BxErr  BxSurfaceErrChkVertex::
IsEnable( bool& isEnable )
{
	BxErr  result;

	IsEnableVertexPos( isEnable );
	if( isEnable == false )
		return kBxErrSuccess;

	IsEnablePairNormal( isEnable );
	if( isEnable == false )
		return kBxErrSuccess;

	result = IsEnableEdgeNormal( isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	IsEnableBoundRot( isEnable );
	if( isEnable == false )
		return kBxErrSuccess;

	result = IsEnableRadial( isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	result = IsEnableSub0( isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	return kBxErrSuccess;
}

BxErr  BxSurfaceErrChkVertex::
IsEnableSub0( bool& isEnable )
{
	for( int i=0; i<2; i++ ) {
		BxErr  result = IsEnableSub00( i, isEnable );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		else if( isEnable == false )
			return kBxErrSuccess;
	}

	return kBxErrSuccess;
}

BxErr  BxSurfaceErrChkVertex::
IsEnableSub00( const short beginEndNo, bool& isEnable )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ], handle, wingHandle[ 2 ];
	BxErr   errResult, result;
	double  sinDt, cosDt;

	isEnable = true;

	normal[ 0 ]     = wing[ 0 ].edge[ beginEndNo ].wingNormal[ 0 ];
	pairNormal[ 0 ] = wing[ 1 ].edge[ beginEndNo ].wingNormal[ 0 ];

	normal[ 1 ]     = wing[ 0 ].edge[ 1-beginEndNo ].wingNormal[ 0 ];
	pairNormal[ 1 ] = wing[ 1 ].edge[ 1-beginEndNo ].wingNormal[ 0 ];

	const BxVec3&  radialVec = vertex[ 1-beginEndNo ].vertexPos - vertex[ beginEndNo ].vertexPos;
	const BxDir&   boundRot  = vertex[ beginEndNo ].boundRot;

	result = GBxOrgTo3D()->BHandleVec( normal, radialVec, boundRot, pairEnableBHandle, pairNormal, handle );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	handle = handle.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	for( int i=0; i<2; i++ ) {
		BxVec3*  normal             = wing[ i ].edge[ beginEndNo ].wingNormal;
		BxVec3*  pairNormal         = wing[ i ].edge[ beginEndNo ].wingPairNormal;
		BxVec3&  wRadialVec         = wing[ i ].edge[ beginEndNo ].wingVec;
		BxDir&   wBoundRot          = wing[ i ].edge[ beginEndNo ].wingBoundRot;
		bool     wPairEnableBHandle = wing[ i ].edge[ beginEndNo ].wingPairEnableBHandle;

		result = GBxOrgTo3D()->BHandleVec( normal, wRadialVec, wBoundRot, wPairEnableBHandle, pairNormal,
						   wingHandle[ i ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		wingHandle[ i ] = wingHandle[ i ].Normalize( &errResult );
		if( errResult == kBxErrFailure )
			return kBxErrFailure;

		if( pairEnable == false )
			break;
	}

	sinDt = Cross( wingHandle[ 0 ], handle ).Length();
	cosDt = Dot( wingHandle[ 0 ], handle );
	if( ( cosDt > 0.0 && sinDt < MinAngSinTwoHandle() ) ||
	    ( cosDt < 0.0 && sinDt < MaxAngSinTwoHandle() ) )
	{
		isEnable = false;
		return kBxErrSuccess;
	}

	if( pairEnable == true ) {
		sinDt = Cross( handle, wingHandle[ 1 ] ).Length();
		cosDt = Dot( handle, wingHandle[ 1 ] );

		if( ( cosDt > 0.0 && sinDt < MinAngSinTwoHandle() ) ||
		    ( cosDt < 0.0 && sinDt < MaxAngSinTwoHandle() ) )
		{
			pairEnable = false;
		}
	}

	return kBxErrSuccess;
}

void  BxSurfaceErrChkVertex::
IsEnableVertexPos( bool& isEnable )
{
	int     i;
	double  len;

	isEnable = true;

	len = ( vertex[ 0 ].vertexPos - vertex[ 1 ].vertexPos ).Length();
	if( len < GBxNearZero() ) {
		isEnable = false;
		return;
	}

	for( i=0; i<2; i++ ) {
		len = wing[ 0 ].edge[ i ].wingVec.Length();
		if( len < GBxNearZero() ) {
			isEnable = false;
			return;
		}
	}

	if( pairEnable == true ) {
		for( i=0; i<2; i++ ) {
			len = wing[ 1 ].edge[ i ].wingVec.Length();
			if( len < GBxNearZero() ) {
				pairEnable = false;
				break;
			}
		}
	}
}

void  BxSurfaceErrChkVertex::
IsEnablePairNormal( bool& isEnable )
{
	int     i;
	double  cosDt, sinDt;

	isEnable = true;

	for( i=0; i<2; i++ ) {
		if( wing[ 0 ].edge[ i ].wingPairEnable == false )
			continue;

		for( int j=0; j<2; j++ ) {
			BxVec3&  normal0 = wing[ 0 ].edge[ i ].wingNormal[ j ];
			BxVec3&  normal1 = wing[ 0 ].edge[ i ].wingPairNormal[ j ];

			cosDt = Dot( normal0, normal1 );
			sinDt = Cross( normal0, normal1 ).Length();
			if( cosDt < 0 && sinDt < MaxAngSinPairNormal() ) {
				isEnable = false;
				return;
			}
		}
	}

	if( pairEnable == true ) {
		for( i=0; i<2; i++ ) {
			if( wing[ 1 ].edge[ i ].wingPairEnable == false )
				continue;

			for( int j=0; j<2; j++ ) {
				BxVec3&  normal0 = wing[ 1 ].edge[ i ].wingNormal[ j ];
				BxVec3&  normal1 = wing[ 1 ].edge[ i ].wingPairNormal[ j ];

				cosDt = Dot( normal0, normal1 );
				sinDt = Cross( normal0, normal1 ).Length();
				if( cosDt < 0 && sinDt < MaxAngSinPairNormal() ) {
					pairEnable        = false;
					pairEnableBHandle = false;
					break;
				}
			}
			if( pairEnable == false )
				break;
		}
	}

	if( pairEnable == false )
		return;

	for( i=0; i<2; i++ ) {
		BxVec3&  normal0 = wing[ 0 ].edge[ i ].wingNormal[ 0 ];
		BxVec3&  normal1 = wing[ 1 ].edge[ i ].wingNormal[ 0 ];

		cosDt = Dot( normal0, normal1 );
		sinDt = Cross( normal0, normal1 ).Length();
		if( cosDt < 0 && sinDt < MaxAngSinPairNormal() ) {
			isEnable = false;
			return;
		}
	}
}

BxErr  BxSurfaceErrChkVertex::
IsEnableEdgeNormal( bool& isEnable )
{
	BxVec3  vertexNormal[ 2 ], *normal0, *normal1;
	BxErr   errResult;
	double  cosDt;
	int     i;

	isEnable = true;

	for( i=0; i<2; i++ ) {
		if( wing[ 0 ].edge[ i ].wingPairEnable == false )
			continue;

		for( int j=0; j<2; j++ ) {
			normal0 = &wing[ 0 ].edge[ i ].wingNormal[ j ];
			normal1 = &wing[ 0 ].edge[ i ].wingPairNormal[ j ];

			vertexNormal[ j ] = ( *normal0 + *normal1 ).Normalize( &errResult );
			if( errResult == kBxErrFailure )
				return kBxErrFailure;
		}

		cosDt = Dot( vertexNormal[ 0 ], vertexNormal[ 1 ] );
		if( cosDt < MaxAngCosEdgeNormal() ) {
			isEnable = false;
			return kBxErrSuccess;
		}
	}

	if( pairEnable == true ) {
		for( i=0; i<2; i++ ) {
			if( wing[ 1 ].edge[ i ].wingPairEnable == false )
				continue;

			for( int j=0; j<2; j++ ) {
				normal0 = &wing[ 1 ].edge[ i ].wingNormal[ j ];
				normal1 = &wing[ 1 ].edge[ i ].wingPairNormal[ j ];

				vertexNormal[ j ] = ( *normal0 + *normal1 ).Normalize( &errResult );
				if( errResult == kBxErrFailure )
					return kBxErrFailure;
			}

			cosDt = Dot( vertexNormal[ 0 ], vertexNormal[ 1 ] );
			if( cosDt < MaxAngCosEdgeNormal() ) {
				pairEnable = false;
				break;
			}
		}
	}

	if( pairEnable == false )
		return kBxErrSuccess;

	for( i=0; i<2; i++ ) {
		normal0 = &wing[ 0 ].edge[ i ].wingNormal[ 0 ];
		normal1 = &wing[ 1 ].edge[ i ].wingNormal[ 0 ];

		vertexNormal[ i ] = ( *normal0 + *normal1 ).Normalize( &errResult );
		if( errResult == kBxErrFailure )
			return kBxErrFailure;
	}

	cosDt = Dot( vertexNormal[ 0 ], vertexNormal[ 1 ] );
	if( cosDt < MaxAngCosEdgeNormal() )
		isEnable = false;

	return kBxErrSuccess;
}

BxErr  BxSurfaceErrChkVertex::
IsEnableBoundRot( bool& isEnable )
{
	int    i;
	BxDir  boundRot;
	BxErr  errResult0, errResult1;

	isEnable = true;
	
	for( i=0; i<2; i++ ) {
		boundRot = vertex[ i ].boundRot;
		bool  chk0 = boundRot.GreaterSignedAngle( -MaxAngBoundRot(), &errResult0 );
		bool  chk1 = boundRot.LessSignedAngle(     MaxAngBoundRot(), &errResult1 );
		if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure )
			return kBxErrFailure;
		if( chk0 == false || chk1 == false ) {
			isEnable = false;
			return kBxErrSuccess;
		}
	}

	for( i=0; i<2; i++ ) {
		boundRot = wing[ 0 ].edge[ i ].wingBoundRot;
		bool  chk0 = boundRot.GreaterSignedAngle( -MaxAngBoundRot(), &errResult0 );
		bool  chk1 = boundRot.LessSignedAngle(     MaxAngBoundRot(), &errResult1 );
		if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure )
			return kBxErrFailure;
		if( chk0 == false || chk1 == false ) {
			isEnable = false;
			return kBxErrSuccess;
		}
	}

	if( pairEnable == true ) {
		for( i=0; i<2; i++ ) {
			boundRot = wing[ 0 ].edge[ i ].wingBoundRot;
			bool  chk0 = boundRot.GreaterSignedAngle( -MaxAngBoundRot(), &errResult0 );
			bool  chk1 = boundRot.LessSignedAngle(     MaxAngBoundRot(), &errResult1 );
			if( errResult0 == kBxErrFailure || errResult1 == kBxErrFailure )
				return kBxErrFailure;
			if( chk0 == false || chk1 == false ) {
				pairEnable = false;
				break;
			}
		}
	}

	return kBxErrSuccess;
}

BxErr  BxSurfaceErrChkVertex::
IsEnableRadial( bool& isEnable )
{
	BxErr  result;
	int    i;

	isEnable = true;
	
	for( i=0; i<2; i++ ) {
		BxVec3*  normal         = wing[ 0 ].edge[ i ].wingNormal;
		BxVec3*  pairNormal     = wing[ 0 ].edge[ i ].wingPairNormal;
		BxVec3&  wRadialVec     = wing[ 0 ].edge[ i ].wingVec;
		BxDir&   wBoundRot      = wing[ 0 ].edge[ i ].wingBoundRot;
		bool     wPairEnableOrg = wing[ 0 ].edge[ i ].wingPairEnable;

		result = IsEnableRadialSub( normal, wRadialVec, wBoundRot, wPairEnableOrg, pairNormal, isEnable );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		else if( isEnable == false )
			return kBxErrSuccess;
	}

	if( pairEnable == true ) {
		for( i=0; i<2; i++ ) {
			BxVec3*  normal         = wing[ 1 ].edge[ i ].wingNormal;
			BxVec3*  pairNormal     = wing[ 1 ].edge[ i ].wingPairNormal;
			BxVec3&  wRadialVec     = wing[ 1 ].edge[ i ].wingVec;
			BxDir&   wBoundRot      = wing[ 1 ].edge[ i ].wingBoundRot;
			bool     wPairEnableOrg = wing[ 1 ].edge[ i ].wingPairEnable;

			result = IsEnableRadialSub( normal, wRadialVec, wBoundRot, wPairEnableOrg, pairNormal,
						    pairEnable );
			if( result == kBxErrFailure )
				return kBxErrFailure;
			else if( pairEnable == false )
				break;
		}
	}
			
	for( i=0; i<2; i++ ) {
		BxVec3  normal[ 2 ], pairNormal[ 2 ];

		normal[ 0 ]     = wing[ 0 ].edge[ i ].wingNormal[ 0 ];
		pairNormal[ 0 ] = wing[ 1 ].edge[ i ].wingNormal[ 0 ];

		normal[ 1 ]     = wing[ 0 ].edge[ 1-i ].wingNormal[ 0 ];
		pairNormal[ 1 ] = wing[ 1 ].edge[ 1-i ].wingNormal[ 0 ];

		const BxVec3&  radialVec = vertex[ 1-i ].vertexPos - vertex[ i ].vertexPos;
		const BxDir&   boundRot  = vertex[ i ].boundRot;

		result = IsEnableRadialSub( normal, radialVec, boundRot, pairEnableBHandle, pairNormal, isEnable );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		else if( isEnable == false )
			return kBxErrSuccess;
	}

	return kBxErrSuccess;
}

BxErr  BxSurfaceErrChkVertex::
IsEnableRadialSub( const BxVec3 normal[ 2 ], const BxVec3& radialVec0, const BxDir& boundRot,
	const bool pairEnableBHandle, const BxVec3 pairNormal[ 2 ], bool& isEnable )
{
	BxVec3  handleVec, radialVec;
	BxErr   errResult;

	BxErr  result = GBxOrgTo3D()->BHandleVec( normal, radialVec0, boundRot, pairEnableBHandle, pairNormal,
						  handleVec );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	radialVec = radialVec0.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	handleVec = handleVec.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	double  cosDt = Dot( radialVec, handleVec );
	if( cosDt < MaxAngCosRadialToHandle() )
		isEnable = false;

	return kBxErrSuccess;
}

void  BxSurfaceErrChk::
SetPairEnable( BxCore::SurfaceInfo& dst )
{
	for( int i=0; i<dst.numVertex; i++ )
		dst.vertex[ i ].pairEnable = vertex[ i ].pairEnable;
}


const double  BxSurfaceErrChkVertex::
MinAngSinTwoHandle()
{
	static const double  minDeg = 5.0;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = sin( minDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}

const double  BxSurfaceErrChkVertex::
MaxAngSinTwoHandle()
{
	static const double  maxDeg = 180.0 - 5.0;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = sin( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}

const double  BxSurfaceErrChkVertex::
MaxAngCosRadialToHandle()
{
	static const double  maxDeg = 90.0 - 5.0;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = cos( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}

const double  BxSurfaceErrChkVertex::
MaxAngSinPairNormal()
{
	static const double  maxDeg = 180.0 - 5.0;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = sin( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}

const double  BxSurfaceErrChkVertex::
MaxAngCosEdgeNormal()
{
	static const double  maxDeg = 90.0 + 5.0;

	static bool    enable = false;
	static double  val;

	if( enable == false ) {
		enable = true;
		val = cos( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}

const BxDir&  BxSurfaceErrChkVertex::
MaxAngBoundRot()
{
	static const double  maxDeg = 90.0 - 5.0;

	static bool   enable = false;
	static BxDir  val;

	if( enable == false ) {
		enable = true;
		val.X() = cos( maxDeg * ( GBxPi() / 180.0 ) );
		val.Y() = sin( maxDeg * ( GBxPi() / 180.0 ) );
	}

	return val;
}


////////
