
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxTri.cpp */


#include "bxCoreGbl.h"
#include "bxCoreGblSub.h"
#include "bxTri.h"
#include "bxSurface.h"
#include "bxVector.h"

using namespace NameBeziex;



BxErr  BxTri::
ConvSurface( const BxCore::SurfaceInfoVertex vertex[ 3 ], BxCore::TriPatchInfo& tri )
{
	int    i;
	BxErr  result;
	
	for( i=0; i<3; i++ ) {
		int  m = ( i + 3 - 1 ) % 3;

		result = GBxVector()->Pos2( m, 3, vertex, tri.bound[ i ].p[ 0 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		GBxVector()->Pos0( i, 4, vertex, tri.bound[ i ].p[ 1 ] );
		result = GBxVector()->Pos1( i, 3, vertex, tri.bound[ i ].p[ 2 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	for( i=0; i<3; i++ ) {
		result = GBxSurface()->IHandlePos( i, 3, vertex, tri.in[ i ].dt5.p );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		result = GBxSurface()->TriVHandlePos( i, vertex, tri.in[ i ].dt2.p );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

void  BxTri::
TriToPatch4741T( const BxCore::TriPatchInfo& tri, BxSurface::Patch4741& patch )
{
	patch.v0.p[ 0 ] = tri.bound[ 0 ].p[ 1 ];
	patch.v0.p[ 1 ] = tri.bound[ 0 ].p[ 2 ];
	patch.v0.p[ 2 ] = tri.bound[ 1 ].p[ 0 ];
	patch.v0.p[ 3 ] = tri.bound[ 1 ].p[ 1 ];

	patch.v1.p[ 0 ] = tri.bound[ 0 ].p[ 0 ];
	patch.v1.p[ 1 ] = tri.in[ 0 ].dt5.p[ 0 ];
	patch.v1.p[ 2 ] = tri.in[ 0 ].dt5.p[ 1 ];
	patch.v1.p[ 3 ] = tri.in[ 0 ].dt5.p[ 2 ];
	patch.v1.p[ 4 ] = tri.in[ 0 ].dt5.p[ 3 ];
	patch.v1.p[ 5 ] = tri.in[ 0 ].dt5.p[ 4 ];
	patch.v1.p[ 6 ] = tri.bound[ 1 ].p[ 2 ];

	patch.v2.p[ 0 ] = tri.bound[ 2 ].p[ 2 ];
	patch.v2.p[ 1 ] = tri.in[ 0 ].dt2.p[ 0 ];
	patch.v2.p[ 2 ] = tri.in[ 0 ].dt2.p[ 1 ];
	patch.v2.p[ 3 ] = tri.bound[ 2 ].p[ 0 ];

	patch.v3        = tri.bound[ 2 ].p[ 1 ];
}

void  BxTri::
TriToPatch4741R( const BxCore::TriPatchInfo& tri, BxSurface::Patch4741& patch )
{
	patch.v0.p[ 0 ] = tri.bound[ 1 ].p[ 1 ];
	patch.v0.p[ 1 ] = tri.bound[ 1 ].p[ 2 ];
	patch.v0.p[ 2 ] = tri.bound[ 2 ].p[ 0 ];
	patch.v0.p[ 3 ] = tri.bound[ 2 ].p[ 1 ];

	patch.v1.p[ 0 ] = tri.bound[ 1 ].p[ 0 ];
	patch.v1.p[ 1 ] = tri.in[ 1 ].dt5.p[ 0 ];
	patch.v1.p[ 2 ] = tri.in[ 1 ].dt5.p[ 1 ];
	patch.v1.p[ 3 ] = tri.in[ 1 ].dt5.p[ 2 ];
	patch.v1.p[ 4 ] = tri.in[ 1 ].dt5.p[ 3 ];
	patch.v1.p[ 5 ] = tri.in[ 1 ].dt5.p[ 4 ];
	patch.v1.p[ 6 ] = tri.bound[ 2 ].p[ 2 ];

	patch.v2.p[ 0 ] = tri.bound[ 0 ].p[ 2 ];
	patch.v2.p[ 1 ] = tri.in[ 1 ].dt2.p[ 0 ];
	patch.v2.p[ 2 ] = tri.in[ 1 ].dt2.p[ 1 ];
	patch.v2.p[ 3 ] = tri.bound[ 0 ].p[ 0 ];

	patch.v3        = tri.bound[ 0 ].p[ 1 ];
}

void  BxTri::
TriToPatch4741S( const BxCore::TriPatchInfo& tri, BxSurface::Patch4741& patch )
{
	patch.v0.p[ 0 ] = tri.bound[ 2 ].p[ 1 ];
	patch.v0.p[ 1 ] = tri.bound[ 2 ].p[ 2 ];
	patch.v0.p[ 2 ] = tri.bound[ 0 ].p[ 0 ];
	patch.v0.p[ 3 ] = tri.bound[ 0 ].p[ 1 ];

	patch.v1.p[ 0 ] = tri.bound[ 2 ].p[ 0 ];
	patch.v1.p[ 1 ] = tri.in[ 2 ].dt5.p[ 0 ];
	patch.v1.p[ 2 ] = tri.in[ 2 ].dt5.p[ 1 ];
	patch.v1.p[ 3 ] = tri.in[ 2 ].dt5.p[ 2 ];
	patch.v1.p[ 4 ] = tri.in[ 2 ].dt5.p[ 3 ];
	patch.v1.p[ 5 ] = tri.in[ 2 ].dt5.p[ 4 ];
	patch.v1.p[ 6 ] = tri.bound[ 0 ].p[ 2 ];

	patch.v2.p[ 0 ] = tri.bound[ 1 ].p[ 2 ];
	patch.v2.p[ 1 ] = tri.in[ 2 ].dt2.p[ 0 ];
	patch.v2.p[ 2 ] = tri.in[ 2 ].dt2.p[ 1 ];
	patch.v2.p[ 3 ] = tri.bound[ 1 ].p[ 0 ];

	patch.v3        = tri.bound[ 1 ].p[ 1 ];
}

int  BxTri::
GetTriType( const BxVec2& uv )
{
	double  r, s, t;

	r = uv.U() * ( 1.0 - uv.V() );
	s = uv.V();
	t = ( 1.0 - uv.U() ) * ( 1.0 - uv.V() );

	if( r >= s ) {
		if( r >= t )
			return( kTriTypeR );
		else
			return( kTriTypeT );
	}
	else {
		if( s >= t )
			return( kTriTypeS );
		else
			return( kTriTypeT );
	}
}

void  BxTri::
PosSurface( const BxCore::TriPatchInfo& tri, const BxVec2& uv, BxVec3& pos )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			pos = tri.bound[ 0 ].p[ 1 ];
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			pos = tri.bound[ 1 ].p[ 1 ];
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		pos = tri.bound[ 2 ].p[ 1 ];
		return;
	}

	PosSurfaceMain( tri, uv, pos );
}

void  BxTri::
PosSurfaceMain( const BxCore::TriPatchInfo& tri, const BxVec2& uv0, BxVec3& pos )
{
	BxSurface::Patch4741  patchT, patchR, patchS;
	BxVec2                uv;
	BxVec3                posT, posR, posS;

	ConvInputInfo( tri, uv0, patchT, patchR, patchS, uv );

	GBxSurface()->PosSurface4741( patchT, uv, posT );
	GBxSurface()->PosSurface4741R( patchR, uv, posR );
	GBxSurface()->PosSurface4741S( patchS, uv, posS );

	double  rParam = ( 1.0 - uv.U() ) * uv.V();
	double  sParam = uv.U() * ( 1.0 - uv.U() ) * ( 1.0 - uv.V() );
	double  tParam = uv.U() * uv.V();

	BxVec3&  pr = posS;
	BxVec3&  ps = posT;
	BxVec3&  pt = posR;

	pos = ( rParam * pr + sParam * ps + tParam * pt ) / ( rParam + sParam + tParam );
}

void  BxTri::
ConvInputInfo( const BxCore::TriPatchInfo& tri, const BxVec2& srcUV, BxSurface::Patch4741& patchT,
	BxSurface::Patch4741& patchR, BxSurface::Patch4741& patchS, BxVec2& dstUV )
{
	double  a;

	int  type = GetTriType( srcUV );

	switch( type ) {
	case kTriTypeT:
		TriToPatch4741T( tri, patchT );
		TriToPatch4741R( tri, patchR );
		TriToPatch4741S( tri, patchS );

		dstUV = srcUV;
		break;

	case kTriTypeR:
		TriToPatch4741R( tri, patchT );
		TriToPatch4741S( tri, patchR );
		TriToPatch4741T( tri, patchS );

		dstUV.U() = srcUV.V() / ( srcUV.U() + srcUV.V() - srcUV.U() * srcUV.V() );
		dstUV.V() = ( 1.0 - srcUV.U() ) * ( 1.0 - srcUV.V() );
		break;

	case kTriTypeS:
		TriToPatch4741S( tri, patchT );
		TriToPatch4741T( tri, patchR );
		TriToPatch4741R( tri, patchS );

		a = 1.0 - srcUV.U() + srcUV.U() * srcUV.V();
		dstUV.U() = ( a - srcUV.V() ) / a;
		dstUV.V() = srcUV.U() * ( 1.0 - srcUV.V() );
		break;
	}
}

void  BxTri::
NormalSurface( const BxCore::TriPatchInfo& tri, const BxVec2& uv, BxVec3& normalVec )
{
	BxVec3  du, dv;

	DuSurface( tri, uv, du );
	DvSurface( tri, uv, dv );
	normalVec = Cross( du, dv ).Normalize( GBxZero3() );
}

void  BxTri::
DuSurface( const BxCore::TriPatchInfo& tri, const BxVec2& uv, BxVec3& du )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			du = ( tri.bound[ 0 ].p[ 2 ] - tri.bound[ 0 ].p[ 1 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			du = ( tri.bound[ 1 ].p[ 1 ] - tri.bound[ 1 ].p[ 0 ] ) * 3.0;
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		du = ( tri.bound[ 2 ].p[ 1 ] - tri.bound[ 2 ].p[ 2 ] ) * 3.0;
		return;
	}

	DuSurfaceMain( tri, uv, du );
}

void  BxTri::
DuSurfaceMain( const BxCore::TriPatchInfo& tri, const BxVec2& uv0, BxVec3& du )
{
	BxSurface::Patch4741  patchT, patchR, patchS;
	BxVec2                uv;
	BxVec3                posT, posR, posS, vecT, vecR, vecS;

	ConvInputInfo( tri, uv0, patchT, patchR, patchS, uv );

	GBxSurface()->PosSurface4741( patchT, uv, posT );
	GBxSurface()->PosSurface4741R( patchR, uv, posR );
	GBxSurface()->PosSurface4741S( patchS, uv, posS );

	GBxSurface()->DuSurface4741( patchT, uv, vecT );
	GBxSurface()->DutSurface4741R( patchR, uv, vecR );
	GBxSurface()->DutSurface4741S( patchS, uv, vecS );

	double  rParam0   = ( 1.0 - uv.U() ) * uv.V();
	double  sParam0   = uv.U() * ( 1.0 - uv.U() ) * ( 1.0 - uv.V() );
	double  tParam0   = uv.U() * uv.V();
	double  rstParam0 = rParam0 + sParam0 + tParam0;

	double  rParam1 = -uv.V();
	double  sParam1 =  ( 1.0 - 2.0 * uv.U() ) * ( 1.0 - uv.V() );
	double  tParam1 =  uv.V();

	BxVec3&  pr = posS;
	BxVec3&  ps = posT;
	BxVec3&  pt = posR;

	BxVec3&  vr = vecS;
	BxVec3&  vs = vecT;
	BxVec3&  vt = vecR;

	BxVec3  tmp      = ( rParam0 * pr + sParam0 * ps + tParam0 * pt ) / rstParam0;
	BxVec3  element0 = ( 2.0 * uv.U() - 1.0 ) * ( 1.0 - uv.V() ) * tmp;
	BxVec3  element1 = rParam1 * pr + sParam1 * ps + tParam1 * pt;
	BxVec3  element2 = rParam0 * vr + sParam0 * vs + tParam0 * vt;

	du = ( element0 + element1 + element2 ) / rstParam0;
}

void  BxTri::
DvSurface( const BxCore::TriPatchInfo& tri, const BxVec2& uv, BxVec3& dv )
{
	if( uv.V() < GBxNearZero() ) {
		if( uv.U() < GBxNearZero() ) {
			dv = ( tri.bound[ 0 ].p[ 0 ] - tri.bound[ 0 ].p[ 1 ] ) * 3.0;
			return;
		}
		else if( uv.U() > ( 1.0 - GBxNearZero() ) ) {
			dv = ( tri.bound[ 1 ].p[ 2 ] - tri.bound[ 1 ].p[ 1 ] ) * 3.0;
			return;
		}
	}
	else if( uv.V() > ( 1.0 - GBxNearZero() ) ) {
		dv = ( tri.bound[ 2 ].p[ 1 ] - tri.bound[ 2 ].p[ 0 ] ) * 3.0;
		return;
	}

	DvSurfaceMain( tri, uv, dv );
}

void  BxTri::
DvSurfaceMain( const BxCore::TriPatchInfo& tri, const BxVec2& uv0, BxVec3& dv )
{
	BxSurface::Patch4741  patchT, patchR, patchS;
	BxVec2                uv;
	BxVec3                posT, posR, posS, vecT, vecR, vecS;

	ConvInputInfo( tri, uv0, patchT, patchR, patchS, uv );

	GBxSurface()->PosSurface4741( patchT, uv, posT );
	GBxSurface()->PosSurface4741R( patchR, uv, posR );
	GBxSurface()->PosSurface4741S( patchS, uv, posS );

	GBxSurface()->DvSurface4741( patchT, uv, vecT );
	GBxSurface()->DvtSurface4741R( patchR, uv, vecR );
	GBxSurface()->DvtSurface4741S( patchS, uv, vecS );

	double  rParam0   = ( 1.0 - uv.U() ) * uv.V();
	double  sParam0   = uv.U() * ( 1.0 - uv.U() ) * ( 1.0 - uv.V() );
	double  tParam0   = uv.U() * uv.V();
	double  rstParam0 = rParam0 + sParam0 + tParam0;

	double  rParam1 = 1.0 - uv.U();
	double  sParam1 = uv.U() * ( uv.U() - 1.0 );
	double  tParam1 = uv.U();

	BxVec3&  pr = posS;
	BxVec3&  ps = posT;
	BxVec3&  pt = posR;

	BxVec3&  vr = vecS;
	BxVec3&  vs = vecT;
	BxVec3&  vt = vecR;

	BxVec3  tmp      = ( rParam0 * pr + sParam0 * ps + tParam0 * pt ) / rstParam0;
	BxVec3  element0 = -( uv.U() * uv.U() - uv.U() + 1.0 ) * tmp;
	BxVec3  element1 = rParam1 * pr + sParam1 * ps + tParam1 * pt;
	BxVec3  element2 = rParam0 * vr + sParam0 * vs + tParam0 * vt;

	dv = ( element0 + element1 + element2 ) / rstParam0;
}

void  BxTri::
InitDiv( const int numDiv, BxCore::DivTri& dst )
{
	dst.numVertex = 4;
	dst.numDiv    = numDiv;

	if( dst.tri.boundVertex != 0 )
		delete[] dst.tri.boundVertex;
	if( dst.tri.inVertex != 0 )
		delete[] dst.tri.inVertex;
	if( dst.tri.face != 0 )
		delete[] dst.tri.face;

	dst.tri.boundVertex = new BxCore::DivBoundVertexTri[ GBxCore()->SizeDivBoundVertex( numDiv ) * 3 ];

	int  sizeDivInVertex = GBxCore()->SizeDivInVertex( 3, numDiv );
	if( sizeDivInVertex > 0 )
		dst.tri.inVertex = new BxCore::DivInVertexTri[ sizeDivInVertex ];

	dst.tri.face = new BxCore::DivFaceTri[ GBxCore()->SizeDivFace( 3, numDiv ) ];

	InitDivFaceBound( numDiv, dst.tri );
	InitDivFaceIn( numDiv, dst.tri );
}

void  BxTri::
InitDivFaceBound( const int numDiv, BxCore::InfoDivTri& tri )
{
	int  i, srcOfs, dstIdx0, dstIdx1, dstIdx2, dstBase0, dstBase1;

	int  dstOfs = ( numDiv + 1 ) * numDiv / 2;

	srcOfs = 0;
	for( i=0; i<( numDiv-1 ); i++ ) {
		dstIdx0 = i;
		dstIdx1 = i + 1;
		dstIdx2 = i + dstOfs;

		tri.face[ dstIdx0 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx0 ].vertex[ 1 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 0 ].faceNo       = dstIdx0;
		tri.boundVertex[ i + srcOfs ].face[ 0 ].vertexInFace = 1;

		tri.face[ dstIdx1 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx1 ].vertex[ 0 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 1 ].faceNo       = dstIdx1;
		tri.boundVertex[ i + srcOfs ].face[ 1 ].vertexInFace = 0;

		tri.face[ dstIdx2 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx2 ].vertex[ 2 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 2 ].faceNo       = dstIdx2;
		tri.boundVertex[ i + srcOfs ].face[ 2 ].vertexInFace = 2;
	}

	srcOfs = numDiv - 1;
	for( i=0; i<( numDiv-1 ); i++ ) {
		dstBase0 = i + 1;
		dstBase1 = i + 2;

		dstIdx0 = dstBase0 * ( 2 * numDiv         - dstBase0 + 1 ) / 2 - 1;
		dstIdx1 = dstBase1 * ( 2 * numDiv         - dstBase1 + 1 ) / 2 - 1;
		dstIdx2 = dstBase0 * ( 2 * ( numDiv - 1 ) - dstBase0 + 1 ) / 2 - 1 + dstOfs;

		tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 0 ].faceNo       = dstIdx0;
		tri.boundVertex[ i + srcOfs ].face[ 0 ].vertexInFace = 2;

		tri.face[ dstIdx1 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx1 ].vertex[ 1 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 1 ].faceNo       = dstIdx1;
		tri.boundVertex[ i + srcOfs ].face[ 1 ].vertexInFace = 1;

		tri.face[ dstIdx2 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx2 ].vertex[ 0 ].vertexNo.no   = i + srcOfs;
		tri.boundVertex[ i + srcOfs ].face[ 2 ].faceNo       = dstIdx2;
		tri.boundVertex[ i + srcOfs ].face[ 2 ].vertexInFace = 0;
	}

	srcOfs = ( numDiv - 1 ) * 2;
	for( i=0; i<( numDiv-1 ); i++ ) {
		int  j = ( ( numDiv - 1 ) - 1 ) - i;
		dstBase0 = i;
		dstBase1 = i + 1;

		dstIdx0 = dstBase0 * ( 2 * numDiv         - dstBase0 + 1 ) / 2;
		dstIdx1 = dstBase1 * ( 2 * numDiv         - dstBase1 + 1 ) / 2;
		dstIdx2 = dstBase0 * ( 2 * ( numDiv - 1 ) - dstBase0 + 1 ) / 2 + dstOfs;

		tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.no   = j + srcOfs;
		tri.boundVertex[ j + srcOfs ].face[ 0 ].faceNo       = dstIdx0;
		tri.boundVertex[ j + srcOfs ].face[ 0 ].vertexInFace = 2;

		tri.face[ dstIdx1 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx1 ].vertex[ 0 ].vertexNo.no   = j + srcOfs;
		tri.boundVertex[ j + srcOfs ].face[ 1 ].faceNo       = dstIdx1;
		tri.boundVertex[ j + srcOfs ].face[ 1 ].vertexInFace = 0;

		tri.face[ dstIdx2 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnBound;
		tri.face[ dstIdx2 ].vertex[ 1 ].vertexNo.no   = j + srcOfs;
		tri.boundVertex[ j + srcOfs ].face[ 2 ].faceNo       = dstIdx2;
		tri.boundVertex[ j + srcOfs ].face[ 2 ].vertexInFace = 1;
	}

	tri.face[ 0 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexOnVertex;
	tri.face[ 0 ].vertex[ 0 ].vertexNo.no   = 0;
	tri.vertexVertex[ 0 ].face.faceNo       = 0;
	tri.vertexVertex[ 0 ].face.vertexInFace = 0;

	dstBase0 = 1;
	dstIdx0  = dstBase0 * ( 2 * numDiv - dstBase0 + 1 ) / 2 - 1;
	tri.face[ dstIdx0 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexOnVertex;
	tri.face[ dstIdx0 ].vertex[ 1 ].vertexNo.no   = 1;
	tri.vertexVertex[ 1 ].face.faceNo       = dstIdx0;
	tri.vertexVertex[ 1 ].face.vertexInFace = 1;

	dstIdx0 = dstOfs - 1;
	tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexOnVertex;
	tri.face[ dstIdx0 ].vertex[ 2 ].vertexNo.no   = 2;
	tri.vertexVertex[ 2 ].face.faceNo       = dstIdx0;
	tri.vertexVertex[ 2 ].face.vertexInFace = 2;
}

void  BxTri::
InitDivFaceIn( const int numDiv, BxCore::InfoDivTri& tri )
{
	int  dstIdx, dstBase;

	int  numDiv2 = numDiv - 2;
	int  dstOfs  = ( numDiv + 1 ) * numDiv / 2;

	for( int j=0; j<numDiv2; j++ ) {
		int  srcOfs = j * ( 2 * numDiv2 - j + 1 ) / 2;
		int  num    = numDiv2 - j;

		for( int i=0; i<num; i++ ) {
			dstBase = j;
			dstIdx  = dstBase * ( 2 * numDiv - dstBase + 1 ) / 2;
			tri.face[ dstIdx + i + 1 ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i + 1 ].vertex[ 2 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 0 ].faceNo          = dstIdx + i + 1;
			tri.inVertex[ srcOfs + i ].face[ 0 ].vertexInFace    = 2;

			dstBase = j + 1;
			dstIdx  = dstBase * ( 2 * numDiv - dstBase + 1 ) / 2;

			tri.face[ dstIdx + i     ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i     ].vertex[ 1 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 1 ].faceNo          = dstIdx + i;
			tri.inVertex[ srcOfs + i ].face[ 1 ].vertexInFace    = 1;

			tri.face[ dstIdx + i + 1 ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i + 1 ].vertex[ 0 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 2 ].faceNo          = dstIdx + i + 1;
			tri.inVertex[ srcOfs + i ].face[ 2 ].vertexInFace    = 0;

			dstBase = j;
			dstIdx  = dstBase * ( 2 * ( numDiv - 1 ) - dstBase + 1 ) / 2 + dstOfs;

			tri.face[ dstIdx + i     ].vertex[ 0 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i     ].vertex[ 0 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 3 ].faceNo          = dstIdx + i;
			tri.inVertex[ srcOfs + i ].face[ 3 ].vertexInFace    = 0;

			tri.face[ dstIdx + i + 1 ].vertex[ 1 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i + 1 ].vertex[ 1 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 4 ].faceNo          = dstIdx + i + 1;
			tri.inVertex[ srcOfs + i ].face[ 4 ].vertexInFace    = 1;

			dstBase = j + 1;
			dstIdx  = dstBase * ( 2 * ( numDiv - 1 ) - dstBase + 1 ) / 2 + dstOfs;

			tri.face[ dstIdx + i     ].vertex[ 2 ].vertexNo.kind = BxCore::kVertexInFace;
			tri.face[ dstIdx + i     ].vertex[ 2 ].vertexNo.no   = srcOfs + i;
			tri.inVertex[ srcOfs + i ].face[ 5 ].faceNo          = dstIdx + i;
			tri.inVertex[ srcOfs + i ].face[ 5 ].vertexInFace    = 2;
		}
	}
}

BxErr  BxTri::
Divide( const BxCore::TriPatchInfo& src, BxCore::DivTri& dst )
{
	if( dst.tri.boundVertex == 0 || dst.tri.face == 0 )
		return kBxErrFailure;

	int  sizeInVertex = GBxCore()->SizeDivInVertex( 3, dst.numDiv );
	if( sizeInVertex > 0 && dst.tri.inVertex == 0 )
		return kBxErrFailure;

	DivideBound( src, dst );
	DivideIn( src, dst );

	return kBxErrSuccess;
}

void  BxTri::
DivideBound( const BxCore::TriPatchInfo& src, BxCore::DivTri& dst )
{
	int     i, ofs;
	double  u, v;
	BxCore::DivBoundVertexTri*  bVertex;
	BxCore::DivVertexVertex*    vVertex;

	BxCore::InfoDivTri&  tri = dst.tri;
	int  numDiv = dst.numDiv;

	ofs = 0;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &tri.boundVertex[ i + ofs ];
		u = ( double )( i + 1 ) / numDiv;
		DivideOne( src, u, 0.0, bVertex->pos, 3, bVertex->face, tri.face );
	}

	ofs = numDiv - 1;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &tri.boundVertex[ i + ofs ];
		v = ( double )( i + 1 ) / numDiv;
		DivideOne( src, 1.0, v, bVertex->pos, 3, bVertex->face, tri.face );
	}

	ofs = ( numDiv - 1 ) * 2;
	for( i=0; i<( numDiv-1 ); i++ ) {
		bVertex = &tri.boundVertex[ i + ofs ];
		v = ( double )( numDiv - 1  - i ) / numDiv;
		DivideOne( src, 0.0, v, bVertex->pos, 3, bVertex->face, tri.face );
	}

	vVertex = &tri.vertexVertex[ 0 ];
	DivideOne( src, 0.0, 0.0, vVertex->pos, 1, &vVertex->face, tri.face );

	vVertex = &tri.vertexVertex[ 1 ];
	DivideOne( src, 1.0, 0.0, vVertex->pos, 1, &vVertex->face, tri.face );

	vVertex = &tri.vertexVertex[ 2 ];
	DivideOne( src, 0.0, 1.0, vVertex->pos, 1, &vVertex->face, tri.face );
}

void  BxTri::
DivideIn( const BxCore::TriPatchInfo& src, BxCore::DivTri& dst )
{
	BxCore::DivInVertexTri*  vertex;

	BxCore::InfoDivTri&  tri = dst.tri;
	int  numDiv  = dst.numDiv;
	int  numDiv2 = numDiv - 2;

	for( int j=0; j<numDiv2; j++ ) {
		int  ofs = j * ( 2 * numDiv2 - j + 1 ) / 2;
		int  num = numDiv2 - j;

		for( int i=0; i<num; i++ ) {
			vertex = &tri.inVertex[ ofs + i ];
			double  u = ( double )( i + 1 ) / ( num + 1 );
			double  v = ( double )( j + 1 ) / numDiv;

			DivideOne( src, u, v, vertex->pos, 6, vertex->face, tri.face );
		}
	}
}

void  BxTri::
DivideOne( const BxCore::TriPatchInfo& src, const double u, const double v, BxVec3& pos,
	const int numNormal, BxCore::DivVertexFace vertexFace[], BxCore::DivFaceTri face[] )
{
	BxVec3  normal;

	BxVec2  uv( u, v );

	PosSurface( src, uv, pos );
	NormalSurface( src, uv, normal );

	for( int i=0; i<numNormal; i++ ) {
		int  faceNo       = vertexFace[ i ].faceNo;
		int  vertexInFace = vertexFace[ i ].vertexInFace;
		face[ faceNo ].vertex[ vertexInFace ].normal = normal;
	}
}


////////
