
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxVector.cpp */


#include "bxCoreGblSub.h"
#include "bxVector.h"
#include "bxOrgTo3D.h"

using namespace NameBeziex;



BxErr  BxVector::
Pos( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 pos[ 4 ] )
{
	Pos0( no, numVertex, vertex, pos[ 0 ] );
	if( Pos1( no, numVertex, vertex, pos[ 1 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Pos2( no, numVertex, vertex, pos[ 2 ] ) == kBxErrFailure )
		return kBxErrFailure;
	Pos3( no, numVertex, vertex, pos[ 3 ] );

	return kBxErrSuccess;
}

void  BxVector::
Pos0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& pos )
{
	pos = vertex[ no ].vertexPos;
}

BxErr  BxVector::
Pos1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& pos )
{
	BxVec3  org, tan;

	Pos0( no, numVertex, vertex, org );

	BxErr  result = Tan0( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	pos = org + tan;

	return kBxErrSuccess;
}

BxErr  BxVector::
Pos2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& pos )
{
	BxVec3  org, tan;

	Pos3( no, numVertex, vertex, org );

	BxErr  result = Tan2( no, numVertex, vertex, tan );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	pos = org - tan;

	return kBxErrSuccess;
}

void  BxVector::
Pos3( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& pos )
{
	int  p = ( no + 1 ) % numVertex;

	pos = vertex[ p ].vertexPos;
}

BxErr  BxVector::
Handle( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 vec[ 4 ] )
{
	if( Handle0( no, numVertex, vertex, vec[ 0 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Handle1( no, numVertex, vertex, vec[ 1 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Handle2( no, numVertex, vertex, vec[ 2 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Handle3( no, numVertex, vertex, vec[ 3 ] ) == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxVector::
Handle0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];
	
	int  m = ( no + numVertex - 1 ) % numVertex;

	BxVec3  radialVec = vertex[ m ].vertexPos - vertex[ no ].vertexPos;

	normal[ 0 ]     = vertex[ no ].vertexNormal;
	normal[ 1 ]     = vertex[ m  ].vertexNormal;
	pairNormal[ 0 ] = vertex[ m  ].edge[ 1 ].wingNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ m  ].edge[ 0 ].wingNormal[ 0 ];

	return GBxOrgTo3D()->BHandleVec( normal, radialVec, -vertex[ m ].boundRot, vertex[ m ].pairEnableBHandle,
					 pairNormal, vec );
}

BxErr  BxVector::
Handle1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  pos1, antiPos1, handle0;
	
	int  p2 = ( no + 2 ) % numVertex;

	Pos1( no, numVertex, vertex, pos1 );

	switch( numVertex ) {
	case 3:
		antiPos1 = vertex[ p2 ].vertexPos;
		break;
	case 4:
		Pos2( p2, numVertex, vertex, antiPos1 );
		break;
	default:
		return kBxErrFailure;
	}
	BxVec3  vecHandleToHandle = antiPos1 - pos1;

	Handle0( no, numVertex, vertex, handle0 );

	return GBxOrgTo3D()->IHandleVec( vertex[ no ].vertexNormal, vertex[ no ].edge[ 0 ].bNormal, handle0,
			vecHandleToHandle, vertex[ no ].edge[ 0 ].ihandle, vec );
}

BxErr  BxVector::
Handle2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  pos2, antiPos2, handle3;
	
	int  p  = ( no + 1 ) % numVertex;
	int  p2 = ( no + 2 ) % numVertex;

	Pos2( no, numVertex, vertex, pos2 );

	switch( numVertex ) {
	case 3:
		antiPos2 = vertex[ p2 ].vertexPos;
		break;
	case 4:
		Pos1( p2, numVertex, vertex, antiPos2 );
		break;
	default:
		return kBxErrFailure;
	}
	BxVec3  vecHandleToHandle = antiPos2 - pos2;

	Handle3( no, numVertex, vertex, handle3 );

	return GBxOrgTo3D()->IHandleVec( vertex[ p ].vertexNormal, vertex[ no ].edge[ 1 ].bNormal, handle3,
			vecHandleToHandle, vertex[ no ].edge[ 1 ].ihandle, vec );
}

BxErr  BxVector::
Handle3( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];
	
	int  p0 = ( no + 1 ) % numVertex;
	int  p1 = ( no + 2 ) % numVertex;

	BxVec3  radialVec = vertex[ p1 ].vertexPos - vertex[ p0 ].vertexPos;

	normal[ 0 ]     = vertex[ p0 ].vertexNormal;
	normal[ 1 ]     = vertex[ p1 ].vertexNormal;
	pairNormal[ 0 ] = vertex[ p0 ].edge[ 0 ].wingNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ p0 ].edge[ 1 ].wingNormal[ 0 ];

	return GBxOrgTo3D()->BHandleVec( normal, radialVec, vertex[ p0 ].boundRot, vertex[ p0 ].pairEnableBHandle,
					 pairNormal, vec );
}

BxErr  BxVector::
AntiHandle( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 vec[ 2 ] )
{
	if( AntiHandle0( no, numVertex, vertex, vec[ 0 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( AntiHandle3( no, numVertex, vertex, vec[ 1 ] ) == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxVector::
AntiHandle0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];

	if( vertex[ no ].pairEnable == false )
		return kBxErrFailure;

	normal[ 0 ]     = vertex[ no ].edge[ 0 ].wingNormal[ 0 ];
	normal[ 1 ]     = vertex[ no ].edge[ 0 ].wingNormal[ 1 ];
	pairNormal[ 0 ] = vertex[ no ].edge[ 0 ].wingPairNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ no ].edge[ 0 ].wingPairNormal[ 1 ];

	return GBxOrgTo3D()->BHandleVec( normal, vertex[ no ].edge[ 0 ].wingVec, vertex[ no ].edge[ 0 ].wingBoundRot,
					 vertex[ no ].edge[ 0 ].wingPairEnableBHandle, pairNormal, vec );
}

BxErr  BxVector::
AntiHandle3( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];

	if( vertex[ no ].pairEnable == false )
		return kBxErrFailure;

	normal[ 0 ]     = vertex[ no ].edge[ 1 ].wingNormal[ 0 ];
	normal[ 1 ]     = vertex[ no ].edge[ 1 ].wingNormal[ 1 ];
	pairNormal[ 0 ] = vertex[ no ].edge[ 1 ].wingPairNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ no ].edge[ 1 ].wingPairNormal[ 1 ];

	return GBxOrgTo3D()->BHandleVec( normal, vertex[ no ].edge[ 1 ].wingVec, vertex[ no ].edge[ 1 ].wingBoundRot,
					 vertex[ no ].edge[ 1 ].wingPairEnableBHandle, pairNormal, vec );
}

BxErr  BxVector::
Tan( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3 vec[ 3 ] )
{
	if( Tan0( no, numVertex, vertex, vec[ 0 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Tan1( no, numVertex, vertex, vec[ 1 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( Tan2( no, numVertex, vertex, vec[ 2 ] ) == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxVector::
Tan0( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];
	
	int  p = ( no + 1 ) % numVertex;

	BxVec3  radialVec = vertex[ p ].vertexPos - vertex[ no ].vertexPos;

	normal[ 0 ]     = vertex[ no ].vertexNormal;
	normal[ 1 ]     = vertex[ p  ].vertexNormal;
	pairNormal[ 0 ] = vertex[ no ].edge[ 0 ].wingNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ no ].edge[ 1 ].wingNormal[ 0 ];

	return GBxOrgTo3D()->BHandleVec( normal, radialVec, vertex[ no ].boundRot, vertex[ no ].pairEnableBHandle,
					 pairNormal, vec );
}

BxErr  BxVector::
Tan1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxErr   result;
	BxVec3  p1, p2;

	result = Pos1( no, numVertex, vertex, p1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	result = Pos2( no, numVertex, vertex, p2 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	
	vec = p2 - p1;

	return kBxErrSuccess;
}

BxErr  BxVector::
Tan2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  normal[ 2 ], pairNormal[ 2 ];
	
	int  p = ( no + 1 ) % numVertex;

	BxVec3  radialVec = vertex[ no ].vertexPos - vertex[ p ].vertexPos;

	normal[ 0 ]     = vertex[ p  ].vertexNormal;
	normal[ 1 ]     = vertex[ no ].vertexNormal;
	pairNormal[ 0 ] = vertex[ no ].edge[ 1 ].wingNormal[ 0 ];
	pairNormal[ 1 ] = vertex[ no ].edge[ 0 ].wingNormal[ 0 ];

	BxErr  result = GBxOrgTo3D()->BHandleVec( normal, radialVec, -vertex[ no ].boundRot,
						  vertex[ no ].pairEnableBHandle, pairNormal, vec );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	vec = -vec;

	return kBxErrSuccess;
}

BxErr  BxVector::
TriVHandle( const int no, const BxCore::SurfaceInfoVertex vertex[], BxVec3 vec[ 2 ] )
{
	if( TriVHandle1( no, vertex, vec[ 0 ] ) == kBxErrFailure )
		return kBxErrFailure;
	if( TriVHandle2( no, vertex, vec[ 1 ] ) == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxVector::
TriVHandle1( const int no, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  pos1, handle0;
	
	int  m = ( no + 3 - 1 ) % 3;

	Pos1( no, 3, vertex, pos1 );

	BxVec3  antiPos = vertex[ m ].vertexPos;
	BxVec3  vecHandleToHandle = antiPos - pos1;

	Handle0( no, 3, vertex, handle0 );

	return GBxOrgTo3D()->TriVHandleVec( vertex[ no ].vertexNormal, vertex[ no ].edge[ 0 ].bNormal, handle0,
			vecHandleToHandle, vertex[ no ].edge[ 0 ].ihandle, vertex[ m ].vertexNormal, vec );
}

BxErr  BxVector::
TriVHandle2( const int no, const BxCore::SurfaceInfoVertex vertex[], BxVec3& vec )
{
	BxVec3  pos2, handle3;
	
	int  m = ( no + 3 - 1 ) % 3;
	int  p = ( no + 1 ) % 3;

	Pos2( no, 3, vertex, pos2 );

	BxVec3  antiPos = vertex[ m ].vertexPos;
	BxVec3  vecHandleToHandle = antiPos - pos2;

	Handle3( no, 3, vertex, handle3 );

	return GBxOrgTo3D()->TriVHandleVec( vertex[ p ].vertexNormal, vertex[ no ].edge[ 1 ].bNormal, handle3,
			vecHandleToHandle, vertex[ no ].edge[ 1 ].ihandle, vertex[ m ].vertexNormal, vec );
}

BxErr  BxVector::
Normal1( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& normal )
{
	BxVec3  pos0, pos1, pos2, antiPos1;
	
	int  p2 = ( no + 2 ) % numVertex;
	int  m  = ( no + numVertex - 1 ) % numVertex;

	BxVec3  radialVec = vertex[ m ].vertexPos - vertex[ no ].vertexPos;

	Pos0( no, numVertex, vertex, pos0 );
	Pos1( no, numVertex, vertex, pos1 );
	Pos2( no, numVertex, vertex, pos2 );
	BxVec3  iHandleAxis = ( ( pos1 + pos2 ) / 2.0 ) - ( ( pos0 + 3.0 * pos1 ) / 4.0 );

	switch( numVertex ) {
	case 3:
		antiPos1 = vertex[ p2 ].vertexPos;
		break;
	case 4:
		Pos2( p2, numVertex, vertex, antiPos1 );
		break;
	default:
		return kBxErrFailure;
	}
	BxVec3  vecHandleToHandle = antiPos1 - pos1;

	return GBxOrgTo3D()->BNormal( vertex[ no ].vertexNormal, radialVec, iHandleAxis, vecHandleToHandle, normal );
}

BxErr  BxVector::
Normal2( const int no, const int numVertex, const BxCore::SurfaceInfoVertex vertex[], BxVec3& normal )
{
	BxVec3  pos1, pos2, pos3, antiPos2;
	
	int  p  = ( no + 1 ) % numVertex;
	int  p2 = ( no + 2 ) % numVertex;

	BxVec3  radialVec = vertex[ p2 ].vertexPos - vertex[ p ].vertexPos;

	Pos1( no, numVertex, vertex, pos1 );
	Pos2( no, numVertex, vertex, pos2 );
	Pos3( no, numVertex, vertex, pos3 );
	BxVec3  iHandleAxis = ( ( pos3 + 3.0 * pos2 ) / 4.0 ) - ( ( pos2 + pos1 ) / 2.0 );

	switch( numVertex ) {
	case 3:
		antiPos2 = vertex[ p2 ].vertexPos;
		break;
	case 4:
		Pos1( p2, numVertex, vertex, antiPos2 );
		break;
	default:
		return kBxErrFailure;
	}
	BxVec3  vecHandleToHandle = antiPos2 - pos2;

	return GBxOrgTo3D()->BNormal( vertex[ p ].vertexNormal, radialVec, iHandleAxis, vecHandleToHandle, normal );
}


////////
