
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxObjBoundRot.cpp */


#include "bxObjBoundRot.h"
#include "bxCoreGbl.h"

using namespace NameBeziex;



BxObjSub::ObjInfo  BxObjBoundRot::gDmyDt;	// not used

BxObjBoundRot::
BxObjBoundRot() : fDt( gDmyDt )
{
}

BxObjBoundRot::
BxObjBoundRot( BxObjSub::ObjInfo& dt )
	: fDt( dt )
{
}

BxErr  BxObjBoundRot::
Exec()
{
	BxErr  result;

	Init();

	result = CheckDoCorrect();
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = SetVHandle();
	if( result == kBxErrFailure )
		return kBxErrFailure;

	CorrectVHandle();

	result = SetBoundRot();
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

void  BxObjBoundRot::
Init()
{
	int  i;

	fInfo.vertex = new InfoVertex[ fDt.numVertex ];
	for( i=0; i<fDt.numVertex; i++ ) {
		fInfo.vertex[ i ].startFaceNo       = -1;
		fInfo.vertex[ i ].startVertexInFace = -1;
	}

	fInfo.faceVertex = new InfoFaceVertex[ fDt.numFaceVertex ];
	for( i=0; i<fDt.numFaceVertex; i++ ) {
		fInfo.faceVertex[ i ].correctOrg = false;

		for( int j=0; j<2; j++ )
			fInfo.faceVertex[ i ].vHandle[ j ].enable = false;
	}
}

BxErr  BxObjBoundRot::
CheckDoCorrect()
{
	int  i;

	for( i=0; i<fDt.numVertex; i++ )
		fInfo.vertex[ i ].doExec = false;

	for( i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( i );
		int  num = NumVertex( i );

		if( num != 3 && num != 4 )
			continue;

		BxErr  result = CheckDoCorrectOne( i, &fDt.faceVertex[ ofs ], num );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

int  BxObjBoundRot::
Ofs( const int faceNo )
{
	int  ofs;

	if( faceNo == 0 )
		ofs = 0;
	else
		ofs = fDt.faceHeader[ faceNo - 1 ];

	return ofs;
}

int  BxObjBoundRot::
NumVertex( const int faceNo )
{
	int  numVertex;

	if( faceNo == 0 )
		numVertex = fDt.faceHeader[ 0 ];
	else if( faceNo < ( fDt.numFace - 1 ) )
		numVertex = fDt.faceHeader[ faceNo ] - fDt.faceHeader[ faceNo - 1 ];
	else
		numVertex = fDt.numFaceVertex - fDt.faceHeader[ faceNo - 1 ];

	return numVertex;
}

BxErr  BxObjBoundRot::
CheckDoCorrectOne( const int faceNo, const BxObjSub::ObjInfoFaceVertex dt[], const int numVertex )
{
	int    startFaceNo, startVertexInFace, numFace;
	bool   isClose;
	BxErr  result;

	for( int i=0; i<numVertex; i++ ) {
		if( fInfo.vertex[ dt[ i ].vertexNo ].doExec == true )
			continue;

		fInfo.vertex[ dt[ i ].vertexNo ].doExec = true;

		result = CheckDoCorrectOne0( faceNo, i, numVertex, startFaceNo, startVertexInFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		result = CheckDoCorrectOne1( startFaceNo, startVertexInFace, numFace, isClose );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		if( ( numFace == 2 && isClose == false ) ||
		    ( numFace == 3                     ) ||
		    ( numFace == 4 && isClose == true  ) )
		{
			CheckDoCorrectOne2( numFace, isClose, startFaceNo, startVertexInFace );

			fInfo.vertex[ dt[ i ].vertexNo ].startFaceNo       = startFaceNo;
			fInfo.vertex[ dt[ i ].vertexNo ].startVertexInFace = startVertexInFace;
		}
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
CheckDoCorrectOne0( const int orgFaceNo, const int orgVertexInFace, const int orgNumVertex,
	int& startFaceNo, int& startVertexInFace )
{
	int  orgVertexInFaceM   = ( orgVertexInFace + orgNumVertex - 1 ) % orgNumVertex;

	int  startFaceNo0       = startFaceNo       = orgFaceNo;
	int  startVertexInFace0 = startVertexInFace = orgVertexInFace;
	
	for( int i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( startFaceNo0 );

		BxObjSub::ObjInfoFaceVertex&  faceVertex = fDt.faceVertex[ ofs + startVertexInFace0 ];
		if( faceVertex.pairFaceNo < 0 ) {
			startFaceNo       = startFaceNo0;
			startVertexInFace = startVertexInFace0;
			return kBxErrSuccess;
		}
		if( faceVertex.pairFaceNo == orgFaceNo && faceVertex.pairVertexInFace == orgVertexInFaceM )
			return kBxErrSuccess;

		startFaceNo0       = faceVertex.pairFaceNo;
		int  numVertex     = NumVertex( startFaceNo0 );

		startVertexInFace0 = ( faceVertex.pairVertexInFace + 1 ) % numVertex;
	}

	return kBxErrFailure;		
}

BxErr  BxObjBoundRot::
CheckDoCorrectOne1( const int startFaceNo, const int startVertexInFace, int& numFace, bool& isClose )
{
	TurnVertexForFace_CheckDoCorrectOne1  obj;

	return TurnVertexForFace( startFaceNo, startVertexInFace, &obj, -1, &numFace, &isClose );
}

BxErr  BxObjBoundRot::
TurnVertexForFace( const int startFaceNo, const int startVertexInFace, TurnVertexForFace_Root* turnVertex,
	const int numExec, int* numFace0, bool* isClose0 )
{
	int  faceNo       = startFaceNo;
	int  vertexInFace = startVertexInFace;

	int  numFace = 0;

	for( int i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( faceNo );

		BxErr  result = turnVertex->Proc( this, faceNo, vertexInFace, ofs, numFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		numFace++;

		if( numExec > 0 && numFace >= numExec ) {
			ResultTurnVertexForFace( numFace, false, numFace0, isClose0 );
			return kBxErrSuccess;
		}

		int  numVertex     = NumVertex( faceNo );
		int  vertexInFaceM = ( vertexInFace + numVertex - 1 ) % numVertex;

		BxObjSub::ObjInfoFaceVertex&  faceVertex = fDt.faceVertex[ ofs + vertexInFaceM ];
		if( faceVertex.pairFaceNo < 0 ) {
			ResultTurnVertexForFace( numFace, false, numFace0, isClose0 );
			return kBxErrSuccess;
		}
		if( faceVertex.pairFaceNo == startFaceNo && faceVertex.pairVertexInFace == startVertexInFace ) {
			ResultTurnVertexForFace( numFace, true, numFace0, isClose0 );
			return kBxErrSuccess;
		}

		faceNo       = faceVertex.pairFaceNo;
		vertexInFace = faceVertex.pairVertexInFace;
	}

	return kBxErrFailure;		
}

void  BxObjBoundRot::
ResultTurnVertexForFace( const int iNumFace, const bool iIsClose, int* oNumFace, bool* oIsClose )
{
	if( oNumFace != 0 )
		*oNumFace = iNumFace;

	if( oIsClose != 0 )
		*oIsClose = iIsClose;
}

void  BxObjBoundRot::
CheckDoCorrectOne2( const int numFace, const bool isClose, int& startFaceNo, int& startVertexInFace )
{
	BxVec3  normal[ 4 ];

	GetNormal( startFaceNo, startVertexInFace, normal );

	bool  doExec = false;
	if(      numFace == 3 && isClose == true )
		doExec = IsSmoothForClose3( normal, startFaceNo, startVertexInFace );
	else if( numFace == 4 && isClose == true )
		doExec = IsSmoothForClose4( normal, startFaceNo, startVertexInFace );
	else
		doExec = IsAllSmooth( numFace, normal );

	if( doExec == false )
		return;

	int  numCorrect = 0;
	if( ( numFace == 2 && isClose == false ) ||
	    ( numFace == 3 && isClose == true  ) )
	{
		numCorrect = 1;
	}
	if( ( numFace == 3 && isClose == false ) ||
	    ( numFace == 4 && isClose == true  ) )
	{
		numCorrect = 2;
	}

	SetCorrectOrg( startFaceNo, startVertexInFace, numCorrect );
}

void  BxObjBoundRot::
GetNormal( const int startFaceNo, const int startVertexInFace, BxVec3 normal[] )
{
	TurnVertexForFace_GetNormal  obj( fDt, normal );

	TurnVertexForFace( startFaceNo, startVertexInFace, &obj );
}

BxObjBoundRot::TurnVertexForFace_GetNormal::
TurnVertexForFace_GetNormal( const BxObjSub::ObjInfo& dt, BxVec3 normal[] )
	: fDt( dt )
{
	fNormal = normal;
}

BxErr  BxObjBoundRot::TurnVertexForFace_GetNormal::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
{
	fNormal[ numFace ] = fDt.faceVertex[ ofs + vertexInFace ].vertexNormal;

	return kBxErrSuccess;
}

bool  BxObjBoundRot::
IsSmoothForClose3( const BxVec3 normal[], int& startFaceNo, int& startVertexInFace )
{
	int  numSmooth = -1;
	for( int i=0; i<3; i++ ) {
		int  m = ( i + 3 - 1 ) % 3;
		if( ( normal[ i ] - normal[ m ] ).Length() <= GBxNearZero() ) {
			if( numSmooth < 0 )
				numSmooth = m;
			else
				return false;
		}
	}
	if( numSmooth < 0 )
		return false;

	bool  result;
	TurnVertexForFace_IsSmoothForClose3  obj( numSmooth, startFaceNo, startVertexInFace, result );
	TurnVertexForFace( startFaceNo, startVertexInFace, &obj );

	return result;
}

BxObjBoundRot::TurnVertexForFace_IsSmoothForClose3::
TurnVertexForFace_IsSmoothForClose3( const int numSmooth, int& startFaceNo, int& startVertexInFace, bool& result )
	: fNumSmooth( numSmooth ),
	  fStartFaceNo( startFaceNo ),
	  fStartVertexInFace( startVertexInFace ),
	  fResult( result )
{
	fResult = false;
}

BxErr  BxObjBoundRot::TurnVertexForFace_IsSmoothForClose3::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
{
	if( numFace == fNumSmooth ) {
		fStartFaceNo       = faceNo;
		fStartVertexInFace = vertexInFace;
		fResult            = true;
	}

	return kBxErrSuccess;
}

bool  BxObjBoundRot::
IsSmoothForClose4( const BxVec3 normal[], int& startFaceNo, int& startVertexInFace )
{
	int   i;
	bool  isSmooth[ 4 ];

	int  cnt = 0;
	for( i=0; i<4; i++ ) {
		int  m = ( i + 4 - 1 ) % 4;
		if( ( normal[ i ] - normal[ m ] ).Length() <= GBxNearZero() ) {
			isSmooth[ i ] = true;
			cnt++;
		}
	}

	switch( cnt ) {
	case 4:
		return true;
	case 2:
		break;
	default:
		return false;
	}

	int  no = -1;
	for( i=0; i<4; i++ ) {
		if( isSmooth[ i ] == true ) {
			no = i;
			break;
		}
	}
	if( no < 0 )
		return false;
	
	int  p0 = ( no + 2 ) % 4;
	if( isSmooth[ p0 ] == false )
		return false;

	int  numSmooth = ( no + 4 - 1 ) % 4;

	bool  result;
	TurnVertexForFace_IsSmoothForClose4  obj( numSmooth, startFaceNo, startVertexInFace, result );
	TurnVertexForFace( startFaceNo, startVertexInFace, &obj );

	return result;
}

BxObjBoundRot::TurnVertexForFace_IsSmoothForClose4::
TurnVertexForFace_IsSmoothForClose4( const int numSmooth, int& startFaceNo, int& startVertexInFace, bool& result )
	: fNumSmooth( numSmooth ),
	  fStartFaceNo( startFaceNo ),
	  fStartVertexInFace( startVertexInFace ),
	  fResult( result )
{
	fResult = false;
}

BxErr  BxObjBoundRot::TurnVertexForFace_IsSmoothForClose4::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
{
	if( numFace == fNumSmooth ) {
		fStartFaceNo       = faceNo;
		fStartVertexInFace = vertexInFace;
		fResult            = true;
	}

	return kBxErrSuccess;
}

bool  BxObjBoundRot::
IsAllSmooth( const int numFace, const BxVec3 normal[] )
{
	for( int i=0; i<numFace; i++ ) {
		int  m = ( i + numFace - 1 ) % numFace;
		if( ( normal[ i ] - normal[ m ] ).Length() > GBxNearZero() )
			return false;
	}

	return true;
}

void  BxObjBoundRot::
SetCorrectOrg( const int startFaceNo, const int startVertexInFace, const int numCorrect )
{
	TurnVertexForFace_SetCorrectOrg  obj( fInfo );

	TurnVertexForFace( startFaceNo, startVertexInFace, &obj, numCorrect );
}

BxObjBoundRot::TurnVertexForFace_SetCorrectOrg::
TurnVertexForFace_SetCorrectOrg( InfoLocal& info )
	: fInfo( info )
{
}

BxErr  BxObjBoundRot::TurnVertexForFace_SetCorrectOrg::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
{
	fInfo.faceVertex[ ofs + vertexInFace ].correctOrg = true;

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
SetVHandle()
{
	int  i;

	for( i=0; i<fDt.numVertex; i++ )
		fInfo.vertex[ i ].doExec = false;

	for( i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( i );
		int  num = NumVertex( i );

		if( num != 3 && num != 4 )
			continue;

		BxErr  result = SetVHandleOne( &fDt.faceVertex[ ofs ], num );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
SetVHandleOne( const BxObjSub::ObjInfoFaceVertex dt[], const int numVertex )
{
	for( int i=0; i<numVertex; i++ ) {
		if( fInfo.vertex[ dt[ i ].vertexNo ].doExec == true )
			continue;

		fInfo.vertex[ dt[ i ].vertexNo ].doExec = true;

		int  startFaceNo = fInfo.vertex[ dt[ i ].vertexNo ].startFaceNo;
		if( startFaceNo < 0 )
			continue;

		int  startVertexInFace = fInfo.vertex[ dt[ i ].vertexNo ].startVertexInFace;

		BxErr  result = SetVHandleOne0( startFaceNo, startVertexInFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
SetVHandleOne0( const int startFaceNo, const int startVertexInFace )
{
	TurnVertexForFace_SetVHandleOne0  obj( fDt, fInfo );

	return TurnVertexForFace( startFaceNo, startVertexInFace, &obj );
}

BxObjBoundRot::TurnVertexForFace_SetVHandleOne0::
TurnVertexForFace_SetVHandleOne0( const BxObjSub::ObjInfo& dt, InfoLocal& info )
	: fDt( dt ), fInfo( info )
{
}

BxErr  BxObjBoundRot::TurnVertexForFace_SetVHandleOne0::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
{
	BxErr   result, errResult;
	BxVec3  vHandle[ 2 ];
	bool    isEnable[ 2 ];

	if( fInfo.faceVertex[ ofs + vertexInFace ].correctOrg == false )
		return kBxErrSuccess;

	for( int i=0; i<2; i++ ) {
		isEnable[ i ] = false;
		vHandle[ i ]  = GBxZero3();
	}

	TurnVertexForEdge_SetVHandleOne0Sub0  obj0( fDt, isEnable, vHandle );
	result = This->TurnVertexForEdge( faceNo, vertexInFace, &obj0 );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable[ 0 ] == false || isEnable[ 1 ] == false )
		return kBxErrSuccess;

	BxVec3  vHandleAve = ( vHandle[ 0 ] - vHandle[ 1 ] ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	TurnVertexForEdge_SetVHandleOne0Sub1  obj1( vHandleAve, fInfo );
	result = This->TurnVertexForEdge( faceNo, vertexInFace, &obj1 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
TurnVertexForEdge( const int startFaceNo, const int startVertexInFace, TurnVertexForEdge_Root* turnVertex,
	const int numExec, int* numEdge0, bool* isClose0 )
{
	BxErr  result;

	int  faceNo       = startFaceNo;
	int  vertexInFace = startVertexInFace;

	int  numEdge = 0;

	for( int i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( faceNo );

		result = turnVertex->Proc( this, faceNo, vertexInFace, 0, ofs, numEdge );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		numEdge++;

		if( numExec > 0 && numEdge >= numExec ) {
			ResultTurnVertexForEdge( numEdge, false, numEdge0, isClose0 );
			return kBxErrSuccess;
		}

		int  numVertex     = NumVertex( faceNo );
		int  vertexInFaceM = ( vertexInFace + numVertex - 1 ) % numVertex;

		BxObjSub::ObjInfoFaceVertex&  faceVertex = fDt.faceVertex[ ofs + vertexInFaceM ];
		if( faceVertex.pairFaceNo < 0 ) {
			result = turnVertex->Proc( this, faceNo, vertexInFaceM, 1, ofs, numEdge );
			if( result == kBxErrFailure )
				return kBxErrFailure;

			numEdge++;

			ResultTurnVertexForEdge( numEdge, false, numEdge0, isClose0 );
			return kBxErrSuccess;
		}

		if( faceVertex.pairFaceNo == startFaceNo && faceVertex.pairVertexInFace == startVertexInFace ) {
			ResultTurnVertexForEdge( numEdge, true, numEdge0, isClose0 );
			return kBxErrSuccess;
		}

		faceNo       = faceVertex.pairFaceNo;
		vertexInFace = faceVertex.pairVertexInFace;
	}

	return kBxErrFailure;		
}

void  BxObjBoundRot::
ResultTurnVertexForEdge( const int iNumEdge, const bool iIsClose, int* oNumEdge, bool* oIsClose )
{
	if( oNumEdge != 0 )
		*oNumEdge = iNumEdge;

	if( oIsClose != 0 )
		*oIsClose = iIsClose;
}

BxObjBoundRot::TurnVertexForEdge_SetVHandleOne0Sub0::
TurnVertexForEdge_SetVHandleOne0Sub0( const BxObjSub::ObjInfo& dt, bool* isEnable, BxVec3* vHandle )
	: fDt( dt )
{
	fIsEnable = isEnable;
	fVHandle  = vHandle;
}

BxErr  BxObjBoundRot::TurnVertexForEdge_SetVHandleOne0Sub0::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int beginEndNo, const int ofs,
	const int numEdge )
{
	BxCore::EdgeVertexInfoBoundRot0  src;
	BxVec3  dst, vertexNormal[ 2 ];
	bool    isEnable;
	BxErr   result;

	if( numEdge != 0 && numEdge != 2 )
		return kBxErrSuccess;

	int  numVertex     = This->NumVertex( faceNo );
	int  vertexInFaceP = ( vertexInFace + 1 ) % numVertex;

	int  vertexNo  = fDt.faceVertex[ ofs + vertexInFace  ].vertexNo;
	int  vertexNoP = fDt.faceVertex[ ofs + vertexInFaceP ].vertexNo;

	result = This->GetVertexNormal( ofs, faceNo, vertexInFace, 0, vertexNormal[ 0 ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = This->GetVertexNormal( ofs, faceNo, vertexInFace, 1, vertexNormal[ 1 ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	
	if( beginEndNo == 0 ) {
		src.radialVec = fDt.vertex[ vertexNoP ] - fDt.vertex[ vertexNo ];
		src.vertexNormal[ 0 ] = vertexNormal[ 0 ];
		src.vertexNormal[ 1 ] = vertexNormal[ 1 ];
	}
	else {
		src.radialVec         = fDt.vertex[ vertexNo ] - fDt.vertex[ vertexNoP ];
		src.vertexNormal[ 0 ] = vertexNormal[ 1 ];
		src.vertexNormal[ 1 ] = vertexNormal[ 0 ];
	}

	result = GBxCore()->MakeVHandle( src, isEnable, dst );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	BxErr  errResult;
	dst = dst.Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	if( numEdge == 0 ) {
		fIsEnable[ 0 ] = true;
		fVHandle[ 0 ]  = dst;
	}
	else {
		fIsEnable[ 1 ] = true;
		fVHandle[ 1 ]  = dst;
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
GetVertexNormal( const int ofs, const int faceNo, const int vertexInFace, const int beginEndNo, BxVec3& vertexNormal )
{
	BxVec3  normal;
	BxErr   errResult;

	if( beginEndNo == 0 )
		normal = fDt.faceVertex[ ofs + vertexInFace ].vertexNormal;
	else {
		int  vertexInFaceP = ( vertexInFace + 1 ) % NumVertex( faceNo );
		normal = fDt.faceVertex[ ofs + vertexInFaceP ].vertexNormal;
	}

	if( fDt.faceVertex[ ofs + vertexInFace ].pairFaceNo >= 0 ) {
		int  pairFaceNo        = fDt.faceVertex[ ofs + vertexInFace ].pairFaceNo;
		int  pairOfs           = Ofs( pairFaceNo );
		int  pairVertexInFace  = fDt.faceVertex[ ofs + vertexInFace ].pairVertexInFace;
		if( beginEndNo == 0 )
			pairVertexInFace = ( pairVertexInFace + 1 ) % NumVertex( pairFaceNo );

		BxVec3  pairNormal = fDt.faceVertex[ pairOfs + pairVertexInFace ].vertexNormal;

		normal = ( normal + pairNormal ).Normalize( &errResult );
		if( errResult == kBxErrFailure )
			return kBxErrFailure;
	}

	vertexNormal = normal;

	return kBxErrSuccess;
}

BxObjBoundRot::TurnVertexForEdge_SetVHandleOne0Sub1::
TurnVertexForEdge_SetVHandleOne0Sub1( const BxVec3& vHandle, InfoLocal& info )
	: fVHandle( vHandle ), fInfo( info )
{
}

BxErr  BxObjBoundRot::TurnVertexForEdge_SetVHandleOne0Sub1::
Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int beginEndNo, const int ofs,
	const int numEdge )
{
	if( numEdge != 0 && numEdge != 2 )
		return kBxErrSuccess;

	fInfo.faceVertex[ ofs + vertexInFace ].vHandle[ beginEndNo ].enable = true;

	if( numEdge == 0 )
		fInfo.faceVertex[ ofs + vertexInFace ].vHandle[ beginEndNo ].vec =  fVHandle;
	else
		fInfo.faceVertex[ ofs + vertexInFace ].vHandle[ beginEndNo ].vec = -fVHandle;

	return kBxErrSuccess;
}

void  BxObjBoundRot::
CorrectVHandle()
{
	for( int i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( i );
		int  num = NumVertex( i );

		if( num != 3 && num != 4 )
			continue;

		CorrectVHandleOne( i, num, ofs );
	}
}

void  BxObjBoundRot::
CorrectVHandleOne( const int faceNo, const int numVertex, const int ofs )
{
	for( int i=0; i<numVertex; i++ ) {
		BxObjSub::ObjInfoFaceVertex&  faceVertex = fDt.faceVertex[ ofs + i ];
		if( faceVertex.pairFaceNo < 0 )
			continue;
		
		for( int j=0; j<2; j++ ) {
			if( fInfo.faceVertex[ ofs + i ].vHandle[ j ].enable == true ) {
				int      pairOfs = Ofs( faceVertex.pairFaceNo ) + faceVertex.pairVertexInFace;
				BxVec3&  vec     = fInfo.faceVertex[ ofs + i ].vHandle[ j ].vec;

				if( fInfo.faceVertex[ pairOfs ].vHandle[ 1-j ].enable == false ) {
					fInfo.faceVertex[ pairOfs ].vHandle[ 1-j ].enable = true;
					fInfo.faceVertex[ pairOfs ].vHandle[ 1-j ].vec    = vec;
				}
			}
		}
	}
}

BxErr  BxObjBoundRot::
SetBoundRot()
{
	for( int i=0; i<fDt.numFace; i++ ) {
		int  ofs = Ofs( i );
		int  num = NumVertex( i );

		if( num != 3 && num != 4 )
			continue;

		BxErr  result = SetBoundRotFace( i, num, ofs );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
SetBoundRotFace( const int faceNo, const int numVertex, const int ofs )
{
	for( int i=0; i<numVertex; i++ ) {
		BxObjSub::ObjInfoFaceVertex&  faceVertex = fDt.faceVertex[ ofs + i ];
		if( faceVertex.edgeNo < 0 )
			continue;

		BxErr  result = SetBoundRotEdge( ofs, faceNo, i, numVertex );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjBoundRot::
SetBoundRotEdge( const int ofs, const int faceNo, const int vertexInFace, const int numVertex )
{
	BxVec3  vertexNormalOrg[ 2 ], vertexPos[ 2 ], normal[ 2 ], radialVec, cross;
	BxErr   result, errResult;

	int  vertexInFaceP = ( vertexInFace + 1 ) % numVertex;

	result = GetVertexNormal( ofs, faceNo, vertexInFace, 0, vertexNormalOrg[ 0 ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = GetVertexNormal( ofs, faceNo, vertexInFace, 1, vertexNormalOrg[ 1 ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	BxObjSub::ObjInfoFaceVertex&  faceVertex  = fDt.faceVertex[ ofs + vertexInFace  ];
	BxObjSub::ObjInfoFaceVertex&  faceVertexP = fDt.faceVertex[ ofs + vertexInFaceP ];

	BxVec3  normalAveOrg = vertexNormalOrg[ 0 ] + vertexNormalOrg[ 1 ];

	vertexPos[ 0 ] = fDt.vertex[ faceVertex.vertexNo ];
	vertexPos[ 1 ] = fDt.vertex[ faceVertexP.vertexNo ];

	radialVec    = vertexPos[ 1 ] - vertexPos[ 0 ];
	normalAveOrg = Cross( Cross( radialVec, normalAveOrg ), radialVec ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	InfoFaceVertex&  info = fInfo.faceVertex[ ofs + vertexInFace ];

	int  no = -1;
	for( int i=0; i<2; i++ ) {
		if( info.vHandle[ i ].enable == false )
			return kBxErrSuccess;
		
		radialVec = vertexPos[ 1-i ] - vertexPos[ i ];
		cross     = Cross( radialVec, info.vHandle[ i ].vec );
		if( cross.Length() < GBxNearZero() ) {
			if( no >= 0 )
				return kBxErrSuccess;

			no = i;
			continue;
		}

		normal[ i ] = Cross( cross, info.vHandle[ i ].vec ).Normalize( &errResult );
		if( errResult == kBxErrFailure )
			return kBxErrFailure;
	}
	
	if( no >= 0 )
		normal[ no ] = normal[ 1 - no ];

	BxVec3  normalAveCorrect = normal[ 0 ] + normal[ 1 ];

	radialVec        = vertexPos[ 1 ] - vertexPos[ 0 ];
	normalAveCorrect = Cross( Cross( radialVec, normalAveCorrect ), radialVec ).Normalize( &errResult );
	if( errResult == kBxErrFailure )
		return kBxErrFailure;

	BxDir  boundRot;
	boundRot.X() = Dot( normalAveOrg, normalAveCorrect );
	cross = Cross( normalAveOrg, normalAveCorrect );
	boundRot.Y() = cross.Length();
	if( Dot( radialVec, cross ) < 0.0 )
		boundRot = -boundRot;

	fDt.edge[ faceVertex.edgeNo ].boundRot = boundRot;

	return kBxErrSuccess;	
}


////////
