
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxObjBoundRot.h */


#if !defined( BX_OBJBOUNDROT_H )
#define BX_OBJBOUNDROT_H

#include "bxObjSub.h"


class BxObjBoundRot {
private:
	struct SortInfo {
		int                 faceNo;
		int                 vertexInFace;
		BxObjSub::ObjInfo*  objInfo;
	};

	struct InfoVertex {
		bool  doExec;
		int   startFaceNo;
		int   startVertexInFace;
	};
	struct InfoFaceHandle {
		bool    enable;
		BxVec3  vec;
	};
	struct InfoFaceVertex {
		bool            correctOrg;
		InfoFaceHandle  vHandle[ 2 ];
	};
	struct InfoLocal {
		InfoVertex*      vertex;
		InfoFaceVertex*  faceVertex;
	};

private:
	class TurnVertexForFace_Root {
	public:
		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs,
			const int numFace ) = 0;
	};
	class TurnVertexForEdge_Root {
	public:
		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int beginEndNo, const int ofs,
			const int numEdge ) = 0;
	};

private:
	BxObjSub::ObjInfo&        fDt;
	InfoLocal                 fInfo;

	static BxObjSub::ObjInfo  gDmyDt;	// not used

private:
	BxObjBoundRot();	// not used

public:
	BxObjBoundRot( BxObjSub::ObjInfo& dt );

	BxErr
	Exec();

	int
	Ofs( const int faceNo );
	int
	NumVertex( const int faceNo );
	BxErr
	TurnVertexForFace( const int startFaceNo, const int startVertexInFace, TurnVertexForFace_Root* turnVertex,
		const int numExec = -1, int* numFace0 = 0, bool* isClose0 = 0 );
	BxErr
	TurnVertexForEdge( const int startFaceNo, const int startVertexInFace, TurnVertexForEdge_Root* turnVertex,
		const int numExec = -1, int* numEdge0 = 0, bool* isClose0 = 0 );
	BxErr
	GetVertexNormal( const int ofs, const int faceNo, const int vertexInFace, const int beginEndNo,
		BxVec3& vertexNormal );

private:
	void
	Init();
	BxErr
	CheckDoCorrect();
	BxErr
	CheckDoCorrectOne( const int faceNo, const BxObjSub::ObjInfoFaceVertex dt[], const int numVertex );
	BxErr
	CheckDoCorrectOne0( const int orgFaceNo, const int orgVertexInFace, const int orgNumVertex,
		int& startFaceNo, int& startVertexInFace );
	BxErr
	CheckDoCorrectOne1( const int startFaceNo, const int startVertexInFace, int& numFace, bool& isClose );
	void
	ResultTurnVertexForFace( const int iNumFace, const bool iIsClose, int* oNumFace, bool* oIsClose );
	void
	CheckDoCorrectOne2( const int numFace, const bool isClose, int& startFaceNo, int& startVertexInFace );
	void
	GetNormal( const int startFaceNo, const int startVertexInFace, BxVec3 normal[] );
	bool
	IsSmoothForClose3( const BxVec3 normal[], int& startFaceNo, int& startVertexInFace );
	bool
	IsSmoothForClose4( const BxVec3 normal[], int& startFaceNo, int& startVertexInFace );
	bool
	IsAllSmooth( const int numFace, const BxVec3 normal[] );
	void
	SetCorrectOrg( const int startFaceNo, const int startVertexInFace, const int numCorrect );
	BxErr
	SetVHandle();
	BxErr
	SetVHandleOne( const BxObjSub::ObjInfoFaceVertex dt[], const int numVertex );
	BxErr
	SetVHandleOne0( const int startFaceNo, const int startVertexInFace );
	void
	ResultTurnVertexForEdge( const int iNumEdge, const bool iIsClose, int* oNumEdge, bool* oIsClose );
	void
	CorrectVHandle();
	void
	CorrectVHandleOne( const int faceNo, const int numVertex, const int ofs );
	BxErr
	SetBoundRot();
	BxErr
	SetBoundRotFace( const int faceNo, const int numVertex, const int ofs );
	BxErr
	SetBoundRotEdge( const int ofs, const int faceNo, const int vertexInFace, const int numVertex );

private:
	class TurnVertexForFace_CheckDoCorrectOne1 : public TurnVertexForFace_Root {
	public:
		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace )
		{
			return kBxErrSuccess;
		}
	};

	class TurnVertexForFace_GetNormal : public TurnVertexForFace_Root {
	private:
		const BxObjSub::ObjInfo&  fDt;
		BxVec3*                   fNormal;

	public:
		TurnVertexForFace_GetNormal( const BxObjSub::ObjInfo& dt, BxVec3 normal[] );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace );
	};

	class TurnVertexForFace_IsSmoothForClose3 : public TurnVertexForFace_Root {
	private:
		const int  fNumSmooth;
		int&       fStartFaceNo;
		int&       fStartVertexInFace;
		bool&      fResult;

	public:
		TurnVertexForFace_IsSmoothForClose3( const int numSmooth, int& startFaceNo, int& startVertexInFace,
			bool& result );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace );
	};

	class TurnVertexForFace_IsSmoothForClose4 : public TurnVertexForFace_Root {
	private:
		const int  fNumSmooth;
		int&       fStartFaceNo;
		int&       fStartVertexInFace;
		bool&      fResult;

	public:
		TurnVertexForFace_IsSmoothForClose4( const int numSmooth, int& startFaceNo, int& startVertexInFace,
			bool& result );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace );
	};

	class TurnVertexForFace_SetCorrectOrg : public TurnVertexForFace_Root {
	private:
		InfoLocal&  fInfo;

	public:
		TurnVertexForFace_SetCorrectOrg( InfoLocal& info );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace );
	};

	class TurnVertexForFace_SetVHandleOne0 : public TurnVertexForFace_Root {
	private:
		const BxObjSub::ObjInfo&  fDt;
		InfoLocal&                fInfo;

	public:
		TurnVertexForFace_SetVHandleOne0( const BxObjSub::ObjInfo& dt, InfoLocal& info );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int ofs, const int numFace );
	};

	class TurnVertexForEdge_SetVHandleOne0Sub0 : public TurnVertexForEdge_Root {
	private:
		const BxObjSub::ObjInfo&  fDt;
		bool*                     fIsEnable;
		BxVec3*                   fVHandle;

	public:
		TurnVertexForEdge_SetVHandleOne0Sub0( const BxObjSub::ObjInfo& dt, bool* isEnable, BxVec3* vHandle );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int beginEndNo, const int ofs,
			const int numEdge );
	};

	class TurnVertexForEdge_SetVHandleOne0Sub1 : public TurnVertexForEdge_Root {
	private:
		const BxVec3&  fVHandle;
		InfoLocal&     fInfo;

	public:
		TurnVertexForEdge_SetVHandleOne0Sub1( const BxVec3& vHandle, InfoLocal& info );

		virtual BxErr
		Proc( BxObjBoundRot* This, const int faceNo, const int vertexInFace, const int beginEndNo, const int ofs,
			const int numEdge );
	};
};


#endif // #if !defined( BX_OBJBOUNDROT_H )

////////
