
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxObjDivSurface.cpp */


#include "bxObjDivSurface.h"
#include "bxCoreGbl.h"

using namespace NameBeziex;



const BxObjSub::ObjInfo  BxObjDivSurface::gDmySrc;	// not used
BxObj::FaceSetInfo       BxObjDivSurface::gDmyDst;	// not used

BxObjDivSurface::
BxObjDivSurface() : fSrc( gDmySrc ), fDst( gDmyDst )
{
}

BxObjDivSurface::
BxObjDivSurface( const BxObjSub::ObjInfo& src, const int numDiv, BxObj::FaceSetInfo& dst )
	: fSrc( src ), fDst( dst )
{
	fDiv.numDiv = numDiv;

	fDiv.vDstVertex = 0;
	fDiv.bDstVertex = 0;
}


BxErr  BxObjDivSurface::
Exec()
{
	BxErr  result;

	result = ExecMain0();
	if( result == kBxErrFailure ) {
		delete[] fBNormal.faceVertex;
		return kBxErrFailure;
	}

	result = ExecMain1();

	delete[] fBNormal.faceVertex;
	delete[] fDiv.vDstVertex;
	delete[] fDiv.bDstVertex;

	return result;
}

BxErr  BxObjDivSurface::
ExecMain0()
{
	BxErr  result;

	fBNormal.faceVertex = new BNormalInfoFaceVertex[ fSrc.numFaceVertex ];
	for( int i=0; i<fSrc.numFaceVertex; i++ ) {
		for( int j=0; j<2; j++ )
			fBNormal.faceVertex[ i ].bNormal[ j ] = GBxZero3();
	}

	result = BNormalProc0();
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = BNormalProc1();
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
BNormalProc0()
{
	int  ofs, num;

	for( int i=0; i<fSrc.numFace; i++ ) {
		if( i == 0 )
			ofs = 0;
		else
			ofs = fSrc.faceHeader[ i-1 ];

		if( i == 0 )
			num = fSrc.faceHeader[ 0 ];
		else if( i < ( fSrc.numFace - 1 ) )
			num = fSrc.faceHeader[ i ] - fSrc.faceHeader[ i-1 ];
		else
			num = fSrc.numFaceVertex - fSrc.faceHeader[ i-1 ];

		if( num != 3 && num != 4 )
			continue;

		BxErr  result = BNormalOne0( &fSrc.faceVertex[ ofs ], &fBNormal.faceVertex[ ofs ], num );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
BNormalOne0( BxObjSub::ObjInfoFaceVertex faceVertex[], BNormalInfoFaceVertex faceVertexEx[], const int numVertex )
{
	BxCore::SurfaceInfo  info;
	bool                 isEnable;
	BxErr                result;

	result = MakeSurfaceInfo( faceVertex, faceVertexEx, numVertex, info );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = GBxCore()->MakeBNormal( info, isEnable );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	for( int i=0; i<numVertex; i++ ) {
		faceVertexEx[ i ].bNormal[ 0 ] = info.vertex[ i ].edge[ 0 ].bNormal;
		faceVertexEx[ i ].bNormal[ 1 ] = info.vertex[ i ].edge[ 1 ].bNormal;
	}
	
	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
MakeSurfaceInfo( const BxObjSub::ObjInfoFaceVertex faceVertex[], const BNormalInfoFaceVertex faceVertexEx[],
	const int numVertex, BxCore::SurfaceInfo& info )
{
	int    ofs, pairNumVertex;
	BxErr  result;

	info.numVertex = numVertex;

	for( int i=0; i<numVertex; i++ ) {
		info.vertex[ i ].vertexPos    = fSrc.vertex[ faceVertex[ i ].vertexNo ];
		info.vertex[ i ].vertexNormal = faceVertex[ i ].vertexNormal;

		info.vertex[ i ].boundRot = GBxZeroDir();
		if( faceVertex[ i ].edgeNo >= 0 )
			info.vertex[ i ].boundRot = fSrc.edge[ faceVertex[ i ].edgeNo ].boundRot;

		info.vertex[ i ].edge[ 0 ].ihandle = faceVertex[ i ].ihandle[ 0 ];
		info.vertex[ i ].edge[ 1 ].ihandle = faceVertex[ i ].ihandle[ 1 ];

		info.vertex[ i ].edge[ 0 ].bNormal = faceVertexEx[ i ].bNormal[ 0 ];
		info.vertex[ i ].edge[ 1 ].bNormal = faceVertexEx[ i ].bNormal[ 1 ];

		info.vertex[ i ].edge[ 0 ].wingPairEnable        = false;
		info.vertex[ i ].edge[ 1 ].wingPairEnable        = false;
		info.vertex[ i ].edge[ 0 ].wingPairEnableBHandle = false;
		info.vertex[ i ].edge[ 1 ].wingPairEnableBHandle = false;

		info.vertex[ i ].pairEnableBHandle = info.vertex[ i ].pairEnable = false;
		if( faceVertex[ i ].pairFaceNo < 0 )
			continue;

		info.vertex[ i ].pairEnableBHandle = info.vertex[ i ].pairEnable = true;

		int  j = faceVertex[ i ].pairFaceNo;
		if( j < 0 ) {
			if( faceVertex[ i ].edgeNo < 0 || faceVertex[ i ].pairVertexInFace >= 0 )
				return kBxErrFailure;
			continue;
		}

		if( faceVertex[ i ].pairVertexInFace < 0 )
			return kBxErrFailure;

		if( j == 0 )
			ofs = 0;
		else
			ofs = fSrc.faceHeader[ j-1 ];

		if( j == 0 )
			pairNumVertex = fSrc.faceHeader[ 0 ];
		else if( j < ( fSrc.numFace - 1 ) )
			pairNumVertex = fSrc.faceHeader[ j ] - fSrc.faceHeader[ j-1 ];
		else
			pairNumVertex = fSrc.numFaceVertex - fSrc.faceHeader[ j-1 ];

		BxObjSub::ObjInfoFaceVertex*  pairFaceVertex   = &fSrc.faceVertex[ ofs ];
		int                           pairVertexInFace = faceVertex[ i ].pairVertexInFace;

		if( faceVertex[ i ].edgeNo < 0 ) {
			int  edgeNo = pairFaceVertex[ pairVertexInFace ].edgeNo;
			if( edgeNo < 0 )
				return kBxErrFailure;

			info.vertex[ i ].boundRot = -fSrc.edge[ edgeNo ].boundRot;
		}

		result = MakeSurfaceInfoSub0( pairFaceVertex, pairVertexInFace, pairNumVertex,
				info.vertex[ i ].edge[ 0 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		result = MakeSurfaceInfoSub1( pairFaceVertex, pairVertexInFace, pairNumVertex,
				info.vertex[ i ].edge[ 1 ] );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
MakeSurfaceInfoSub0( const BxObjSub::ObjInfoFaceVertex pairFaceVertex[], const int pairVertexInFace,
	const int pairNumVertex, BxCore::SurfaceInfoEdge& info )
{
	int  ofs, pairPairNumVertex;

	int  p1 = ( pairVertexInFace + 1 ) % pairNumVertex;
	int  p2 = ( pairVertexInFace + 2 ) % pairNumVertex;

	info.wingVec = fSrc.vertex[ pairFaceVertex[ p2 ].vertexNo ] - fSrc.vertex[ pairFaceVertex[ p1 ].vertexNo ];

	info.wingNormal[ 0 ] = pairFaceVertex[ p1 ].vertexNormal;
	info.wingNormal[ 1 ] = pairFaceVertex[ p2 ].vertexNormal;

	info.wingBoundRot = 0.0;
	if( pairFaceVertex[ p1 ].edgeNo >= 0 )
		info.wingBoundRot = fSrc.edge[ pairFaceVertex[ p1 ].edgeNo ].boundRot;

	info.wingPairEnable        = false;
	info.wingPairEnableBHandle = false;
	if( pairFaceVertex[ p1 ].pairFaceNo < -1 )
		return kBxErrSuccess;

	int  j = pairFaceVertex[ p1 ].pairFaceNo;
	if( j < 0 ) {
		if( pairFaceVertex[ p1 ].edgeNo < 0 || pairFaceVertex[ p1 ].pairVertexInFace >= 0 )
			return kBxErrFailure;
		return kBxErrSuccess;
	}

	info.wingPairEnable        = true;
	info.wingPairEnableBHandle = true;

	if( pairFaceVertex[ p1 ].pairVertexInFace < 0 )
		return kBxErrFailure;

	if( j == 0 )
		ofs = 0;
	else
		ofs = fSrc.faceHeader[ j-1 ];

	if( j == 0 )
		pairPairNumVertex = fSrc.faceHeader[ 0 ];
	else if( j < ( fSrc.numFace - 1 ) )
		pairPairNumVertex = fSrc.faceHeader[ j ] - fSrc.faceHeader[ j-1 ];
	else
		pairPairNumVertex = fSrc.numFaceVertex - fSrc.faceHeader[ j-1 ];

	BxObjSub::ObjInfoFaceVertex*  pairPairFaceVertex   = &fSrc.faceVertex[ ofs ];
	int                           pairPairVertexInFace = pairFaceVertex[ p1 ].pairVertexInFace;

	if( pairFaceVertex[ p1 ].edgeNo < 0 ) {
		int  edgeNo = pairPairFaceVertex[ pairPairVertexInFace ].edgeNo;
		if( edgeNo < 0 )
			return kBxErrFailure;

		info.wingBoundRot = -fSrc.edge[ edgeNo ].boundRot;
	}

	int  pairNo = pairPairVertexInFace;
	int  pairP  = ( pairPairVertexInFace + 1 ) % pairPairNumVertex;
	info.wingPairNormal[ 0 ] = pairPairFaceVertex[ pairP  ].vertexNormal;
	info.wingPairNormal[ 1 ] = pairPairFaceVertex[ pairNo ].vertexNormal;

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
MakeSurfaceInfoSub1( const BxObjSub::ObjInfoFaceVertex pairFaceVertex[], const int pairVertexInFace,
	const int pairNumVertex, BxCore::SurfaceInfoEdge& info )
{
	int  ofs, pairPairNumVertex;

	int  m  = ( pairVertexInFace + pairNumVertex - 1 ) % pairNumVertex;
	int  no = pairVertexInFace;

	info.wingVec = fSrc.vertex[ pairFaceVertex[ m ].vertexNo ] - fSrc.vertex[ pairFaceVertex[ no ].vertexNo ];

	info.wingNormal[ 0 ] = pairFaceVertex[ no ].vertexNormal;
	info.wingNormal[ 1 ] = pairFaceVertex[ m  ].vertexNormal;

	info.wingBoundRot = 0.0;
	if( pairFaceVertex[ m ].edgeNo >= 0 )
		info.wingBoundRot = -fSrc.edge[ pairFaceVertex[ m ].edgeNo ].boundRot;

	info.wingPairEnable        = false;
	info.wingPairEnableBHandle = false;
	if( pairFaceVertex[ m ].pairFaceNo < -1 )
		return kBxErrSuccess;

	int  j = pairFaceVertex[ m ].pairFaceNo;
	if( j < 0 ) {
		if( pairFaceVertex[ m ].edgeNo < 0 || pairFaceVertex[ m ].pairVertexInFace >= 0 )
			return kBxErrFailure;
		return kBxErrSuccess;
	}

	info.wingPairEnable        = true;
	info.wingPairEnableBHandle = true;

	if( pairFaceVertex[ m ].pairVertexInFace < 0 )
		return kBxErrFailure;

	if( j == 0 )
		ofs = 0;
	else
		ofs = fSrc.faceHeader[ j-1 ];

	if( j == 0 )
		pairPairNumVertex = fSrc.faceHeader[ 0 ];
	else if( j < ( fSrc.numFace - 1 ) )
		pairPairNumVertex = fSrc.faceHeader[ j ] - fSrc.faceHeader[ j-1 ];
	else
		pairPairNumVertex = fSrc.numFaceVertex - fSrc.faceHeader[ j-1 ];

	BxObjSub::ObjInfoFaceVertex*  pairPairFaceVertex   = &fSrc.faceVertex[ ofs ];
	int                           pairPairVertexInFace = pairFaceVertex[ m ].pairVertexInFace;

	if( pairFaceVertex[ m ].edgeNo < 0 ) {
		int  edgeNo = pairPairFaceVertex[ pairPairVertexInFace ].edgeNo;
		if( edgeNo < 0 )
			return kBxErrFailure;

		info.wingBoundRot = fSrc.edge[ edgeNo ].boundRot;
	}

	int  pairNo = pairPairVertexInFace;
	int  pairP  = ( pairPairVertexInFace + 1 ) % pairPairNumVertex;
	info.wingPairNormal[ 0 ] = pairPairFaceVertex[ pairNo ].vertexNormal;
	info.wingPairNormal[ 1 ] = pairPairFaceVertex[ pairP  ].vertexNormal;

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
BNormalProc1()
{
	int  ofs, numVertex;

	for( int i=0; i<fSrc.numFace; i++ ) {
		if( i == 0 )
			ofs = 0;
		else
			ofs = fSrc.faceHeader[ i-1 ];

		if( i == 0 )
			numVertex = fSrc.faceHeader[ 0 ];
		else if( i < ( fSrc.numFace - 1 ) )
			numVertex = fSrc.faceHeader[ i ] - fSrc.faceHeader[ i-1 ];
		else
			numVertex = fSrc.numFaceVertex - fSrc.faceHeader[ i-1 ];

		if( numVertex != 3 && numVertex != 4 )
			continue;

		BxObjSub::ObjInfoFaceVertex*  faceVertex   = &fSrc.faceVertex[ ofs ];
		BNormalInfoFaceVertex*        faceVertexEx = &fBNormal.faceVertex[ ofs ];

		for( int j=0; j<numVertex; j++ ) {
			int  p = ( j + 1 ) % numVertex;
			BxErr  result = BNormalOne1( faceVertex[ j ], faceVertex[ p ], faceVertexEx[ j ],
						     faceVertexEx[ p ] );
			if( result == kBxErrFailure )
				return kBxErrFailure;
		}
	}
	
	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
BNormalOne1( BxObjSub::ObjInfoFaceVertex& faceVertex0a, BxObjSub::ObjInfoFaceVertex& faceVertex0b,
	BNormalInfoFaceVertex& faceVertexEx0a, BNormalInfoFaceVertex& faceVertexEx0b )
{
	int     ofs, numVertex;
	double  sinDt;
	BxVec3  aveNormal;
	BxErr   errResult;

	if( faceVertex0a.pairFaceNo < 0 || faceVertex0a.edgeNo < 0 )
		return kBxErrSuccess;
	
	if( faceVertex0a.pairVertexInFace < 0 )
		return kBxErrFailure;

	if( faceVertex0a.pairFaceNo == 0 )
		ofs = 0;
	else
		ofs = fSrc.faceHeader[ faceVertex0a.pairFaceNo - 1 ];

	if( faceVertex0a.pairFaceNo == 0 )
		numVertex = fSrc.faceHeader[ 0 ];
	else if( faceVertex0a.pairFaceNo < ( fSrc.numFace - 1 ) )
		numVertex = fSrc.faceHeader[ faceVertex0a.pairFaceNo ] - fSrc.faceHeader[ faceVertex0a.pairFaceNo - 1 ];
	else
		numVertex = fSrc.numFaceVertex - fSrc.faceHeader[ faceVertex0a.pairFaceNo - 1 ];

	int  no = faceVertex0a.pairVertexInFace;
	int  p  = ( no + 1 ) % numVertex;

	BxObjSub::ObjInfoFaceVertex&  faceVertex1a = fSrc.faceVertex[ ofs + no ];
	BxObjSub::ObjInfoFaceVertex&  faceVertex1b = fSrc.faceVertex[ ofs + p  ];

	BNormalInfoFaceVertex&  faceVertexEx1a = fBNormal.faceVertex[ ofs + no ];
	BNormalInfoFaceVertex&  faceVertexEx1b = fBNormal.faceVertex[ ofs + p  ];

	sinDt = Cross( faceVertex0a.vertexNormal, faceVertex1b.vertexNormal ).Length();
	if( sinDt < GBxNearZero() ) {
		aveNormal = ( faceVertexEx0a.bNormal[ 0 ] + faceVertexEx1a.bNormal[ 1 ] ).Normalize( &errResult );
		if( errResult == kBxErrSuccess ) {
			faceVertexEx0a.bNormal[ 0 ] = aveNormal;
			faceVertexEx1a.bNormal[ 1 ] = aveNormal;
		}
	}

	sinDt = Cross( faceVertex0b.vertexNormal, faceVertex1a.vertexNormal ).Length();
	if( sinDt < GBxNearZero() ) {
		aveNormal = ( faceVertexEx0a.bNormal[ 1 ] + faceVertexEx1a.bNormal[ 0 ] ).Normalize( &errResult );
		if( errResult == kBxErrSuccess ) {
			faceVertexEx0a.bNormal[ 1 ] = aveNormal;
			faceVertexEx1a.bNormal[ 0 ] = aveNormal;
		}
	}
	
	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
ExecMain1()
{
	BxErr  err;
	int    numFace;

	Init( numFace );

	err = ExecSub0( numFace );
	if( err == kBxErrFailure )
		return kBxErrFailure;

	err = ExecSub1();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

void  BxObjDivSurface::
Init( int& numFace )
{
	int  i;

	fDiv.vDstVertex = new int[ fSrc.numVertex ];
	fDiv.bDstVertex = new int[ fSrc.numEdge ];

	for( i=0; i<fSrc.numVertex; i++ )
		fDiv.vDstVertex[ i ] = -1;

	for( i=0; i<fSrc.numEdge; i++ )
		fDiv.bDstVertex[ i ] = -1;

	GBxCore()->InitDivRect( fDiv.numDiv, fDiv.divRect );
	GBxCore()->InitDivTri( fDiv.numDiv, fDiv.divTri );

	InitFaceSetInfoEx( numFace );
}

void  BxObjDivSurface::
InitFaceSetInfoEx( int& numFace )
{
	int  i, ofs, num;

	int  sizeInVertexRect = GBxCore()->SizeDivInVertex( 4, fDiv.numDiv );
	int  sizeInVertexTri  = GBxCore()->SizeDivInVertex( 3, fDiv.numDiv );
	int  sizeFaceRect     = GBxCore()->SizeDivFace( 4, fDiv.numDiv );
	int  sizeFaceTri      = GBxCore()->SizeDivFace( 3, fDiv.numDiv );

	int  numVertex     = fSrc.numVertex;
	int  numFaceVertex = 0;

	numFace = 0;

	numVertex += fSrc.numEdge * GBxCore()->SizeDivBoundVertex( fDiv.numDiv );

	for( i=0; i<fSrc.numFace; i++ ) {
		if( i == 0 )
			ofs = 0;
		else
			ofs = fSrc.faceHeader[ i-1 ];

		if( i == 0 )
			num = fSrc.faceHeader[ 0 ];
		else if( i < ( fSrc.numFace - 1 ) )
			num = fSrc.faceHeader[ i ] - fSrc.faceHeader[ i-1 ];
		else
			num = fSrc.numFaceVertex - fSrc.faceHeader[ i-1 ];

		if( num != 3 && num != 4 )
			continue;

		switch( num ) {
		case 4:
			numVertex     += sizeInVertexRect;
			numFace       += sizeFaceRect;
			numFaceVertex += sizeFaceRect * num;
			break;
		case 3:
			numVertex     += sizeInVertexTri;
			numFace       += sizeFaceTri;
			numFaceVertex += sizeFaceTri * num;
			break;
		}
	}

	fTmp.numVertex     = 0;
	fTmp.numFace       = 0;
	fTmp.numFaceVertex = 0;

	fTmp.vertex     = new BxVec3[ numVertex ];
	if( numFace >= 2 )
		fTmp.faceHeader = new int[ numFace - 1 ];
	fTmp.faceVertex = new BxObj::FaceSetInfoFaceVertex[ numFaceVertex ];
}

BxErr  BxObjDivSurface::
ExecSub0( const int numFace )
{
	int  ofs, num;

	for( int i=0; i<fSrc.numFace; i++ ) {
		if( i == 0 )
			ofs = 0;
		else
			ofs = fSrc.faceHeader[ i-1 ];

		if( i == 0 )
			num = fSrc.faceHeader[ 0 ];
		else if( i < ( fSrc.numFace - 1 ) )
			num = fSrc.faceHeader[ i ] - fSrc.faceHeader[ i-1 ];
		else
			num = fSrc.numFaceVertex - fSrc.faceHeader[ i-1 ];

		if( num != 3 && num != 4 )
			continue;

		BxErr  result = ExecOne( &fSrc.faceVertex[ ofs ], &fBNormal.faceVertex[ ofs ], num, numFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
ExecOne( const BxObjSub::ObjInfoFaceVertex faceVertex[], const BNormalInfoFaceVertex faceVertexEx[], const int numVertex,
	const int numFace )
{
	BxCore::SurfaceInfo  src;
	bool                 isEnable;
	BxCore::PatchRec*    patch;
	BxCore::RectPatch    rectPatch;
	BxCore::TriPatch     triPatch;
	BxErr                result;

	result = MakeSurfaceInfo( faceVertex, faceVertexEx, numVertex, src );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	switch( numVertex ) {
	case 4:
		patch = &rectPatch;
		break;
	case 3:
		patch = &triPatch;
		break;
	default:
		return kBxErrSuccess;
	}

	result = GBxCore()->ConvSurface( src, isEnable, *patch );
	if( result == kBxErrFailure )
		return kBxErrFailure;
	else if( isEnable == false )
		return kBxErrSuccess;

	result = DivideProc( *patch );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	result = SetFaceSetInfoEx( faceVertex, numVertex, numFace );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
DivideProc( const BxCore::PatchRec& patch )
{
	BxCore::RectPatch*  rectPatch;
	BxCore::TriPatch*   triPatch;
	BxErr                result;

	switch( patch.numVertex ) {
	case 4:
		rectPatch = ( BxCore::RectPatch* )&patch;
		result = GBxCore()->DivideRect( rectPatch->rect, fDiv.divRect );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		break;

	case 3:
		triPatch = ( BxCore::TriPatch* )&patch;
		result = GBxCore()->DivideTri( triPatch->tri, fDiv.divTri );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		break;

	default:
		return kBxErrSuccess;
	};		

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoEx( const BxObjSub::ObjInfoFaceVertex faceVertex[], const int numVertex, const int numFace )
{
	int    faceTop;
	BxErr  result;

	switch( numVertex ) {
	case 4:
		result = SetFaceSetInfoExRect0( faceVertex, faceTop );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		result = SetFaceSetInfoExRect1( faceVertex, faceTop, numFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		break;

	case 3:
		result = SetFaceSetInfoExTri0( faceVertex, faceTop );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		result = SetFaceSetInfoExTri1( faceVertex, faceTop, numFace );
		if( result == kBxErrFailure )
			return kBxErrFailure;
		break;
	};

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoExRect0( const BxObjSub::ObjInfoFaceVertex faceVertex[], int& faceTop )
{
	int  edgeNo, i;

	int  numVertex       = 4;
	int  sizeBoundVertex = GBxCore()->SizeDivBoundVertex( fDiv.numDiv );
	int  sizeInVertex    = GBxCore()->SizeDivInVertex( numVertex, fDiv.numDiv );

	int  srcBoundVertexNo = 0;
	for( i=0; i<numVertex; i++ ) {
		if( fDiv.vDstVertex[ faceVertex[ i ].vertexNo ] < 0 ) {
			fDiv.vDstVertex[ faceVertex[ i ].vertexNo ] = fTmp.numVertex;
			fTmp.vertex[ fTmp.numVertex ] = fDiv.divRect.rect.vertexVertex[ i ].pos;

			fTmp.numVertex++;
		}

		bool  reverse;
		BxErr  result = GetEdgeNo( faceVertex[ i ], edgeNo, reverse );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		if( fDiv.bDstVertex[ edgeNo ] < 0 ) {
			fDiv.bDstVertex[ edgeNo ] = fTmp.numVertex;

			int  srcBoundVertexNo2;
			if( reverse == false )
				srcBoundVertexNo2 = srcBoundVertexNo;
			else
				srcBoundVertexNo2 = srcBoundVertexNo + sizeBoundVertex - 1;
			for( int j=0; j<sizeBoundVertex; j++ ) {
				fTmp.vertex[ fTmp.numVertex ] = fDiv.divRect.rect.boundVertex[ srcBoundVertexNo2 ].pos;

				if( reverse == false )
					srcBoundVertexNo2++;
				else
					srcBoundVertexNo2--;

				fTmp.numVertex++;
			}
		}
		
		srcBoundVertexNo += sizeBoundVertex;
	}

	faceTop = fTmp.numVertex;

	for( i=0; i<sizeInVertex; i++ ) {
		fTmp.vertex[ fTmp.numVertex ] = fDiv.divRect.rect.inVertex[ i ].pos;
		fTmp.numVertex++;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoExRect1( const BxObjSub::ObjInfoFaceVertex faceVertex[], const int faceTop, const int numFace )
{
	int  numVertex = 4;
	int  sizeFace  = GBxCore()->SizeDivFace( numVertex, fDiv.numDiv );

	for( int i=0; i<sizeFace; i++ ) {
		for( int j=0; j<numVertex; j++ ) {
			BxCore::DivFace&  vertex = fDiv.divRect.rect.face[ i ].vertex[ j ];

			BxErr  result = SetFaceSetInfoExSub1( faceVertex, vertex, faceTop );
			if( result == kBxErrFailure )
				return kBxErrFailure;
		}

		if( fTmp.numFace < ( numFace - 1 ) ) {
			if( fTmp.numFace == 0 )
				fTmp.faceHeader[ 0 ] = numVertex;
			else if( fTmp.numFace >= 1 )
				fTmp.faceHeader[ fTmp.numFace ] = fTmp.faceHeader[ fTmp.numFace-1 ] + numVertex;
		}

		fTmp.numFace++;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoExTri0( const BxObjSub::ObjInfoFaceVertex faceVertex[], int& faceTop )
{
	int  edgeNo, i;

	int  numVertex       = 3;
	int  sizeBoundVertex = GBxCore()->SizeDivBoundVertex( fDiv.numDiv );
	int  sizeInVertex    = GBxCore()->SizeDivInVertex( numVertex, fDiv.numDiv );

	int  srcBoundVertexNo = 0;
	for( i=0; i<numVertex; i++ ) {
		if( fDiv.vDstVertex[ faceVertex[ i ].vertexNo ] < 0 ) {
			fDiv.vDstVertex[ faceVertex[ i ].vertexNo ] = fTmp.numVertex;
			fTmp.vertex[ fTmp.numVertex ] = fDiv.divTri.tri.vertexVertex[ i ].pos;

			fTmp.numVertex++;
		}

		bool  reverse;
		BxErr  result = GetEdgeNo( faceVertex[ i ], edgeNo, reverse );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		if( fDiv.bDstVertex[ edgeNo ] < 0 ) {
			fDiv.bDstVertex[ edgeNo ] = fTmp.numVertex;

			int  srcBoundVertexNo2;
			if( reverse == false )
				srcBoundVertexNo2 = srcBoundVertexNo;
			else
				srcBoundVertexNo2 = srcBoundVertexNo + sizeBoundVertex - 1;
			for( int j=0; j<sizeBoundVertex; j++ ) {
				fTmp.vertex[ fTmp.numVertex ] = fDiv.divTri.tri.boundVertex[ srcBoundVertexNo2 ].pos;

				if( reverse == false )
					srcBoundVertexNo2++;
				else
					srcBoundVertexNo2--;

				fTmp.numVertex++;
			}
		}

		srcBoundVertexNo += sizeBoundVertex;
	}

	faceTop = fTmp.numVertex;

	for( i=0; i<sizeInVertex; i++ ) {
		fTmp.vertex[ fTmp.numVertex ] = fDiv.divTri.tri.inVertex[ i ].pos;
		fTmp.numVertex++;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoExTri1( const BxObjSub::ObjInfoFaceVertex faceVertex[], const int faceTop, const int numFace )
{
	int  numVertex = 3;
	int  sizeFace  = GBxCore()->SizeDivFace( numVertex, fDiv.numDiv );

	for( int i=0; i<sizeFace; i++ ) {
		for( int j=0; j<numVertex; j++ ) {
			BxCore::DivFace&  vertex = fDiv.divTri.tri.face[ i ].vertex[ j ];

			BxErr  result = SetFaceSetInfoExSub1( faceVertex, vertex, faceTop );
			if( result == kBxErrFailure )
				return kBxErrFailure;
		}

		if( fTmp.numFace < ( numFace - 1 ) ) {
			if( fTmp.numFace == 0 )
				fTmp.faceHeader[ 0 ] = numVertex;
			else if( fTmp.numFace >= 1 )
				fTmp.faceHeader[ fTmp.numFace ] = fTmp.faceHeader[ fTmp.numFace-1 ] + numVertex;
		}

		fTmp.numFace++;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
SetFaceSetInfoExSub1( const BxObjSub::ObjInfoFaceVertex faceVertex[], const BxCore::DivFace& vertex, const int faceTop )
{
	int    vertexNo, vertexInFace, subNo, edgeNo;
	BxErr  result;

	int  sizeBoundVertex = GBxCore()->SizeDivBoundVertex( fDiv.numDiv );

	switch( vertex.vertexNo.kind ) {
	case BxCore::kVertexOnVertex:
		vertexNo = faceVertex[ vertex.vertexNo.no ].vertexNo;
		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNo     = fDiv.vDstVertex[ vertexNo ];
		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNormal = vertex.normal;
		fTmp.numFaceVertex++;
		break;

	case BxCore::kVertexOnBound:
		vertexInFace = vertex.vertexNo.no / sizeBoundVertex;
		subNo        = vertex.vertexNo.no % sizeBoundVertex;

		bool  reverse;
		result = GetEdgeNo( faceVertex[ vertexInFace ], edgeNo, reverse );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		if( reverse == true )
			subNo = sizeBoundVertex - 1 - subNo;

		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNo     = fDiv.bDstVertex[ edgeNo ] + subNo;
		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNormal = vertex.normal;
		fTmp.numFaceVertex++;
		break;

	case BxCore::kVertexInFace:
		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNo     = faceTop + vertex.vertexNo.no;
		fTmp.faceVertex[ fTmp.numFaceVertex ].vertexNormal = vertex.normal;
		fTmp.numFaceVertex++;
		break;
	}

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
GetEdgeNo( const BxObjSub::ObjInfoFaceVertex& faceVertex, int& edgeNo, bool& reverse )
{
	int  ofs;

	edgeNo = faceVertex.edgeNo;
	if( edgeNo >= 0 ) {
		reverse = false;
		return kBxErrSuccess;
	}

	int  j = faceVertex.pairFaceNo;
	if( j < 0 || faceVertex.pairVertexInFace < 0 )
		return kBxErrFailure;

	if( j == 0 )
		ofs = 0;
	else
		ofs = fSrc.faceHeader[ j-1 ];

	BxObjSub::ObjInfoFaceVertex*  pairFaceVertex   = &fSrc.faceVertex[ ofs ];
	int                           pairVertexInFace = faceVertex.pairVertexInFace;

	edgeNo = pairFaceVertex[ pairVertexInFace ].edgeNo;
	if( edgeNo < 0 )
		return kBxErrFailure;

	reverse = true;

	return kBxErrSuccess;
}

BxErr  BxObjDivSurface::
ExecSub1()
{
	int  i;

	if( fTmp.numVertex == 0 || fTmp.numFace == 0 )
		return kBxErrFailure;

	fDst.numVertex = fTmp.numVertex;
	fDst.vertex = new BxVec3[ fTmp.numVertex ];
	for( i=0; i<fTmp.numVertex; i++ )
		fDst.vertex[ i ] = fTmp.vertex[ i ];	

	fDst.numFace = fTmp.numFace;
	if( fTmp.numFace >= 2 ) {
		fDst.faceHeader = new int[ fTmp.numFace - 1 ];

		for( i=0; i<( fTmp.numFace - 1 ); i++ )
			fDst.faceHeader[ i ] = fTmp.faceHeader[ i ];
	}

	fDst.numFaceVertex = fTmp.numFaceVertex;
	fDst.faceVertex = new BxObj::FaceSetInfoFaceVertex[ fTmp.numFaceVertex ];
	for( i=0; i<fTmp.numFaceVertex; i++ )
		fDst.faceVertex[ i ] = fTmp.faceVertex[ i ];	

	return kBxErrSuccess;
}


////////
