
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* bxObjMakeObj.cpp */


#include <stdlib.h>
#include "bxObjMakeObj.h"

using namespace NameBeziex;



const BxObj::FaceSetInfo  BxObjMakeObj::gDmySrc;	// not used
BxObjSub::ObjInfo         BxObjMakeObj::gDmyDst;	// not used

BxObjMakeObj::
BxObjMakeObj() : fSrc( gDmySrc ), fDst( gDmyDst )
{
}

BxObjMakeObj::
BxObjMakeObj( const BxObj::FaceSetInfo& src, BxObjSub::ObjInfo& dst )
	: fSrc( src ), fDst( dst )
{
	fSortInfo = 0;
}

BxErr  BxObjMakeObj::
Exec()
{
	BxErr  result;
	int    numVertex;

	int*  vertexNoTable = new int[ fSrc.numVertex ];

	result = Init0( vertexNoTable, numVertex );
	if( result == kBxErrFailure ) {
		delete[] vertexNoTable;
		return kBxErrFailure;
	}

	Alloc( numVertex );

	result = ExecMain( vertexNoTable );

	delete[] fSortInfo;
	delete[] vertexNoTable;

	return result;
}

BxErr  BxObjMakeObj::
ExecMain( const int* vertexNoTable )
{
	BxErr  result;
	int    numEdge0;

	result = Init1( vertexNoTable );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	Sort();
	result = Correct( numEdge0 );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	EdgeProc( numEdge0 );

	return kBxErrSuccess;
}

BxErr  BxObjMakeObj::
Init0( int* vertexNoTable, int& numVertex )
{
	int  i;

	for( i=0; i<fSrc.numVertex; i++ )
		vertexNoTable[ i ] = -1;

	for( i=0; i<fSrc.numFaceVertex; i++ ) {
		int  no = fSrc.faceVertex[ i ].vertexNo;
		if( 0 > no || no >= fSrc.numVertex )
			return kBxErrFailure;

		vertexNoTable[ no ] = 0;
	}

	numVertex = 0;
	for( i=0; i<fSrc.numVertex; i++ ) {
		if( vertexNoTable[ i ] >= 0 ) {
			vertexNoTable[ i ] = numVertex;
			numVertex++;
		}
	}

	return kBxErrSuccess;
}


void  BxObjMakeObj::
Alloc( const int numVertex )
{
	fDst.numVertex     = numVertex;
	fDst.numFace       = fSrc.numFace;
	fDst.numFaceVertex = fSrc.numFaceVertex;

	fDst.vertex     = new BxVec3[ fDst.numVertex ];
	fDst.faceHeader = new int[ fDst.numFace - 1 ];
	fDst.faceVertex = new BxObjSub::ObjInfoFaceVertex[ fDst.numFaceVertex ];

	fSortInfo = new SortInfo[ fDst.numFaceVertex ];
}

BxErr  BxObjMakeObj::
Init1( const int* vertexNoTable )
{
	int  i, num, cnt;

	cnt = 0;
	for( i=0; i<fSrc.numVertex; i++ ) {
		if( vertexNoTable[ i ] >= 0 ) {
			fDst.vertex[ cnt ] = fSrc.vertex[ i ];
			cnt++;
		}
	}

	for( i=0; i<( fSrc.numFace - 1 ); i++ )
		fDst.faceHeader[ i ] = fSrc.faceHeader[ i ];

	for( i=0; i<fSrc.numFaceVertex; i++ ) {
		fDst.faceVertex[ i ].vertexNo     = vertexNoTable[ fSrc.faceVertex[ i ].vertexNo ];
		fDst.faceVertex[ i ].vertexNormal = fSrc.faceVertex[ i ].vertexNormal;

		fDst.faceVertex[ i ].pairFaceNo       = -1;
		fDst.faceVertex[ i ].pairVertexInFace = -1;

		fDst.faceVertex[ i ].edgeNo = -2;

		for( int j=0; j<2; j++ )
			fDst.faceVertex[ i ].ihandle[ j ] = GBxZeroDir();
	}

	cnt = 0;
	for( i=0; i<fDst.numFace; i++ ) {
		if( i == 0 )
			num = fDst.faceHeader[ 0 ];
		else if( i < ( fDst.numFace - 1 ) )
			num = fDst.faceHeader[ i ] - fDst.faceHeader[ i-1 ];
		else
			num = fDst.numFaceVertex - fDst.faceHeader[ i-1 ];

		for( int j=0; j<num; j++ ) {
			if( cnt > fDst.numFaceVertex )
				return kBxErrFailure;

			fSortInfo[ cnt ].faceNo       = i;
			fSortInfo[ cnt ].vertexInFace = j;
			fSortInfo[ cnt ].objInfo      = &fDst;

			cnt++;
		}
	}
	if( cnt != fDst.numFaceVertex )
		return kBxErrFailure;

	return kBxErrSuccess;
}

void  BxObjMakeObj::
Sort()
{
	qsort( fSortInfo, fDst.numFaceVertex, sizeof( SortInfo ), SortProc );
}

int  BxObjMakeObj::
SortProc( const void* idt0, const void* idt1 )
{
	int  ofs0, ofs1;

	SortInfo*  dt0 = ( SortInfo* )idt0;
	SortInfo*  dt1 = ( SortInfo* )idt1;

	BxObjSub::ObjInfo*  objInfo = dt0->objInfo;

	if( dt0->faceNo == 0 )
		ofs0 = 0;
	else
		ofs0 = objInfo->faceHeader[ dt0->faceNo - 1 ];

	if( dt1->faceNo == 0 )
		ofs1 = 0;
	else
		ofs1 = objInfo->faceHeader[ dt1->faceNo - 1 ];

	int  faceVertex0 = ofs0 + dt0->vertexInFace;
	int  faceVertex1 = ofs1 + dt1->vertexInFace;

	int  vertexNo0 = objInfo->faceVertex[ faceVertex0 ].vertexNo;
	int  vertexNo1 = objInfo->faceVertex[ faceVertex1 ].vertexNo;

	if( vertexNo0 < vertexNo1 )
		return -1;
	else if( vertexNo0 > vertexNo1 )
		return  1;

	return 0;
}

BxErr  BxObjMakeObj::
Correct( int& numEdge0 )
{
	int  edgeCnt = 0;
	int  j       = 0;

	for( int i=0; i<fDst.numFaceVertex; i=j ) {
		int  vertexNo0 = GetVertexNo( fSortInfo[ i ] );

		for( j=i+1; j<fDst.numFaceVertex; j++ ) {
			int  vertexNo1 = GetVertexNo( fSortInfo[ j ] );
			if( vertexNo0 != vertexNo1 )
				break;
		}
		int  num = j - i;
			
		BxErr  result = CorrectSub0( &fSortInfo[ i ], num, edgeCnt );
		if( result == kBxErrFailure )
			return kBxErrFailure;
	}

	numEdge0 = edgeCnt;

	return kBxErrSuccess;
}

int  BxObjMakeObj::
GetVertexNo( const SortInfo& sortInfo )
{
	int  ofs;

	if( sortInfo.faceNo == 0 )
		ofs = 0;
	else
		ofs = fDst.faceHeader[ sortInfo.faceNo - 1 ];

	int  faceVertex = ofs + sortInfo.vertexInFace;

	return fDst.faceVertex[ faceVertex ].vertexNo;
}

BxErr  BxObjMakeObj::
CorrectSub0( SortInfo sortInfo[], const int num, int& edgeCnt )
{
	for( int i=0; i<num; i++ ) {
		BxObjSub::ObjInfoFaceVertex&  faceVertex0 = GetFaceVertex( sortInfo[ i ], -1 );

		if( faceVertex0.pairFaceNo >= 0 ) {
			if( faceVertex0.pairVertexInFace < 0 )
				return kBxErrFailure;

			int  vertexInFace0 = GetVertexInFace( sortInfo[ i ], -1 );
			BxObjSub::ObjInfoFaceVertex  faceVertex1;
			PairFaceVertex( faceVertex0, faceVertex1 );
			if( sortInfo[ i ].faceNo != faceVertex1.pairFaceNo       ||
			    vertexInFace0        != faceVertex1.pairVertexInFace )
			{
				return kBxErrFailure;
			}
			
			continue;
		}

		int  vertexNo0 = faceVertex0.vertexNo;

		for( int j=0; j<num; j++ ) {
			if( i == j )
				continue;

			BxObjSub::ObjInfoFaceVertex&  faceVertex1 = GetFaceVertex( sortInfo[ j ], 0 );
			if( faceVertex1.pairFaceNo >= 0 ) {
				if( faceVertex1.pairVertexInFace < 0 )
					return kBxErrFailure;

				continue;
			}

			int  vertexNo1 = GetVertexNoOne( sortInfo[ j ], 1 );
			if( vertexNo0 != vertexNo1 )
				continue;

			CorrectSub00( sortInfo[ i ], -1, edgeCnt, faceVertex1 );
			CorrectSub00( sortInfo[ j ],  0,      -1, faceVertex0 );

			edgeCnt++;

			break;
		}
	}

	return kBxErrSuccess;
}

int  BxObjMakeObj::
GetVertexNoOne( const SortInfo& sortInfo, const int vertexOfs )
{
	BxObjSub::ObjInfoFaceVertex&  faceVertex = GetFaceVertex( sortInfo, vertexOfs );

	return faceVertex.vertexNo;
}

int  BxObjMakeObj::
GetVertexInFace( const SortInfo& sortInfo, const int vertexOfs )
{
	int  numVertex;

	if( sortInfo.faceNo == 0 )
		numVertex = fDst.faceHeader[ 0 ];
	else if( sortInfo.faceNo < ( fDst.numFace - 1 ) )
		numVertex = fDst.faceHeader[ sortInfo.faceNo ] - fDst.faceHeader[ sortInfo.faceNo - 1 ];
	else
		numVertex = fDst.numFaceVertex - fDst.faceHeader[ sortInfo.faceNo - 1 ];

	int  vertexInFace = ( sortInfo.vertexInFace + numVertex + vertexOfs ) % numVertex;

	return vertexInFace;
}

BxObjSub::ObjInfoFaceVertex&  BxObjMakeObj::
GetFaceVertex( const SortInfo& sortInfo, const int vertexOfs )
{
	int  ofs;

	if( sortInfo.faceNo == 0 )
		ofs = 0;
	else
		ofs = fDst.faceHeader[ sortInfo.faceNo - 1 ];

	int  vertexInFace = GetVertexInFace( sortInfo, vertexOfs );
	int  faceVertexNo = ofs + vertexInFace;

	return fDst.faceVertex[ faceVertexNo ];
}

void  BxObjMakeObj::
PairFaceVertex( const BxObjSub::ObjInfoFaceVertex& src, BxObjSub::ObjInfoFaceVertex& dst )
{
	int  ofs;

	if( src.pairFaceNo == 0 )
		ofs = 0;
	else
		ofs = fDst.faceHeader[ src.pairFaceNo - 1 ];

	int  faceVertex = ofs + src.pairVertexInFace;

	dst = fDst.faceVertex[ faceVertex ];
}

void  BxObjMakeObj::
CorrectSub00( const SortInfo& sortInfo, const int vertexOfs, const int edgeNo, BxObjSub::ObjInfoFaceVertex& faceVertex )
{
	int  numVertex;

	if( sortInfo.faceNo == 0 )
		numVertex = fDst.faceHeader[ 0 ];
	else if( sortInfo.faceNo < ( fDst.numFace - 1 ) )
		numVertex = fDst.faceHeader[ sortInfo.faceNo ] - fDst.faceHeader[ sortInfo.faceNo - 1 ];
	else
		numVertex = fDst.numFaceVertex - fDst.faceHeader[ sortInfo.faceNo - 1 ];

	faceVertex.pairFaceNo       = sortInfo.faceNo;
	faceVertex.pairVertexInFace = ( sortInfo.vertexInFace + numVertex + vertexOfs ) % numVertex;

	faceVertex.edgeNo           = edgeNo;
}

void  BxObjMakeObj::
EdgeProc( const int numEdge0 )
{
	int  i;

	int  numEdge = numEdge0;
	for( i=0; i<fDst.numFaceVertex; i++ ) {
		if( fDst.faceVertex[ i ].edgeNo == -2 ) {
			fDst.faceVertex[ i ].edgeNo = numEdge;
			numEdge++;
		}
	}

	fDst.numEdge = numEdge;
	fDst.edge    = new BxObjSub::ObjInfoEdge[ numEdge ];

	for( i=0; i<numEdge; i++ )
		fDst.edge[ i ].boundRot = GBxZeroDir();
}


////////
