
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* divObj.cpp */


#include <string.h>

#include "divObj.h"
#include "divObjReadObj.h"
#include "divObjWriteObj.h"
#include "divObjWriteVrml.h"



BxErr  DivObj::
Exec( const char* readFname, const int numDiv, const char* writeFname )
{
	BxErr               err;
	BxObj::FaceSetInfo  src, dst;

	DivObjReadObj    objReadObj( readFname, src );
	DivObjMain       objMain( src, numDiv, dst );
	DivObjWriteObj   objWriteObj( writeFname, dst );
	DivObjWriteVrml  objWriteVrml( writeFname, dst );

	err = objReadObj.Exec();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	err = objMain.Exec();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	bool  isVrmlFile = IsVrmlFile( writeFname );
	
	if( isVrmlFile == true ) {
		err = objWriteVrml.Exec();
		if( err == kBxErrFailure )
			return kBxErrFailure;
	}
	else {
		err = objWriteObj.Exec();
		if( err == kBxErrFailure )
			return kBxErrFailure;
	}

	return kBxErrSuccess;
}

bool  DivObj::
IsVrmlFile( const char* writeFname )
{
	enum {
		kLenExtPlus = 4		// == strlen( kStrExtPlus )
	};
	static const char  kStrExtPlus[] = ".wrl";

	int  len = strlen( writeFname );
	if( len < kLenExtPlus )
		return false;

	const char*  str = &writeFname[ len - kLenExtPlus ];

	if( _stricmp( str, kStrExtPlus ) == 0 )
		return true;

	return false;
}


DivObjMain::
DivObjMain( const BxObj::FaceSetInfo& src, const int numDiv, BxObj::FaceSetInfo& dst )
	: fSrc    ( src ),
	  fNumDiv ( numDiv ),
	  fDst    ( dst )
{
}

BxErr  DivObjMain::
Exec()
{
	BxObj  obj;
	BxErr  err;

	err = obj.MakeObj( fSrc );
	if( err == kBxErrFailure )
		return kBxErrFailure;
	
	err = obj.BoundRot();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	err = obj.DivSurface( fNumDiv, fDst );
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}


////////
