
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* divObjReadObj.cpp */


#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <sys\stat.h>

#include "divObjReadObj.h"



DivObjReadObj::
DivObjReadObj( const char* fname, BxObj::FaceSetInfo& faceSetInfo )
	: fFname       ( fname ),
	  fFaceSetInfo ( faceSetInfo )
{
	fNumMemDt      = 0;
	fMemDt         = 0;

	fNumVertex     = 0;
	fNumNormal     = 0;
	fNumFace       = 0;
	fNumFaceVertex = 0;

	fNormal        = 0;
}

BxErr  DivObjReadObj::
Exec()
{
	BxErr  err;

	err = FileToMem();
	if( err == kBxErrFailure ) {
		Delete( 0 );
		return kBxErrFailure;
	}

	err = Pass0();
	if( err == kBxErrFailure ) {
		Delete( 0 );
		return kBxErrFailure;
	}

	err = Pass1();
	if( err == kBxErrFailure ) {
		Delete( fNormal );
		return kBxErrFailure;
	}

	err = Pass2();

	Delete( fNormal );

	return err;
}

void  DivObjReadObj::
Delete( BxVec3* normalBuf )
{
	if( fMemDt != 0 ) {
		delete[] fMemDt;
		fNumMemDt = 0;
		fMemDt    = 0;
	}

	if( normalBuf != 0 ) {
		delete[] normalBuf;
		normalBuf = 0;
	}
}


BxErr  DivObjReadObj::
FileToMem()
{
	struct stat  statbuf;

	FILE*  fp = fopen( fFname, "rb" );
	if( fp == 0 )
		return kBxErrFailure;

	fstat( _fileno( fp ), &statbuf );

	fNumMemDt = statbuf.st_size;
	fMemDt    = new char[ fNumMemDt ];

	fread( fMemDt, 1, fNumMemDt, fp );
	fclose( fp );

	return kBxErrSuccess;
}

BxErr  DivObjReadObj::
Pass0()
{
	DivObjReadObjPass0  obj0( fNumMemDt, fMemDt, fNumVertex, fNumNormal, fNumFace, fNumFaceVertex );

	BxErr  err = obj0.Exec();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObj::
Pass1()
{
	DivObjReadObjPass1  obj1( fNumMemDt, fMemDt, fNumNormal, fNormal );

	BxErr  err = obj1.Exec();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObj::
Pass2()
{
	DivObjReadObjPass2  obj2( fNumMemDt, fMemDt, fNumVertex, fNumNormal, fNumFace, fNumFaceVertex,
			          fNormal, fFaceSetInfo );

	BxErr  err = obj2.Exec();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}


DivObjReadObjMain::
DivObjReadObjMain( const int numMemDt, const char* memDt )
{
	fNumMemDt = numMemDt;
	fMemDt    = memDt;
}

BxErr  DivObjReadObjMain::
Exec()
{
	BxErr  err;

	err = Init();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	int  ofs    = 0;
	int  newOfs = 0;
	for(;;) {
		int  num = GetNextLine( fNumMemDt, fMemDt, newOfs );
		if( num > 0 ) {
			err = ExecMain( ofs, num );
			if( err == kBxErrFailure )
				return kBxErrFailure;
		}

		if( newOfs >= fNumMemDt )
			break;

		ofs = newOfs;
	}

	err = Finish();
	if( err == kBxErrFailure )
		return kBxErrFailure;

	return kBxErrSuccess;
}


int  DivObjReadObjMain::
GetNextLine( const int numMemDt, const char* memDt, int& ofs )
{
	int  num = 0;

	bool  doExec = false;
	while( ofs < numMemDt ) {
		doExec = true;

		if( memDt[ ofs ] == '\x0d' || memDt[ ofs ] == '\x0a' ) {
			ofs++;
			break;
		}

		num++;
		ofs++;
	}

	if( doExec == true && ofs < numMemDt && ( memDt[ ofs - 1 ] == '\x0d' && memDt[ ofs ] == '\x0a' ) )
		ofs++;

	return num;
}


BxErr  DivObjReadObjMain::
ExecMain( const int iOfs, const int iNum )
{
	enum {
		kKindUnknown,
		kKindVertex,
		kKindNormal,
		kKindFace
	};

	BxErr  err;
	char*  str;
	
	str = ( char* )&fMemDt[ iOfs ];
	int  ofsSkip = 0;
	if( SkipSpace( str, iNum, ofsSkip ) == true )
		return kBxErrFailure;

	str  = ( char* )&fMemDt[ iOfs + ofsSkip ];

	int  kind = kKindUnknown;
	int  inc  = 0;
	switch( str[ 0 ] ) {
	case 'v':
		if( ( iNum - ofsSkip ) <= 1 || isspace( str[ 1 ] ) != 0 ) {
			kind = kKindVertex;
			inc  = 1;
		}
		else if( str[ 1 ] == 'n' ) {
			kind = kKindNormal;
			inc  = 2;
		}
		break;

	case 'f':
		kind = kKindFace;
		inc  = 1;
		break;
	}

	int  ofs = iOfs + ( ofsSkip + inc );
	int  num = iNum - ( ofsSkip + inc );
	if( num <= 0 )
		return kBxErrFailure;
	if( kind != kKindUnknown && num > kLenDtStr )
		return kBxErrFailure;

	switch( kind ) {
	case kKindVertex:
		err = VertexProc( ofs, num );
		if( err == kBxErrFailure )
			return kBxErrFailure;
		break;

	case kKindNormal:
		err = NormalProc( ofs, num );
		if( err == kBxErrFailure )
			return kBxErrFailure;
		break;

	case kKindFace:
		err = FaceProc( ofs, num );
		if( err == kBxErrFailure )
			return kBxErrFailure;
		break;
	}

	return kBxErrSuccess;
}

bool  DivObjReadObjMain::
SkipSpace( const char* str, const int num, int& ofs )
{
	while( ofs < num ) {
		if( isspace( str[ ofs ] ) == 0 )
			break;

		ofs++;
	}
	if( ofs == num )
		return true;

	return false;
}

bool  DivObjReadObjMain::
GetNextSpace( const char* str, const int num, int& ofs )
{
	while( ofs < num ) {
		if( isspace( str[ ofs ] ) != 0 )
			break;

		ofs++;
	}
	if( ofs == num )
		return true;

	return false;
}

BxErr  DivObjReadObjMain::
GetBxVec3( const char* str, const int num, BxVec3& dt )
{
	char  dtStr[ kLenDtStr + 1 ];

	int  i   = 0;
	int  cnt = 0;
	while( i < num ) {
		if( SkipSpace( str, num, i ) == true )
			break;

		int  st = i;
		GetNextSpace( str, num, i );
		if( ( i - st ) > kLenDtStr )
			return kBxErrFailure;
		
		memcpy( dtStr, &str[ st ], i - st );
		dtStr[ i - st ] = '\0';

		char*  endptr;
		dt[ cnt ] = strtod( dtStr, &endptr );
		if( endptr != &dtStr[ i - st ] )
			return kBxErrFailure;

		cnt++;
		if( cnt == 3 )
			break;
	}

	if( cnt != 3 )
		return kBxErrFailure;

	return kBxErrSuccess;
}


DivObjReadObjPass0::
DivObjReadObjPass0( const int numMemDt, const char* memDt,
	int& iNumVertex, int& iNumNormal, int& iNumFace, int& iNumFaceVertex )
		: DivObjReadObjMain ( numMemDt, memDt ),

		  fNumVertex     ( iNumVertex ),
		  fNumNormal     ( iNumNormal ),
		  fNumFace       ( iNumFace ),
		  fNumFaceVertex ( iNumFaceVertex )
{
}

BxErr  DivObjReadObjPass0::
Init()
{
	fNumVertex     = 0;
	fNumNormal     = 0;
	fNumFace       = 0;
	fNumFaceVertex = 0;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass0::
VertexProc( const int ofs, const int num )
{
	fNumVertex++;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass0::
NormalProc( const int ofs, const int num )
{
	fNumNormal++;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass0::
FaceProc( const int ofs, const int num )
{
	fNumFace++;

	const char*  str = &fMemDt[ ofs ];

	int  i   = 0;
	int  cnt = 0;
	while( i < num ) {
		if( SkipSpace( str, num, i ) == true )
			break;

		cnt++;

		GetNextSpace( str, num, i );
	}
	if( cnt == 0 )
		return kBxErrFailure;

	fNumFaceVertex += cnt;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass0::
Finish()
{
	return kBxErrSuccess;
}


DivObjReadObjPass1::
DivObjReadObjPass1( const int numMemDt, const char* memDt, const int iNumNormal, BxVec3*& iNormal )
	: DivObjReadObjMain ( numMemDt, memDt ),

	  fNumNormal     ( iNumNormal ),
	  fNormal        ( iNormal )
{
}

BxErr  DivObjReadObjPass1::
Init()
{
	if( fNumNormal <= 0 )
		 return kBxErrFailure;

	fNormal = new BxVec3[ fNumNormal ];

	fCntNormal = 0;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass1::
VertexProc( const int ofs, const int num )
{
	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass1::
NormalProc( const int ofs, const int num )
{
	const char*  str = &fMemDt[ ofs ];

	BxErr  result = GetBxVec3( str, num, fNormal[ fCntNormal ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	fCntNormal++;
	if( fCntNormal > fNumNormal )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass1::
FaceProc( const int ofs, const int num )
{
	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass1::
Finish()
{
	if( fCntNormal != fNumNormal )
		return kBxErrFailure;

	return kBxErrSuccess;
}


DivObjReadObjPass2::
DivObjReadObjPass2( const int numMemDt, const char* memDt,
	const int iNumVertex, const int iNumNormal, const int iNumFace, const int iNumFaceVertex,
	BxVec3*& iNormal, BxObj::FaceSetInfo& iFaceSetInfo )
		: DivObjReadObjMain ( numMemDt, memDt ),

		  fNumVertex     ( iNumVertex ),
		  fNumNormal     ( iNumNormal ),
		  fNumFace       ( iNumFace ),
		  fNumFaceVertex ( iNumFaceVertex ),

		  fNormal        ( iNormal ),
		  fFaceSetInfo   ( iFaceSetInfo )
{
}

BxErr  DivObjReadObjPass2::
Init()
{
	BxObj::FaceSetInfo&  info = fFaceSetInfo;

	if( fNumVertex     <= 0 ||
	    fNumFace       <= 0 ||
	    fNumFaceVertex <= 0 )
	{
		 return kBxErrFailure;
	}

	info.numVertex     = fNumVertex;
	info.numFace       = fNumFace;
	info.numFaceVertex = fNumFaceVertex;

	info.vertex = new BxVec3[ info.numVertex ];
	if( info.numFace > 0 )
		info.faceHeader = new int[ info.numFace - 1 ];
	info.faceVertex = new BxObj::FaceSetInfoFaceVertex[ info.numFaceVertex ];

	fCntVertex     = 0;
	fCntFace       = 0;
	fCntFaceVertex = 0;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass2::
VertexProc( const int ofs, const int num )
{
	BxObj::FaceSetInfo&  info = fFaceSetInfo;
	const char*          str  = &fMemDt[ ofs ];

	BxErr  result = GetBxVec3( str, num, info.vertex[ fCntVertex ] );
	if( result == kBxErrFailure )
		return kBxErrFailure;

	fCntVertex++;
	if( fCntVertex > fNumVertex )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass2::
NormalProc( const int ofs, const int num )
{
	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass2::
FaceProc( const int ofs, const int num )
{
	char  dtStr[ kLenDtStr + 1 ];

	BxObj::FaceSetInfo&  info = fFaceSetInfo;
	const char*          str  = &fMemDt[ ofs ];

	int  i   = 0;
	int  cnt = 0;
	while( i < num ) {
		if( SkipSpace( str, num, i ) == true )
			break;

		int  st = i;
		GetNextSpace( str, num, i );
		if( ( i - st ) > kLenDtStr )
			return kBxErrFailure;
		
		memcpy( dtStr, &str[ st ], i - st );
		dtStr[ i - st ] = '\0';

		BxErr  result = FaceSub( dtStr );
		if( result == kBxErrFailure )
			return kBxErrFailure;

		cnt++;
	}

	if( fCntFace < ( fNumFace - 1 ) ) {
		if( fCntFace == 0 )
			info.faceHeader[ fCntFace ] = cnt;
		else
			info.faceHeader[ fCntFace ] = info.faceHeader[ fCntFace - 1 ] + cnt;
	}

	fCntFace++;
	if( fCntFace > fNumFace )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass2::
FaceSub( const char* str )
{
	enum {
		kLenDtStr = 255
	};
	char   dtStr[ kLenDtStr + 1 ];
	char*  endptr;
	int*   dst;
	int    buf;

	BxObj::FaceSetInfo&  info = fFaceSetInfo;

	int  i        = 0;
	int  cntSlush = 0;
	while( str[ i ] != '\0' ) {
		int  st = i;
		while( str[ i ] != '\0' && str[ i ] != '/' )
			i++;

		if( str[ i ] == '/' ) {
			cntSlush++;
			if( cntSlush > 2 || str[ i+1 ] == '\0' )
				return kBxErrFailure;
		}
		else if( cntSlush != 2 )
			return kBxErrFailure;

		if( cntSlush == 2 && str[ i ] == '/' ) {
			i++;
			continue;
		}

		if( ( i - st ) > kLenDtStr )
			return kBxErrFailure;
		
		memcpy( dtStr, &str[ st ], i - st );
		dtStr[ i - st ] = '\0';

		if( str[ i ] == '/' )
			dst = &info.faceVertex[ fCntFaceVertex ].vertexNo;
		else
			dst = &buf;

		*dst = strtol( dtStr, &endptr, 10 );
		if( endptr != &dtStr[ i - st ] )
			return kBxErrFailure;

		( *dst )--;
			
		if( str[ i ] == '/' ) {
			if( 0 > *dst || *dst >= fNumVertex )
				return kBxErrFailure;
		}
		else {
			if( 0 > *dst || *dst >= fNumNormal )
				return kBxErrFailure;

			BxErr  errResult;
			info.faceVertex[ fCntFaceVertex ].vertexNormal = fNormal[ *dst ].Normalize( &errResult );
			if( errResult == kBxErrFailure )
				return kBxErrFailure;
		}

		if( str[ i ] == '/' )
			i++;
	}

	if( cntSlush != 2 )
		return kBxErrFailure;

	fCntFaceVertex++;
	if( fCntFaceVertex > fNumFaceVertex )
		return kBxErrFailure;

	return kBxErrSuccess;
}

BxErr  DivObjReadObjPass2::
Finish()
{
	if( fCntVertex     != fNumVertex )
		return kBxErrFailure;
	if( fCntFace       != fNumFace )
		return kBxErrFailure;
	if( fCntFaceVertex != fNumFaceVertex )
		return kBxErrFailure;

	return kBxErrSuccess;
}


////////
