
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* divObjReadObj.h */


#if !defined( DIVOBJREADOBJ_H )
#define DIVOBJREADOBJ_H

#include "bxObj.h"



class DivObjReadObj {
private:
	const char*          fFname;
	BxObj::FaceSetInfo&  fFaceSetInfo;

	int                  fNumMemDt;
	char*                fMemDt;

	int                  fNumVertex;
	int                  fNumNormal;
	int                  fNumFace;
	int                  fNumFaceVertex;

	BxVec3*              fNormal;

public:
	DivObjReadObj( const char* fname, BxObj::FaceSetInfo& faceSetInfo );

	BxErr
	Exec();

private:
	void
	Delete( BxVec3* normal );
	BxErr
	FileToMem();
	BxErr
	Pass0();
	BxErr
	Pass1();
	BxErr
	Pass2();
};

class DivObjReadObjMain {
protected:
	enum {
		kLenDtStr = 255
	};

protected:
	int          fNumMemDt;
	const char*  fMemDt;

public:
	DivObjReadObjMain( const int numMemDt, const char* memDt );

	BxErr
	Exec();

protected:
	virtual BxErr
	Init()                                      = 0;
	virtual BxErr
	VertexProc( const int ofs, const int num )  = 0;
	virtual BxErr
	NormalProc( const int ofs, const int num )  = 0;
	virtual BxErr
	FaceProc( const int ofs, const int num )    = 0;
	virtual BxErr
	Finish()                                    = 0;

protected:
	int
	GetNextLine( const int numMemDt, const char* memDt, int& ofs );
	bool
	SkipSpace( const char* str, const int num, int& ofs );
	bool
	GetNextSpace( const char* str, const int num, int& ofs );
	BxErr
	GetBxVec3( const char* str, const int num, BxVec3& dt );

private:
	BxErr
	ExecMain( const int ofs, const int num );
};

class DivObjReadObjPass0 : public DivObjReadObjMain {
private:
	int&  fNumVertex;
	int&  fNumNormal;
	int&  fNumFace;
	int&  fNumFaceVertex;

public:
	DivObjReadObjPass0( const int numMemDt, const char* memDt,
		int& iNumVertex, int& iNumNormal, int& iNumFace, int& iNumFaceVertex );

private:
	virtual BxErr
	Init();
	virtual BxErr
	VertexProc( const int ofs, const int num );
	virtual BxErr
	NormalProc( const int ofs, const int num );
	virtual BxErr
	FaceProc( const int ofs, const int num );
	virtual BxErr
	Finish();
};

class DivObjReadObjPass1 : public DivObjReadObjMain {
private:
	const int  fNumNormal;
	BxVec3*&   fNormal;

	int        fCntNormal;

public:
	DivObjReadObjPass1( const int numMemDt, const char* memDt, const int iNumNormal, BxVec3*& iNormal );

private:
	virtual BxErr
	Init();
	virtual BxErr
	VertexProc( const int ofs, const int num );
	virtual BxErr
	NormalProc( const int ofs, const int num );
	virtual BxErr
	FaceProc( const int ofs, const int num );
	virtual BxErr
	Finish();
};

class DivObjReadObjPass2 : public DivObjReadObjMain {
private:
	const int            fNumVertex;
	const int            fNumNormal;
	const int            fNumFace;
	const int            fNumFaceVertex;

	BxVec3*&             fNormal;
	BxObj::FaceSetInfo&  fFaceSetInfo;

	int                  fCntVertex;
	int                  fCntFace;
	int                  fCntFaceVertex;

public:
	DivObjReadObjPass2( const int numMemDt, const char* memDt,
		const int iNumVertex, const int iNumNormal, const int iNumFace, const int iNumFaceVertex,
		BxVec3*& iNormal, BxObj::FaceSetInfo& iFaceSetInfo );

private:
	virtual BxErr
	Init();
	virtual BxErr
	VertexProc( const int ofs, const int num );
	virtual BxErr
	NormalProc( const int ofs, const int num );
	virtual BxErr
	FaceProc( const int ofs, const int num );
	virtual BxErr
	Finish();

private:
	BxErr
	FaceSub( const char* dtStr );
};


#endif  // !defined( DIVOBJREADOBJ_H )

////////
