
//
// Copyright 2004 Eishin Matsui. All rights reserved.
//
//     first release day : 2004/May/23
//

/* divObjWriteVrml.cpp */


#include "divObjWriteVrml.h"



DivObjWriteVrml::
DivObjWriteVrml( const char* fname, BxObj::FaceSetInfo& faceSetInfo )
	: fFname       ( fname ),
	  fFaceSetInfo ( faceSetInfo )
{
	fFp = 0;
}

BxErr  DivObjWriteVrml::
Exec()
{
	fFp = fopen( fFname, "wt" );
	if( fFp == 0 )
		return kBxErrFailure;

	HeaderProc();
	VertexProc();
	NormalProc();
	FaceProc();
	NormalIndexProc();
	FooterProc();

	fclose( fFp );

	return kBxErrSuccess;
}

void  DivObjWriteVrml::
HeaderProc()
{
	fprintf( fFp, "#VRML V2.0 utf8\n" );
	fprintf( fFp, "\n" );
	fprintf( fFp, "Group {\n" );
	fprintf( fFp, "\tchildren [\n" );
	fprintf( fFp, "\t\tShape {\n" );
	fprintf( fFp, "\t\t\tappearance Appearance {\n" );
	fprintf( fFp, "\t\t\t\tmaterial Material {\n" );
	fprintf( fFp, "\t\t\t\t\tdiffuseColor 0.8 0.8 0.8\n" );
	fprintf( fFp, "\t\t\t\t}\n" );
	fprintf( fFp, "\t\t\t}\n" );
	fprintf( fFp, "\t\t\tgeometry IndexedFaceSet {\n" );
	fprintf( fFp, "\t\t\t\tsolid FALSE\n" );
}

void  DivObjWriteVrml::
VertexProc()
{
	fprintf( fFp, "\t\t\t\tcoord Coordinate {\n" );
	fprintf( fFp, "\t\t\t\t\tpoint [\n" );

	for( int i=0; i<fFaceSetInfo.numVertex; i++ ) {
		fprintf( fFp, "\t\t\t\t\t\t%g %g %g",
			 fFaceSetInfo.vertex[ i ].X(), fFaceSetInfo.vertex[ i ].Y(), fFaceSetInfo.vertex[ i ].Z() );

		if( i < ( fFaceSetInfo.numVertex - 1 ) )
			fprintf( fFp, "," );

		fprintf( fFp, "\n" );
	}

	fprintf( fFp, "\t\t\t\t\t]\n" );
	fprintf( fFp, "\t\t\t\t}\n" );
}

void  DivObjWriteVrml::
NormalProc()
{
	int  num;

	fprintf( fFp, "\t\t\t\tnormal Normal {\n" );
	fprintf( fFp, "\t\t\t\t\tvector [\n" );

	int  cnt = 0;
	for( int i=0; i<fFaceSetInfo.numFace; i++ ) {
		if( i == 0 )
			num = fFaceSetInfo.faceHeader[ 0 ];
		else if( i < ( fFaceSetInfo.numFace - 1 ) )
			num = fFaceSetInfo.faceHeader[ i ] - fFaceSetInfo.faceHeader[ i-1 ];
		else
			num = fFaceSetInfo.numFaceVertex - fFaceSetInfo.faceHeader[ i-1 ];

		for( int j=0; j<num; j++ ) {
			fprintf( fFp, "\t\t\t\t\t\t%g %g %g",
				 fFaceSetInfo.faceVertex[ cnt ].vertexNormal.X(),
				 fFaceSetInfo.faceVertex[ cnt ].vertexNormal.Y(),
				 fFaceSetInfo.faceVertex[ cnt ].vertexNormal.Z() );
			cnt++;

			if( i < ( fFaceSetInfo.numFace - 1 ) || j < ( num - 1 ) )
				fprintf( fFp, "," );

			fprintf( fFp, "\n" );
		}
	}

	fprintf( fFp, "\t\t\t\t\t]\n" );
	fprintf( fFp, "\t\t\t\t}\n" );
}

void  DivObjWriteVrml::
FaceProc()
{
	int  num;

	fprintf( fFp, "\t\t\t\tcoordIndex [\n" );

	int  cnt = 0;
	for( int i=0; i<fFaceSetInfo.numFace; i++ ) {
		if( i == 0 )
			num = fFaceSetInfo.faceHeader[ 0 ];
		else if( i < ( fFaceSetInfo.numFace - 1 ) )
			num = fFaceSetInfo.faceHeader[ i ] - fFaceSetInfo.faceHeader[ i-1 ];
		else
			num = fFaceSetInfo.numFaceVertex - fFaceSetInfo.faceHeader[ i-1 ];

		if( num > 0 )
			fprintf( fFp, "\t\t\t\t\t" );

		for( int j=0; j<num; j++ ) {
			fprintf( fFp, "%d, ", fFaceSetInfo.faceVertex[ cnt ].vertexNo );
			cnt++;
		}

		if( num > 0 ) {
			fprintf( fFp, "-1" );

			if( i < ( fFaceSetInfo.numFace - 1 ) )
				fprintf( fFp, "," );

			fprintf( fFp, "\n" );
		}
	}

	fprintf( fFp, "\t\t\t\t]\n" );
}

void  DivObjWriteVrml::
NormalIndexProc()
{
	int  num;

	fprintf( fFp, "\t\t\t\tnormalIndex [\n" );

	int  cnt = 0;
	for( int i=0; i<fFaceSetInfo.numFace; i++ ) {
		if( i == 0 )
			num = fFaceSetInfo.faceHeader[ 0 ];
		else if( i < ( fFaceSetInfo.numFace - 1 ) )
			num = fFaceSetInfo.faceHeader[ i ] - fFaceSetInfo.faceHeader[ i-1 ];
		else
			num = fFaceSetInfo.numFaceVertex - fFaceSetInfo.faceHeader[ i-1 ];

		if( num > 0 )
			fprintf( fFp, "\t\t\t\t\t" );

		for( int j=0; j<num; j++ ) {
			fprintf( fFp, "%d, ", cnt );
			cnt++;
		}

		if( num > 0 ) {
			fprintf( fFp, "-1" );

			if( i < ( fFaceSetInfo.numFace - 1 ) )
				fprintf( fFp, "," );

			fprintf( fFp, "\n" );
		}
	}

	fprintf( fFp, "\t\t\t\t]\n" );
}

void  DivObjWriteVrml::
FooterProc()
{
	fprintf( fFp, "\t\t\t}\n" );
	fprintf( fFp, "\t\t}\n" );
	fprintf( fFp, "\t]\n" );
	fprintf( fFp, "}\n" );
}


////////
